#!/bin/bash

# For Version 2.2
# This script creates an smbmount two levels deep in the FS root, then attempts to rename a file under it.
# Another variation is creating a share that points to the root of the FS, and defining fs case sensitive = no

SMBCONF="/cluster/config/samba/smb.conf"
TMPSMBCONF="/tmp/smb.conf"
export lockroot="/cluster"
export SAMBA_SERVICE=smbd

#cleanup of possibly old runs
rm -rf /mnt/mgmt/sanity
umount /tmp/test
rm -rf /tmp/test
umount /tmp/test2
rm -rf /tmp/test2

#create new dirs
casedir create sanity
casedir create sanity/rename_test
chmod -R 777 /mnt/mgmt/sanity
mkdir -m 777 /tmp/test
mkdir -m 777 /tmp/test2

su - fse -c "exadmin access shares cifs-shares add rename_test /sanity/rename_test"
su - fse -c "exadmin access shares cifs-shares add rename_test_root /"
su - fse -c "exadmin system protocols cifs general set -security guest -allowguests"

grep -q "fs case sensitive" $SMBCONF
if [ $? ] ; then
	sed 's:fs case sensitive = yes:fs case sensitive = no:' $SMBCONF > $TMPSMBCONF
	cp $TMPSMBCONF $SMBCONF
else
	# The string is not in the smb.conf at all
	sed 's:deadtime = 15:fs case sensitive = no:' $SMBCONF > $TMPSMBCONF
	cp $TMPSMBCONF $SMBCONF
fi

/cluster/config/failover/scripts/etc/init.d/smb_server restart


# Now the test on the two levels-deep rename
smbmount //node0-client/rename_test /tmp/test -o username=guest,password=nopassword
cd /tmp/test
touch test_file1
mv test_file1 test_file2
res=$?
if [ "$res" == "0" ] ; then
	echo Succeeded in Two levels deep rename test. Continuing...
else
	echo Two levels deep rename result: Failure !
	rm -rf test_file1 test_file2
	cd /
	umount /tmp/test
	exit 1
fi
rm -rf test_file1 test_file2
cd /
umount /tmp/test

# Now the test on the root level rename
smbmount //node0-client/rename_test_root /tmp/test2 -o username=guest,password=nopassword
cd /tmp/test2/sanity/rename_test
touch test_file1
mv test_file1 test_file2
res=$?
if [ "$res" == "0" ] ; then
	echo Succeeded in Root levels deep rename test.
else
	echo Root level rename result: Failure !
	rm -rf test_file1 test_file2
	cd /
	umount /tmp/test2
	exit 1
fi
rm -rf test_file1 test_file2
cd /
umount /tmp/test2




