/* 
   wbstub.c 
   Listen to winbind port as a stub. Simply accept connections on the
   relevant port.
   
   (c) 2002, Exanet Inc.
   
   Written by <nirs@exanet.com>
   
*/
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <syslog.h>

#define WINBINDD_SERVICE "winbind"

int mygetservbyname(char *service, char *protocol) {

    struct servent *sp;
    int port;

    sp=(struct servent *)getservbyname(service,protocol);
    if (sp==NULL) {

        if (1==sscanf(service, "%d", &port)) {
            return htons(port);
        } else {
            syslog(LOG_ERR, "MAJOR: getservbyname error for service %s protocol %s\n",service, protocol);
            return -1;
        }
    }

    return sp->s_port;

}

/* Create winbindd socket */

//EXANET
int create_sock(void)
{
		
    int                 listenfd;
    int			port;
    struct sockaddr_in  servaddr;
    const int on=1;

    listenfd = socket(AF_INET, SOCK_STREAM, 0);

    setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family      = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    
    port = mygetservbyname(WINBINDD_SERVICE, "tcp");
    if ( port < 0 ) {
    	syslog(LOG_ERR, "MAJOR: Cannot find winbind service.\n");
	return -1;
    }
    servaddr.sin_port = port;
    
    if (bind(listenfd, (struct sockaddr *) &servaddr, sizeof(servaddr))!=0) {
        syslog(LOG_ERR, "MAJOR: bind failed on winbind socket : %s\n",
                  strerror(errno));
        close(listenfd);
        return -1;
    };

    if (listen(listenfd, 5)!=0) {
         syslog(LOG_ERR, "MAJOR: listen failed on winbind socket : %s\n",
                  strerror(errno));
        close(listenfd);
        return -1;
    }
     /* Success! */
        
    return listenfd;
}

int main() {
    int fd,acceptfd;
    fd = create_sock();
    if (fd>=0) {
	for (;;) {
          acceptfd=accept(fd, 0 ,0);
          close(acceptfd);
	}
    } else {
	syslog(LOG_ERR, "MAJOR: Error occured in create_sock, aborting.\n");
	exit(1);
    }
}
