#include "includes.h"
#include "exasmb_if.h"
extern int              last_lock_counter;
static const char *prtrace();



static void exasmb_if_share_check_error(exasmb_message_error err, const char * what, uint64_t dev, uint64_t inode)
{
	switch (err) {
	case EXASMB_OK:
		return;
	case EXASMB_FATAL_ERROR:
	case EXASMB_LRLOCK_ERROR:
	case EXASMB_ERROR:
		DEBUG(0, ("%s failed: err %d dev %u inode %llx %s - Terminating\n",
			  what, err, (unsigned) dev, (long long) inode, prtrace()));
		break;
	default:
		DEBUG(0, ("UNKNOWN ERROR - %s failed: err %d dev %u inode %llx %s - Terminating\n",
			  what, err, (unsigned) dev, (long long) inode, prtrace()));
		break;
	}

	smb_panic(what);
}

static BOOL
exasmb_if_share_lock_lr_int(char* nfs_fh,unsigned int nfs_fh_len, uint32_t pid, uint16_t tid, uint64_t dev, uint64_t inode, unsigned long share_file_id, BOOL do_lock)
{
	exasmb_share_lock_lr_request request;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(pid, tid, nfs_fh, nfs_fh_len, dev, inode, share_file_id, request.lr, False);

	request.lock_count = last_lock_counter;
	
	if (do_lock) {
		verify_no_double_lock (dev,inode);
	}
	
	

	do {
		err = do_lock
			? exasmb_share_lock_lr(client, &request)
			: exasmb_share_unlock_lr(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_share_check_error(err, do_lock ? "exasmb_share_lock_lr" : "exasmb_share_unlock_lr", dev, inode);

	if (do_lock) {
		exasmb_lock_increase(dev,inode);
	} else {
		exasmb_lock_decrease(dev,inode);
	}

	return True;
}

BOOL
exasmb_if_share_lock_lr(char* nfs_fh,unsigned int nfs_fh_len, uint32_t pid, uint16_t tid, uint64_t dev, uint64_t inode, unsigned long share_file_id)
{
	DEBUG(10, ("exasmb_if_share_lock_lr\n"));

	return exasmb_if_share_lock_lr_int(nfs_fh,nfs_fh_len,pid, tid, dev, inode, share_file_id, True);
}

BOOL
exasmb_if_share_unlock_lr(char* nfs_fh,unsigned int nfs_fh_len, uint32_t pid, uint16_t tid, uint64_t dev, uint64_t inode, unsigned long share_file_id)
{
	DEBUG(10, ("exasmb_if_share_unlock_lr\n"));
	
	return exasmb_if_share_lock_lr_int(nfs_fh,nfs_fh_len, pid, tid, dev, inode, share_file_id, False);
}

static void
unpack_share_entry(exasmb_shared_lock *exasmb_share_entry, struct share_mode_entry *share_entry)
{
	share_entry->pid = node_pid_to_procid(exasmb_share_entry->node, exasmb_share_entry->pid);
	share_entry->smbid = exasmb_share_entry->smbid;
	share_entry->op_mid = exasmb_share_entry->op_mid;
	share_entry->access_mask = exasmb_share_entry->access_mask;
	share_entry->op_type = exasmb_share_entry->op_type;
	share_entry->dev = exasmb_share_entry->dev;
	share_entry->inode = exasmb_share_entry->inode;
	share_entry->share_file_id = exasmb_share_entry->share_file_id;
	share_entry->share_access = exasmb_share_entry->share_access;
	share_entry->time.tv_usec = exasmb_share_entry->open_time.tv_usec;
	share_entry->time.tv_sec = exasmb_share_entry->open_time.tv_sec;
	share_entry->private_options = exasmb_share_entry->private_options;
	share_entry->uid = exasmb_share_entry->uid;
	share_entry->flags = exasmb_share_entry->flags;
}

int32_t
exasmb_if_share_get_locks(char* nfs_fh,unsigned int nfs_fh_len, uint32_t pid, uint16_t tid,  uint64_t dev, uint64_t inode, uint32_t file_id,
			  struct share_mode_entry **pp_shares, uint32_t num_of_entries,
			  uint64_t cookie_in, uint64_t *cookie_out, BOOL *delete_on_close)
{
	exasmb_share_get_locks_request request;
	exasmb_share_get_locks_reply reply;
	dlm_client client;
	int num_share_modes = -1;
	struct share_mode_entry *shares;
	exasmb_message_error err;
	int i, j;

	*delete_on_close = False;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(pid, tid, nfs_fh, nfs_fh_len, dev, inode, file_id, request.lr, -1);
	request.cookie = cookie_in;
	request.num_locks = MIN(num_of_entries, MAX_FETCH_NUM_ENTRIES);

	do {
		reply.body.entries = NULL;

		err = exasmb_share_get_locks(client, &request, &reply);
	} while( err == EXASMB_COMM_ERROR);
	exasmb_if_share_check_error(err, "exasmb_if_share_get_locks", dev, inode);

	num_share_modes = reply.header.num_entries;
	*delete_on_close = (reply.header.delete_on_close == 0) ? False : True;

	DEBUG(3, ("num_share_modes is %d\n", num_share_modes));
	DEBUG(3, ("%s: delete_on_close is %d\n", prtrace(), *delete_on_close));

	j = 0;
	if (num_share_modes > 0) {
		/* Paranoia */
		if (!reply.body.entries) {
			DEBUG(1,("No entries returned in reply"));
			return -1;
		}

		shares = *pp_shares = malloc(num_share_modes * sizeof(struct share_mode_entry));
		if (!*pp_shares)
			return -1;

		for (i = 0; i < num_share_modes; i++) {
			/* !!! lame process existance code was here. it did not work cross-cluster */

			unpack_share_entry(reply.body.entries + i, shares + j);
			j++;
		}

		*cookie_out = reply.header.cookie;
	}

	return j;
}

BOOL
exasmb_if_share_add_lock( files_struct *fsp, const struct share_mode_entry *entry)
{
	exasmb_share_add_lock_request request;
	exasmb_shared_lock *plock = NULL;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(entry->pid.pid, fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.header.lr, False);
	request.header.num_entries = 1;

	plock = (exasmb_shared_lock *) exasmb_dlm_client_get_req_buf(client, sizeof(exasmb_shared_lock));

	if(!plock) {
		DEBUG(0, ("Unable to allocate request buffer"));
		return False;
	}

	request.body.entries = plock;

	plock->pid = entry->pid.pid;
	plock->node = entry->pid.node;
	plock->smbid = exasmb_getsmbid();
	plock->op_mid = entry->op_mid;
	plock->op_type = entry->op_type;
	plock->share_access = fsp->share_access;
	plock->access_mask = fsp->access_mask;
	memcpy(&plock->open_time, &fsp->open_time, sizeof(struct timeval));
	plock->dev = fsp->dev;
	plock->inode = fsp->inode;
	plock->share_file_id = fsp->fh->file_id;
	plock->private_options = entry->private_options;
	plock->uid = entry->uid;
	plock->flags = entry->flags;

	do {
		err = exasmb_share_add_lock(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_share_check_error(err, "exasmb_share_add_lock", fsp->dev, fsp->inode);

	exasmb_dlm_client_release_req_buf(client, plock);

	return True;
}

BOOL
exasmb_if_share_delete_lock( files_struct *fsp, const struct share_mode_entry *entry)
{
	exasmb_share_delete_lock_request request;
	exasmb_shared_lock *pdel_entry = NULL;
	exasmb_share_delete_lock_reply reply;
	dlm_client client;
	exasmb_message_error err;

	/* !!!
	if (( entry->pid.pid != pid ) ||
		( entry->smbid != exasmb_getsmbid() ) ||
		( entry->dev != dev ) ||
		( entry->inode != inode ) ||
		( entry->share_file_id != fsp->fh->file_id )) {
		DEBUG(3,("Requesting a delete lock on a lock that's not mine. Returning error.\n"));

		DEBUG(3,("My data   : pid=%u, smbid=%u dev=%llu, inode=%llu, file_id=%lu .\n",
					pid,exasmb_getsmbid(),dev,inode,fsp->fh->file_id));

		DEBUG(3,("Their data: pid=%u, smbid=%u dev=%llu, inode=%llu, file_id=%lu\n",
					entry->pid,entry->smbid,entry->dev,entry->inode,entry->share_file_id));
		return -1;
	}*/

	client = exasmb_if_client_open();

	/* *delete_on_close = False; */
	PACK_LR_CONTEXT(entry->pid.pid, fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH , fsp->dev, fsp->inode, entry->share_file_id, request.header.lr, False);

	request.header.num_entries = 1;

	pdel_entry = (exasmb_shared_lock *) exasmb_dlm_client_get_req_buf(client, sizeof(exasmb_shared_lock));

	if(!pdel_entry) {
		DEBUG(0, ("Unable to allocate request buffer"));
		return False;
	}

	request.body.entries = pdel_entry;
	memset(pdel_entry, 0, sizeof(exasmb_shared_lock));

	pdel_entry->smbid = exasmb_getsmbid();
	pdel_entry->pid = entry->pid.pid;
	pdel_entry->node = entry->pid.node;
	pdel_entry->dev = fsp->dev;
	pdel_entry->inode = fsp->inode;
	pdel_entry->share_file_id = entry->share_file_id;
	pdel_entry->share_access = entry->share_access;

	do {
		err = exasmb_share_delete_lock(client, &request, &reply);
	} while ( err == EXASMB_COMM_ERROR );

	exasmb_if_share_check_error(err, "exasmb_share_delete_lock", fsp->dev, fsp->inode);

	/*
	rc = reply.num_left;
	*delete_on_close = (reply.delete_on_close) ? True : False;
	DEBUG(3, ("%s: delete_on_close is %d\n", prtrace(), *delete_on_close));

	if (ppse) {
		if (reply.locks_were_deleted) {

			SAFE_FREE(*ppse);
			*ppse = malloc(sizeof(struct share_mode_entry));
			if (!*ppse) {
				DEBUG(0,("Unable to allocate share entry"));
				rc = -1;
			} else {
				unpack_share_entry(&reply.lock_deleted, *ppse);
			}
		}
	}*/

	exasmb_dlm_client_release_req_buf(client, pdel_entry);

	return True;
}

BOOL
exasmb_if_share_modify_oplock( files_struct *fsp, const struct share_mode_entry *entry)
{
	exasmb_share_modify_oplock_request request;
	exasmb_shared_lock *plock = &request.entry;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	PACK_LR_CONTEXT(entry->pid.pid, fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.lr, False);
	request.set_port = 0; /* !!! unused */

	plock->pid = entry->pid.pid;
	plock->node = entry->pid.node;
	plock->smbid = exasmb_getsmbid();
	plock->op_mid = entry->op_mid;
	plock->op_type = entry->op_type;
	plock->share_access = fsp->share_access;
	plock->access_mask = fsp->access_mask;
	memcpy(&plock->open_time, &fsp->open_time, sizeof(struct timeval));
	plock->dev = fsp->dev;
	plock->inode = fsp->inode;
	plock->share_file_id = fsp->fh->file_id;
	plock->private_options = entry->private_options;
	plock->uid = entry->uid;
	plock->flags = entry->flags;

	do {
		err = exasmb_share_modify_oplock(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_share_check_error(err, "exasmb_share_modify_oplock", fsp->dev, fsp->inode);

	return True;
}

BOOL
exasmb_if_share_set_delete_on_close( files_struct *fsp, BOOL delete_on_close)
{
	exasmb_share_set_delete_on_close_request request;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();
	PACK_LR_CONTEXT(sys_getpid(), fsp->conn->cnum, fsp->nfs_fh, NFS_FH_LENGTH, fsp->dev, fsp->inode, fsp->fh->file_id, request.lr, False);

	request.delete_on_close = delete_on_close;

	do {
		err = exasmb_share_set_delete_on_close(client, &request);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_share_check_error(err, "exasmb_share_set_delete_on_close", fsp->dev, fsp->inode);

	return True;
}

#if 0
BOOL exasmb_if_share_get_delete_on_close(uint32_t pid, uint16_t tid,  uint64_t dev, uint64_t inode,
					 unsigned long file_id, BOOL* delete_on_close)
{
	exasmb_share_get_delete_on_close_request request;
	exasmb_share_get_delete_on_close_reply reply;
	dlm_client client;
	exasmb_message_error err;

	client = exasmb_if_client_open();

	*delete_on_close = False;

	PACK_LR_CONTEXT(pid, tid, fsp->nfs_fh, NFS_FH_LENGTH, dev, inode, file_id, request.lr, -1);

	do {
		err = exasmb_share_get_delete_on_close(client, &request, &reply);
	} while (err == EXASMB_COMM_ERROR);

	exasmb_if_share_check_error(err, "exasmb_share_get_delete_on_close", dev, inode);

	*delete_on_close = (reply.delete_on_close) ? True : False;

	return True;
}
#endif

#include <execinfo.h>

static const char *prtrace()
{
	#ifdef sprintf
	#	undef sprintf
	#endif
	static char s[5*(11)+1];
	static void *trace[5];
	int i, len, n = backtrace(trace, 5);
	s[0] = 0;
	for (i = len = 0; i < n; ++i) {
		len += sprintf(&s[len], " %p", trace[i]);
	}
	return s;
}
