#include <iostream>
#include <string>
// for select
#include <sys/time.h> 
#include <sys/types.h> 
#include <unistd.h> 

#include "ListenerMgr.h"

extern bool verbose;

	using namespace std;

	ListenerMgr::ListenerMgr(int baseDomain,
							 int step,
							 int noOfDomains,
							 char *serverPrefix) 
	: _baseDomain(baseDomain)
	, _step(step)
	, _noOfDomains(noOfDomains)
	, _serverPrefix(serverPrefix)
	, _maxFd(-1)
	{
	}

	// Create listener objects
	void ListenerMgr::open() {
		DBG(cout << "Creating listeners per domain." << endl);
		int currentDomain = 0;
		char currentDomainName[SERVER_PREFIX_SIZE + 5];
		SambaEamListener *sambaEamListener = NULL;
		int currFd = -1;
		for (int i = 0; i < _noOfDomains; ++i) {
			currentDomain = _baseDomain + (i * _step);
			sprintf(currentDomainName, "%s.%d", _serverPrefix.c_str(), currentDomain);
			DBG(cout << " Creating domain " << currentDomainName << endl);

			sambaEamListener = new SambaEamListener(currentDomainName);
			try {
				currFd = sambaEamListener->open();
				_listenerMap[currFd] = sambaEamListener;
			} catch (...) {
				cout << "BADBAD" << endl;
			}
		}
	}

	// reset _fdset and _maxFd
	void ListenerMgr::scanFds() {

		FD_ZERO(&_fdset);
		_maxFd = -1;

		// Go over all FDs to find MAX and put in set
		BasicListenerIter listenerIter(_listenerMap.begin()),
						  listenerIterEnd(_listenerMap.end());

		for (; listenerIter != listenerIterEnd; ++listenerIter) {
			FD_SET(listenerIter->first, &_fdset);
			if (listenerIter->first > _maxFd) {
				_maxFd = listenerIter->first;
			}
		}

		BasicConnectionIter connectionIter(_connectionMap.begin()),
						    connectionIterEnd(_connectionMap.end());

		for (; connectionIter != connectionIterEnd; ++connectionIter) {
			FD_SET(connectionIter->first, &_fdset);
			if (connectionIter->first > _maxFd) {
				_maxFd = connectionIter->first;
			}
		}
	}

	// Main loop - select and call the appropriate listener (new connection) or connection (new message)
	void ListenerMgr::run() {
		
		// Go over all FDs to find MAX and put in set
		BasicListenerIter listenerIter; 
		BasicListenerIter listenerIterEnd;

		BasicConnectionIter connectionIter;
		BasicConnectionIter connectionIterEnd;

		int noFdsReady;
		int removedFd;

		while (true) {

			// reset _maxFd and _fdset
			scanFds();

			// Lister on all valid FDS
			DBG(cout << "running select on all open listeners" << endl;);
			noFdsReady = select(_maxFd + 1, &_fdset, NULL, NULL, NULL);
			DBG(cout << "select jumped with " << noFdsReady << " sockets ready" << endl;);

			// go over all connections and see if any one jumped.
			connectionIter = _connectionMap.begin();
			connectionIterEnd = _connectionMap.end();

			// check if new message for existing connection
			for (int i = 0; connectionIter != connectionIterEnd && i < noFdsReady; ++connectionIter) {
				if (FD_ISSET(connectionIter->first, &_fdset)) {
					DBG(cout << "connection listener jumped on fd " << connectionIter->first << endl);
					++i;
					exasmb_message_error ret = connectionIter->second->handleMessage();
					if (ret == EXASMB_FATAL_ERROR) { // currently used as connection closed
						_removalList.push_back(connectionIter->first);
						close(connectionIter->first);
					} else if (ret == EXASMB_COMM_ERROR) {
						cerr << "Error getting message" << endl;
					}
				}
			}

			// clear FDs that were closed from the Maps
			while (!_removalList.empty()) {
				removedFd = _removalList.back();
				_removalList.pop_back();
				if ((connectionIter = _connectionMap.find(removedFd)) == connectionIterEnd) {
					cerr << "internal error finding removed Fd during clean up after closed connection" << endl;
				} else {
					DBG(cout << "Removing " << removedFd << endl);
					_connectionMap.erase(removedFd);
				}
			}

			// go over all listeners and see if new connections arrive
			listenerIter = _listenerMap.begin();
			listenerIterEnd = _listenerMap.end();

			// check if new connection
			for (int i = 0; listenerIter != listenerIterEnd && i < noFdsReady; ++listenerIter) {
				if (FD_ISSET(listenerIter->first, &_fdset)) {
					DBG(cout << "listener " << listenerIter->second->getPortName() << " jumped" << endl);
					++i;

					int newFd = listenerIter->second->newConnection();
					_connectionMap[newFd] = new SambaEamConnection(newFd);
				}
			}
		} // end while true
	}

