#include <iostream>
#include <string.h>

#include <errno.h>

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>

#include "SambaEamConnection.h"
#include "smb_acl_message.h"
#include "unixntacl.h"
#include "exasmb_vfsfs_message.h"

	using namespace std;

	// ctor - new connection identified by fd
	SambaEamConnection::SambaEamConnection(int fd) 
	: BasicConnection(fd)
	, rcved(0)
	, left(0)
	, ctrl_buf(NULL)
	, ancillary_fd(-1)
	{
		ctrl_buf = new uint32_t[4096/sizeof(uint32_t)];
	}

	SambaEamConnection::~SambaEamConnection() {
		delete[] ctrl_buf;
		ctrl_buf = NULL;
	}

	// handle incoming message,
	// includes reading, decoding, processing, encoding and replying.
	exasmb_message_error SambaEamConnection::handleMessage() {

		// read using recvmsg
 
		// read the incoming message in varialbe length format
		ssize_t n = read(&left, sizeof(left)); // read the length of the message, classic variable length
		if (n < 0) {
			cerr << "Error reading from fd " << _fd << " with error " << strerror(errno) << endl;
			rcved = left = 0;
			_buf = NULL;
			return EXASMB_COMM_ERROR;
		} else if (n == 0) {
			DBG(cout << "connection closed" << endl);
			rcved = left = 0;
			_buf = NULL;
			return EXASMB_FATAL_ERROR; // == connection closed
		} // else ...

		// left contains the number of the bytes in the message (without left itself)
		left = ntohl(left);
		for (rcved = 0; left; left -= n, rcved += n) {
			n = read(&ctrl_buf[rcved], left);
			if (n < 0) {
				return EXASMB_COMM_ERROR;
			}
		}

		// decode message
		smb_acl_request	req;
		const uint32_t *p = decode_request(&req, ctrl_buf);

		switch (req.type) {
		case MESSAGE_TYPE_REQUEST:
			switch (req.code) {
				case MESSAGE_CODE_GET_ACL:
				{
					cout << "MESSAGE_CODE_GET_ACL" << endl;
					//get_acl(req, p);

					// Send reply to exasmb client
					smb_acl_get_acl_reply rep;
					unixntacl_acl acl;
					unixntacl_acl_ace ace;

					ace.type = ACCESS_ALLOWED_ACE_TYPE;
					ace.flags = 0;
					ace.mask = 0x1f01ff; // FULL CONTROL
					ace.trustee.type = trustee_world;
					ace.trustee.id = 0;

					acl.type = 0x9404;
						/*SEC_DESC_DACL_PRESENT | SE_DESC_DACL_AUTO_INHERITED | SE_DESC_DACL_PROTECTED | SEC_DESC_SELF_RELATIVE */
					acl.ace_count = 1;
					acl.aces = &ace;

					rep.error = SMBACL_MESSAGE_OK;
					rep.acl = &acl;

					p = encode_get_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_SET_ACL:
				{
					cout << "MESSAGE_CODE_SET_ACL" << endl;
					//set_acl(req, p);

					// Send reply to exasmb client
					smb_acl_set_acl_reply rep;
					
					rep.error = SMBACL_MESSAGE_OK;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_GET_ACCESS:
				{
					cout << "MESSAGE_CODE_GET_ACCESS" << endl;
					//get_access(req, p);

					// Send reply to exasmb client
					smb_acl_get_access_reply rep;
					
					rep.error = SMBACL_MESSAGE_OK;
					rep.access = 0x1f01ff; // FULL CONTROL
					rep.access_troubleshooting_details[0] = 0; // Empty access details string
					uint32_t *p = encode_get_access_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_SENDFILE:
				{
					cout << "MESSAGE_CODE_SENDFILE" << endl;
					//sendfile(req, p);

					smb_vfs_sendfile_reply rep;
					rep.error = SMBACL_MESSAGE_NOT_SUPPORTED;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_SHWRITE:
				{
					cout << "MESSAGE_CODE_SHWRITE" << endl;
					//shwrite(req, p);

					smb_vfs_shwrite_reply rep;
					rep.error = SMBACL_MESSAGE_NOT_SUPPORTED;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_RENAME:
				{
					cout << "MESSAGE_CODE_RENAME" << endl;
					//rename(req, p);

					smb_vfs_rename_reply rep;
					rep.error = SMBACL_MESSAGE_NOT_SUPPORTED;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				case MESSAGE_CODE_FSYNC:
				{
					cout << "MESSAGE_CODE_FSYNC" << endl;
					//fsync(req, p);

					smb_vfs_fsync_reply rep;
					rep.error = SMBACL_MESSAGE_NOT_SUPPORTED;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;		
				}
				case MESSAGE_CODE_DIRPLUS:
				{
					cout << "MESSAGE_CODE_DIRPLUS" << endl;
					//dirplus(req, p);

					smb_vfs_dirplus_reply rep;
					rep.error = SMBACL_MESSAGE_NOT_SUPPORTED;
					p = encode_set_acl_reply(&rep, ctrl_buf);
					send((char *) p - (char *) ctrl_buf,
						 ctrl_buf);	
					break;
				}
				default:
					cerr << "Unknown message request code: " << req.code << endl;
					rcved = left = 0;
					_buf = NULL;
					ancillary_fd = -1;
					return EXASMB_FATAL_ERROR;
				}
			break;
		case MESSAGE_TYPE_REPLY:
			cerr << "Message reply recieved when expecting request: " << req.type << endl;
			rcved = left = 0;
			_buf = NULL;
			ancillary_fd = -1;
			return EXASMB_FATAL_ERROR;
		}
		rcved = left = 0;
		_buf = NULL;
		ancillary_fd = -1;
		return EXASMB_OK;
	}

	int SambaEamConnection::read(void *buf, size_t count) { // variable len + recvmsg
		
		char cmsg_buf[CMSG_SPACE(sizeof(int))]; // in case we need it
	
		size_t nread = 0; 
		
		do {
			struct iovec vec;
			vec.iov_base = buf;
			vec.iov_len = count;
			struct msghdr msg;
			msg.msg_name = NULL;
			msg.msg_namelen = 0;
	
			msg.msg_control = cmsg_buf;
			msg.msg_controllen = sizeof cmsg_buf;
			
			msg.msg_flags = 0;
			msg.msg_iov = &vec;
			msg.msg_iovlen = 1;
			int n = recvmsg(_fd, &msg, 0);
			if (n <= 0) {
				return n;
			}
			if ( msg.msg_flags & MSG_CTRUNC) {
				// increase file limit ans such
				return -1;
			}
			if (msg.msg_controllen) {
				cmsghdr *cmsg = CMSG_FIRSTHDR(&msg);			
				if  ( cmsg && ( cmsg->cmsg_len == CMSG_LEN(sizeof(int)) &&
					 cmsg->cmsg_level == SOL_SOCKET &&
					 cmsg->cmsg_type == SCM_RIGHTS)) 
				{
					int tmp_ancillary_fd; // allows cleanup for a complete implementation
					memcpy(&tmp_ancillary_fd, CMSG_DATA(cmsg), sizeof(int));
					if((ancillary_fd != tmp_ancillary_fd) && (tmp_ancillary_fd != -1)) {
						DBG(cout << "got CMSG: old ancillary fd " << ancillary_fd << " new acilliary fd " << tmp_ancillary_fd << endl);
						if(ancillary_fd != -1) {
							// cleanup
						}
						ancillary_fd = tmp_ancillary_fd; 
					}			    				
				}
			}
			buf = static_cast<char*>(buf) + n;
			nread += n;
			count -= n;
		} while ( count );	
		return nread;
	}


/*
	void SambaElmConnection::reply_ok(xdr_func xdr_ctl, 
									  void *p, 
									  xdr_func xdr_dat)
	{
		header.message_type = EXASMB_MESSAGE_REPLY;
		header.message_error = EXASMB_OK;
		
		uint32_t *cp = enc_exasmb_message_header(&header, ctrl_buf);
		if (xdr_ctl)
			cp = xdr_ctl(p, cp);
		if (xdr_dat)
			cp = xdr_dat(p, cp);
		send((char *) cp - (char *) ctrl_buf);	
	}

	void
	SambaElmConnection::decode_message(uint32_t *p)
	{   	
		switch (header.message_code) {	
		case EXASMB_CONNECT_NLM:
			decode_connect(p);
			break;
		case EXASMB_CLEANUP:
			decode_cleanup(p);
			break;
		case EXASMB_SHARE_LOCK_LR:
			decode_share_lock_lr(p);
			break;
		case EXASMB_SHARE_UNLOCK_LR:
			decode_share_unlock_lr(p);
			break;
		case EXASMB_SHARE_GET_LOCKS:
			decode_share_get_locks(p);
			break;
		case EXASMB_SHARE_ADD_LOCK:
			decode_share_add_lock(p);
			break;
		case EXASMB_SHARE_DELETE_LOCK:
			decode_share_delete_lock(p);
			break;

		case EXASMB_SHARE_MODIFY_OPLOCK:
			decode_share_modify_oplock(p);
			break;
		case EXASMB_SHARE_SET_DELETE_ON_CLOSE:
			decode_share_set_delete_on_close(p);
		case EXASMB_SHARE_GET_DELETE_ON_CLOSE:
			decode_share_get_delete_on_close(p);
			break;
		case EXASMB_BR_LOCK_LR:
			decode_br_lock_lr(p);
			break;
		case EXASMB_BR_UNLOCK_LR:
			decode_br_unlock_lr(p);
			break;
		case EXASMB_BR_GET_LOCKS:
			decode_br_get_locks(p);
			break;
		case EXASMB_BR_CHK_LOCK:
			decode_br_chk_lock(p);
			break;
		case EXASMB_BR_ADD_LOCK:
			decode_br_add_lock(p);
			break;
		case EXASMB_BR_DELETE_LOCK:
			decode_br_delete_lock(p);
			break;
		case EXASMB_BR_FILE_CLOSE:
			decode_br_file_close(p);
			break;
		default:
			throw std::exception();
		}
	}

	uint32_t *
	SambaElmConnection::decode_request(uint32_t *buf,
									void *req,
									xdr_func xdr_ctl,
									xdr_func xdr_dat)
	{
		uint32_t *p = xdr_ctl(req, buf);
		if (xdr_dat) {
			if (conn_type != CONN_TYPE_DIRECT)
				p = xdr_dat(req, p);
			else {
				if (shmid == -1 || !data_buf){
					//throw std::exception("Bad Data Buffer");
					cerr << "CRITICAL: Bad Data Buffer" << endl;
				}
				struct typical_vec_repuest {
					exasmb_request_vec_header header;
					struct typical_body {
						void	*entries;
					} body;
				};
				((typical_vec_repuest *) req)->body.entries = data_buf;
			}
		}
		return p;
	}

	void
	SambaElmConnection::decode_connect(uint32_t *p)
	{
		exasmb_connect_request* request = static_cast<exasmb_connect_request*>
			(_decoded_message.reset( sizeof(exasmb_connect_request)));

		p = dec_exasmb_connect_request(request, p);	
	}

	void
	SambaElmConnection::decode_cleanup(uint32_t *p)
	{
		exasmb_cleanup_request* request = static_cast<exasmb_cleanup_request*>
			(_decoded_message.reset( sizeof(exasmb_cleanup_request)));
		p = dec_exasmb_cleanup_request(request, p);	
	}

	void
	SambaElmConnection::decode_share_lock_lr(uint32_t *p)
	{
		exasmb_share_lock_lr_request* request = static_cast<exasmb_share_lock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_share_lock_lr_request)));
		dec_exasmb_share_lock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_share_unlock_lr(uint32_t *p)
	{
		exasmb_share_unlock_lr_request* request = static_cast<exasmb_share_unlock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_share_unlock_lr_request)));
		dec_exasmb_share_unlock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_share_get_locks(uint32_t *p)
	{
		exasmb_share_get_locks_request* request = static_cast<exasmb_share_get_locks_request*>
		(_decoded_message.reset( sizeof(exasmb_share_get_locks_request)));
		dec_exasmb_share_get_locks_request(request, p);
	}

	void
	SambaElmConnection::decode_share_add_lock(uint32_t *p)
	{
		exasmb_share_add_lock_request* request = static_cast<exasmb_share_add_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_add_lock_request)));

		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_share_add_lock_request_header,
						(xdr_func) &dec_exasmb_share_add_lock_request_body);
	}

	void
	SambaElmConnection::decode_share_delete_lock(uint32_t *p)
	{
		exasmb_share_delete_lock_request* request = static_cast<exasmb_share_delete_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_delete_lock_request)));
		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_share_delete_lock_request_header,
						(xdr_func) &dec_exasmb_share_delete_lock_request_body);
	}

	void
	SambaElmConnection::decode_share_modify_oplock(uint32_t *p)
	{
		exasmb_share_modify_oplock_request* request = static_cast<exasmb_share_modify_oplock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_modify_oplock_request)));
		dec_exasmb_share_modify_oplock_request(request, p);
	}

	void
	SambaElmConnection::decode_br_lock_lr(uint32_t *p)
	{
		exasmb_br_lock_lr_request* request = static_cast<exasmb_br_lock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_br_lock_lr_request)));
		dec_exasmb_br_lock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_br_unlock_lr(uint32_t *p)
	{
		exasmb_br_unlock_lr_request* request = static_cast<exasmb_br_unlock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_br_unlock_lr_request)));
		dec_exasmb_br_unlock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_br_get_locks(uint32_t *p)
	{
		exasmb_get_locks_request* request = static_cast<exasmb_get_locks_request*>
		(_decoded_message.reset( sizeof(exasmb_get_locks_request)));
		dec_exasmb_get_locks_request(request, p);	
	}

	void
	SambaElmConnection::decode_br_chk_lock(uint32_t *p)
	{
		exasmb_br_chk_lock_request* request = static_cast<exasmb_br_chk_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_chk_lock_request)));

		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_chk_lock_request_header,
						(xdr_func) &dec_exasmb_br_chk_lock_request_body);	
	}


	void
	SambaElmConnection::decode_share_get_delete_on_close(uint32_t *p)
	{
		exasmb_share_get_delete_on_close_request* request = static_cast<exasmb_share_get_delete_on_close_request*>
		(_decoded_message.reset( sizeof(exasmb_share_get_delete_on_close_request)));		
		dec_exasmb_share_get_delete_on_close_request(request, p);	
	}
	void

	SambaElmConnection::decode_share_set_delete_on_close(uint32_t *p)
	{
		exasmb_share_set_delete_on_close_request* request = static_cast<exasmb_share_set_delete_on_close_request*>
		(_decoded_message.reset( sizeof(exasmb_share_set_delete_on_close_request)));		
		dec_exasmb_share_set_delete_on_close_request(request, p);
	}

	void
	SambaElmConnection::decode_br_add_lock(uint32_t *p)
	{
		exasmb_br_add_lock_request* request = static_cast<exasmb_br_add_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_add_lock_request)));
		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_add_lock_request_header,
						(xdr_func) &dec_exasmb_br_add_lock_request_body);	
	}

	void
	SambaElmConnection::decode_br_delete_lock(uint32_t *p)
	{
		exasmb_br_delete_lock_request* request = static_cast<exasmb_br_delete_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_delete_lock_request)));

		request->body.entries = NULL;
		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_delete_lock_request_header,
						(xdr_func) &dec_exasmb_br_delete_lock_request_body);	
	}

	void
	SambaElmConnection::decode_br_file_close(uint32_t *p)
	{
		exasmb_file_close_request* request = static_cast<exasmb_file_close_request*>
		(_decoded_message.reset( sizeof(exasmb_file_close_request)));	
		dec_exasmb_file_close_request(request, p);
	}
*/
/*
	SambaEamConnection::auto_malloc_ptr::auto_malloc_ptr():
	_ptr(NULL)
	{

	}

	SambaEamConnection::auto_malloc_ptr::~auto_malloc_ptr()
	{
		free(_ptr);
	}

	void* 
	SambaEamConnection::auto_malloc_ptr::reset(size_t size)
	{
		free(_ptr);
		_ptr = malloc(size);
		return _ptr;
	}

	void* 
	SambaEamConnection::auto_malloc_ptr::get()
	{
		return _ptr;
	}
*/
