#include <iostream>
#include <stdlib.h>
#include <unistd.h>

#include "ListenerMgr.h"

#define DBG(x) if (verbose) {x;}
#define SERVER_PREFIX_SIZE 20
bool	verbose;

using namespace std;

/* provide usage info and exit */
int usage(char **argv){
	cerr << "Usage: " << argv[0] << " -b BaseDomain -s Step -n #Domains [-p serverPrefix] [-i] [-v] [-h]" << endl
		 << "  [-p serverPrefix] --> the server prefix will be the supplied prefix, default is eam" << endl
		 << "  [-i] --> interactive, don't become a deamon" << endl
		 << "  [-v] --> verbose, extended info" << endl
		 << "  [-h] --> Show this help screen" << endl
		 << "The listeners will work on unix sockets named serverPrefix.0 etc"
		 << "Example: " << argv[0] << " -b 0 -s 2 -n 8--> will listen on  0 2 4 6 8 10 12 and 14" << endl
		 << "Example: " << argv[0] << " -b 8 -s 2 -n 4--> will listen on 8 10 12 and 14" << endl;

    exit(1);
}

//////////////////////////////////////////////////////////////////////////////////////////////
//																							//
// External ACL Manager skeleton server														//
// ---------------------------------------													//
// Writen by Tsiyon Sadiky from Exanet														//
// Based on work done (mainly) by Nadav Danieli from Exanet									//
// 																							//
// The purpose of this skeleton server is to demonstrate the use of the External ACL		//
// mechanism that was developed at Exanet in order to support proprietry ACL implementation	//
// The mechnism includes additional File-System-API infrastructure for a proprietry FS		//
// 																							//
// The server demonstrate recieving messages and replying the smbd process thruough			//
// a unix socket in a way that completes the unixntacl and fs VFS layers included			//
// 																							//
// The real ACL mechanism at Exanet is completely external to SAMBA and it resides inside	//
// the proprietry file system (which is why it also handles the FS-API calls). The actual	//
// work is distrubted (as in the external lock manager) through a proxy server				//
// 																							//
// Samba <--> ACLServer <--> Internal ACL Implementation									//
// 																							//
//////////////////////////////////////////////////////////////////////////////////////////////

/* parse inputs and start deamon */
int main(int argc, char **argv) {

	// Local configuration params
	int		baseDomain 	= -1;
	int		step 		= -1;
	int		noOfDomains = -1;
	char	serverPrefix[SERVER_PREFIX_SIZE+1] 	= "eam";
	bool	deamon 		= true;
	verbose	= false;
	
	// getopt vars
	int c;
    extern char *optarg;
    extern int optind, optopt;
	// parse input
	while ((c = getopt(argc, argv, "b:s:n:p:ivh")) != -1) {
        switch (c) {
        case 'i':
            deamon = false;
            break;
		case 'v':
			verbose = true;
			break;
		case 'h':
			usage(argv);
			break;
		case 'b':
			if (sscanf(optarg, "%d", &baseDomain) != 1) {
				cerr << "Base domain was in wrong format" << endl;
				exit(1);
			}
			if (baseDomain < 0) {
				cerr << "Base domain must be positive (or 0)" << endl;
				exit(1);
			}
			DBG(cout << "Base Domain set to " << baseDomain << endl);
			break;
		case 's':
			if (sscanf(optarg, "%d", &step) != 1) {
				cerr << "Step was in wrong format" << endl;
				exit(1);
			}
			if (step < 1) {
				cerr << "Step must be positive and non zero" << endl;
				exit(1);
			}
			DBG(cout << "Step set to " << step << endl);
			break;
		case 'n':
			if (sscanf(optarg, "%d", &noOfDomains) != 1) {
				cerr << "Number of Domains was in wrong format" << endl;
				exit(1);
			}
			if (noOfDomains < 0) {
				cerr << "Number of Domains must be positive and non zero" << endl;
				exit(1);
			}
			DBG(cout << "Number of Domains set to " << noOfDomains << endl);
			break;
		case 'p':
			if (strlen(optarg) > SERVER_PREFIX_SIZE) {
				cerr << "serverPrefix too long... exiting" << endl;
			}
			if (sscanf(optarg, "%s", serverPrefix) != 1) {
				cerr << "serverPrefix parsing error... exiting" << endl;
				exit(1);
			}
			DBG(cout << "Server Prefix used is " << serverPrefix << endl);
			break;
		case ':':
			cerr << "option -" << char(optopt) << "requires an operand" << endl;
			usage(argv);
			break;
		case '?':
			cerr << "Unrecognised option -" << char(optopt) << endl;
			usage(argv);
			break;
		default:
            fprintf(stderr,"Invalid option %c\n", c);
            usage(argv);
            break;
        }
	}

	// Check mandatory values
	if (baseDomain < 0) {
		cerr << "Base domain is mandatory" << endl;
		usage(argv);
	}

	if (step < 0) {
		cerr << "Step is mandatory" << endl;
		usage(argv);
	}

	if (noOfDomains < 0) {
		cerr << "Number of Domains is mandatory" << endl;
		usage(argv);
	}

	// demonize
    if (deamon) {
        pid_t pid = fork();
        if (pid == -1) {
            DBG(cerr << "Error forking deamon proccess... exiting" << endl);
            exit(1);
        }
        if (pid != 0) { // parent should exit
			DBG(cout << "Fork succesfull - parent exiting" << endl);
            exit(0);
        } else {
			DBG(cout << "Deamon created and running" << endl);
		}
    } else {
		DBG(cout << "Running in interactive mode" << endl);
	}

	// Create the listenerMgr
	ListenerMgr *listenerMgr = new ListenerMgr(baseDomain,
											   step,
											   noOfDomains,
											   serverPrefix);
	listenerMgr->open();
	listenerMgr->run(); // endless loop
};

