#include <iostream>
#include <string.h>

#include <errno.h>

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>

#include "SambaElmConnection.h"
#include <sys/shm.h>

	using namespace std;

	// ctor - new connection identified by fd
	SambaElmConnection::SambaElmConnection(int fd) 
	: BasicConnection(fd)
	, conn_type(CONN_TYPE_DIRECT)
	, shmid(-1)
	, rcved(0)
	, left(0)
	, ctrl_buf(NULL)
	, data_buf(NULL)
	{
		if (conn_type == CONN_TYPE_DIRECT) {
			ctrl_buf = new uint32_t[1024/sizeof(uint32_t)];
		} else {
			ctrl_buf = new uint32_t[1024*1024/sizeof(uint32_t)];
			data_buf = new uint32_t[1024*1024/sizeof(uint32_t)];
		}
	}

	SambaElmConnection::~SambaElmConnection() {
		if (conn_type != CONN_TYPE_DIRECT) {
			delete[] data_buf;
		} else if (shmid != -1) {
			if (data_buf) {
				if (shmdt(data_buf) < 0) {
					cerr << "shmdt failed with errno " << errno << endl;
				}
			}
			shmid = -1;
		}
		data_buf = NULL;
		delete[] ctrl_buf;
		ctrl_buf = NULL;
	}

	// handle incoming message,
	// includes reading, decoding, processing, encoding and replying.
	exasmb_message_error SambaElmConnection::handleMessage() {

		// read the incoming message in varialbe length format
		ssize_t n = read(_fd, &left, sizeof(left)); // read the length of the message, classic variable length
		if (n < 0) {
			cerr << "Error reading from fd " << _fd << " with error " << strerror(errno) << endl;
			rcved = left = 0;
			_buf = NULL;
			return EXASMB_COMM_ERROR;
		} else if (n == 0) {
			DBG(cout << "connection closed" << endl);
			rcved = left = 0;
			_buf = NULL;
			return EXASMB_FATAL_ERROR; // == connection closed
		} // else ...

		// left contains the number of the bytes in the message (without left itself)
		left = ntohl(left);
		for (rcved = 0; left; left -= n, rcved += n) {
			n = read(_fd, &ctrl_buf[rcved], left);
			if (n < 0) {
				return EXASMB_COMM_ERROR;
			}
		}

		// decode message
		_buf = dec_exasmb_message_header(&header, ctrl_buf);	

		// handle shmid stuff which is part of the messaging mechanism
		if (header.shmid != -1 && header.shmid != shmid) {		
			if (shmid != -1) {
				if (shmdt(data_buf) < 0) {
					cerr << "shmdt returned error " << errno << endl;
				}
				if (shmctl(shmid, IPC_RMID, NULL) < 0) {
					cerr << "shmctl returned error " << errno << endl;
				}
				shmid = -1;
			}
			data_buf = (uint32_t *) shmat(header.shmid, NULL, 0);
			if (data_buf != (uint32_t *)-1)
				shmid = header.shmid;
			else {
				shmid = -1;
				cerr << "shmat failed error " << errno << endl;
				//throw xnt_exception("shmat failed");
			}
		}else if((header.shmid == -1) &&(shmid != -1)){			  
			if (data_buf) {
				if (shmdt(data_buf) < 0) {
					cerr << "shmdt failed error " << errno << endl;
				}
			}
			shmid = -1;  
			data_buf = NULL;		
		}

		// decode non header data.
		// will put info in _decoded_message
		decode_message(_buf);
		// process message according to message_code
		switch (header.message_code) {
		case EXASMB_CONNECT_NLM: 
			{
				DBG(cout << "Header.message_code = EXASMB_CONNECT_NLM" << endl);
				exasmb_connect_request *request = static_cast<exasmb_connect_request*> (_decoded_message.get());
				reply_ok();

				break;
			}
		case EXASMB_CLEANUP:
			{
				DBG(cout << "Header.message_code = EXASMB_CLEANUP" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_LOCK_LR:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_LOCK_LR" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_UNLOCK_LR:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_UNLOCK_LR" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_GET_LOCKS:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_GET_LOCKS" << endl);
				exasmb_share_get_locks_reply reply;

				memset(&reply, 0, sizeof(reply));
				reply.header.cookie = 0;
				reply.header.num_entries = 0;
				reply.header.delete_on_close = 0;
				reply.body.entries = NULL;

				// no handling of entries in this dummy impelementation
				// should handle each entry

				reply_ok((xdr_func) enc_exasmb_share_get_locks_reply_header, &reply,
						 (xdr_func) enc_exasmb_share_get_locks_reply_body);

				break;
			}
		case EXASMB_SHARE_ADD_LOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_ADD_LOCK" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_DELETE_LOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_DELETE_LOCK" << endl);

				exasmb_share_delete_lock_reply reply;
				exasmb_share_delete_lock_request *request = static_cast<exasmb_share_delete_lock_request*> (_decoded_message.get());

				memset(&reply, 0, sizeof(reply));
				reply.delete_on_close = 0;
				reply.num_left = 0;
				reply.locks_were_deleted = 1;
				reply.lock_deleted = *(request->body.entries);
                
				reply_ok((xdr_func) &enc_exasmb_share_delete_lock_reply, &reply);	

				break;
			}
		case EXASMB_SHARE_MODIFY_OPLOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_MODIFY_OPLOCK" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_SET_DELETE_ON_CLOSE:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_SET_DELETE_ON_CLOSE" << endl);
				reply_ok();

				break;
			}
		case EXASMB_SHARE_GET_DELETE_ON_CLOSE:
			{
				DBG(cout << "Header.message_code = EXASMB_SHARE_GET_DELETE_ON_CLOSE" << endl);
				exasmb_share_get_delete_on_close_reply reply;

				memset(&reply, 0, sizeof(reply));
				reply.delete_on_close = 0;
				reply_ok((xdr_func) &enc_exasmb_share_get_delete_on_close_reply, &reply);

				break;
			}

		case EXASMB_BR_LOCK_LR:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_LOCK_LR" << endl);
				reply_ok();

				break;
			}
		case EXASMB_BR_UNLOCK_LR:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_UNLOCK_LR" << endl);
				reply_ok();

				break;
			}
		case EXASMB_BR_GET_LOCKS:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_GET_LOCKS" << endl);
				exasmb_br_get_locks_reply reply;
				
				memset(&reply, 0, sizeof(reply));
				reply.header.cookie = 0;
				reply.header.num_entries = 0;
				reply.body.entries = NULL;

				reply_ok((xdr_func) enc_exasmb_br_get_locks_reply_header, &reply,
						 (xdr_func) enc_exasmb_br_get_locks_reply_body);

				break;
			}
		case EXASMB_BR_CHK_LOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_CHK_LOCK" << endl);
				reply_ok();

				break;
			}
		case EXASMB_BR_ADD_LOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_ADD_LOCK" << endl);
				reply_ok();

				break;
			}
		case EXASMB_BR_DELETE_LOCK:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_DELETE_LOCK" << endl);
				reply_ok();

				break;
			}
		case EXASMB_BR_FILE_CLOSE:
			{
				DBG(cout << "Header.message_code = EXASMB_BR_FILE_CLOSE" << endl);
				reply_ok();

				break;
			}
		default:
			cerr << "Unknown message code: " << header.message_code << endl;
			rcved = left = 0;
			_buf = NULL;
			return EXASMB_FATAL_ERROR;
		}

		rcved = left = 0;
		_buf = NULL;
		return EXASMB_OK;
	}

	void SambaElmConnection::reply_ok(xdr_func xdr_ctl, 
									  void *p, 
									  xdr_func xdr_dat)
	{
		header.message_type = EXASMB_MESSAGE_REPLY;
		header.message_error = EXASMB_OK;
		
		uint32_t *cp = enc_exasmb_message_header(&header, ctrl_buf);
		if (xdr_ctl)
			cp = xdr_ctl(p, cp);
		if (xdr_dat)
			cp = xdr_dat(p, cp);
		send((char *) cp - (char *) ctrl_buf,
			 ctrl_buf);	
	}

	void
	SambaElmConnection::decode_message(uint32_t *p)
	{   	
		switch (header.message_code) {	
		case EXASMB_CONNECT_NLM:
			decode_connect(p);
			break;
		case EXASMB_CLEANUP:
			decode_cleanup(p);
			break;
		case EXASMB_SHARE_LOCK_LR:
			decode_share_lock_lr(p);
			break;
		case EXASMB_SHARE_UNLOCK_LR:
			decode_share_unlock_lr(p);
			break;
		case EXASMB_SHARE_GET_LOCKS:
			decode_share_get_locks(p);
			break;
		case EXASMB_SHARE_ADD_LOCK:
			decode_share_add_lock(p);
			break;
		case EXASMB_SHARE_DELETE_LOCK:
			decode_share_delete_lock(p);
			break;

		case EXASMB_SHARE_MODIFY_OPLOCK:
			decode_share_modify_oplock(p);
			break;
		case EXASMB_SHARE_SET_DELETE_ON_CLOSE:
			decode_share_set_delete_on_close(p);
		case EXASMB_SHARE_GET_DELETE_ON_CLOSE:
			decode_share_get_delete_on_close(p);
			break;
		case EXASMB_BR_LOCK_LR:
			decode_br_lock_lr(p);
			break;
		case EXASMB_BR_UNLOCK_LR:
			decode_br_unlock_lr(p);
			break;
		case EXASMB_BR_GET_LOCKS:
			decode_br_get_locks(p);
			break;
		case EXASMB_BR_CHK_LOCK:
			decode_br_chk_lock(p);
			break;
		case EXASMB_BR_ADD_LOCK:
			decode_br_add_lock(p);
			break;
		case EXASMB_BR_DELETE_LOCK:
			decode_br_delete_lock(p);
			break;
		case EXASMB_BR_FILE_CLOSE:
			decode_br_file_close(p);
			break;
		default:
			throw std::exception();
		}
	}

	uint32_t *
	SambaElmConnection::decode_request(uint32_t *buf,
									void *req,
									xdr_func xdr_ctl,
									xdr_func xdr_dat)
	{
		uint32_t *p = xdr_ctl(req, buf);
		if (xdr_dat) {
			if (conn_type != CONN_TYPE_DIRECT)
				p = xdr_dat(req, p);
			else {
				if (shmid == -1 || !data_buf){
					//throw std::exception("Bad Data Buffer");
					cerr << "CRITICAL: Bad Data Buffer" << endl;
				}
				struct typical_vec_repuest {
					exasmb_request_vec_header header;
					struct typical_body {
						void	*entries;
					} body;
				};
				((typical_vec_repuest *) req)->body.entries = data_buf;
			}
		}
		return p;
	}

	void
	SambaElmConnection::decode_connect(uint32_t *p)
	{
		exasmb_connect_request* request = static_cast<exasmb_connect_request*>
			(_decoded_message.reset( sizeof(exasmb_connect_request)));

		p = dec_exasmb_connect_request(request, p);	
	}

	void
	SambaElmConnection::decode_cleanup(uint32_t *p)
	{
		exasmb_cleanup_request* request = static_cast<exasmb_cleanup_request*>
			(_decoded_message.reset( sizeof(exasmb_cleanup_request)));
		p = dec_exasmb_cleanup_request(request, p);	
	}

	void
	SambaElmConnection::decode_share_lock_lr(uint32_t *p)
	{
		exasmb_share_lock_lr_request* request = static_cast<exasmb_share_lock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_share_lock_lr_request)));
		dec_exasmb_share_lock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_share_unlock_lr(uint32_t *p)
	{
		exasmb_share_unlock_lr_request* request = static_cast<exasmb_share_unlock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_share_unlock_lr_request)));
		dec_exasmb_share_unlock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_share_get_locks(uint32_t *p)
	{
		exasmb_share_get_locks_request* request = static_cast<exasmb_share_get_locks_request*>
		(_decoded_message.reset( sizeof(exasmb_share_get_locks_request)));
		dec_exasmb_share_get_locks_request(request, p);
	}

	void
	SambaElmConnection::decode_share_add_lock(uint32_t *p)
	{
		exasmb_share_add_lock_request* request = static_cast<exasmb_share_add_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_add_lock_request)));

		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_share_add_lock_request_header,
						(xdr_func) &dec_exasmb_share_add_lock_request_body);
	}

	void
	SambaElmConnection::decode_share_delete_lock(uint32_t *p)
	{
		exasmb_share_delete_lock_request* request = static_cast<exasmb_share_delete_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_delete_lock_request)));
		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_share_delete_lock_request_header,
						(xdr_func) &dec_exasmb_share_delete_lock_request_body);
	}

	void
	SambaElmConnection::decode_share_modify_oplock(uint32_t *p)
	{
		exasmb_share_modify_oplock_request* request = static_cast<exasmb_share_modify_oplock_request*>
		(_decoded_message.reset( sizeof(exasmb_share_modify_oplock_request)));
		dec_exasmb_share_modify_oplock_request(request, p);
	}

	void
	SambaElmConnection::decode_br_lock_lr(uint32_t *p)
	{
		exasmb_br_lock_lr_request* request = static_cast<exasmb_br_lock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_br_lock_lr_request)));
		dec_exasmb_br_lock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_br_unlock_lr(uint32_t *p)
	{
		exasmb_br_unlock_lr_request* request = static_cast<exasmb_br_unlock_lr_request*>
		(_decoded_message.reset( sizeof(exasmb_br_unlock_lr_request)));
		dec_exasmb_br_unlock_lr_request(request, p);
	}

	void
	SambaElmConnection::decode_br_get_locks(uint32_t *p)
	{
		exasmb_get_locks_request* request = static_cast<exasmb_get_locks_request*>
		(_decoded_message.reset( sizeof(exasmb_get_locks_request)));
		dec_exasmb_get_locks_request(request, p);	
	}

	void
	SambaElmConnection::decode_br_chk_lock(uint32_t *p)
	{
		exasmb_br_chk_lock_request* request = static_cast<exasmb_br_chk_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_chk_lock_request)));

		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_chk_lock_request_header,
						(xdr_func) &dec_exasmb_br_chk_lock_request_body);	
	}


	void
	SambaElmConnection::decode_share_get_delete_on_close(uint32_t *p)
	{
		exasmb_share_get_delete_on_close_request* request = static_cast<exasmb_share_get_delete_on_close_request*>
		(_decoded_message.reset( sizeof(exasmb_share_get_delete_on_close_request)));		
		dec_exasmb_share_get_delete_on_close_request(request, p);	
	}
	void

	SambaElmConnection::decode_share_set_delete_on_close(uint32_t *p)
	{
		exasmb_share_set_delete_on_close_request* request = static_cast<exasmb_share_set_delete_on_close_request*>
		(_decoded_message.reset( sizeof(exasmb_share_set_delete_on_close_request)));		
		dec_exasmb_share_set_delete_on_close_request(request, p);
	}

	void
	SambaElmConnection::decode_br_add_lock(uint32_t *p)
	{
		exasmb_br_add_lock_request* request = static_cast<exasmb_br_add_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_add_lock_request)));
		request->body.entries = NULL;

		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_add_lock_request_header,
						(xdr_func) &dec_exasmb_br_add_lock_request_body);	
	}

	void
	SambaElmConnection::decode_br_delete_lock(uint32_t *p)
	{
		exasmb_br_delete_lock_request* request = static_cast<exasmb_br_delete_lock_request*>
		(_decoded_message.reset( sizeof(exasmb_br_delete_lock_request)));

		request->body.entries = NULL;
		decode_request(p, request,
						(xdr_func) &dec_exasmb_br_delete_lock_request_header,
						(xdr_func) &dec_exasmb_br_delete_lock_request_body);	
	}

	void
	SambaElmConnection::decode_br_file_close(uint32_t *p)
	{
		exasmb_file_close_request* request = static_cast<exasmb_file_close_request*>
		(_decoded_message.reset( sizeof(exasmb_file_close_request)));	
		dec_exasmb_file_close_request(request, p);
	}

	SambaElmConnection::auto_malloc_ptr::auto_malloc_ptr():
	_ptr(NULL)
	{

	}

	SambaElmConnection::auto_malloc_ptr::~auto_malloc_ptr()
	{
		free(_ptr);
	}

	void* 
	SambaElmConnection::auto_malloc_ptr::reset(size_t size)
	{
		free(_ptr);
		_ptr = malloc(size);
		return _ptr;
	}

	void* 
	SambaElmConnection::auto_malloc_ptr::get()
	{
		return _ptr;
	}

