#ifndef _SAMBA_ELM_CONNECTION
#define _SAMBA_ELM_CONNECTION

#include "BasicConnection.h"
#include "exasmb_xdr.h"

//////////////////////////////////////////////////
// Samba Elm Connection							//
// -------------------------------				//
// Responsible for handling new messages for 	//
// existing connections.						//
//////////////////////////////////////////////////

class SambaElmConnection : public BasicConnection {
public:
	SambaElmConnection(int fd);
	virtual ~SambaElmConnection();

	virtual exasmb_message_error handleMessage();
	void reply_ok(xdr_func xdr_ctl= NULL, 
				  void *p = NULL,
				  xdr_func xdr_dat = NULL);

private:
	// internall class
	class auto_malloc_ptr
	{
		public:
			auto_malloc_ptr();
			~auto_malloc_ptr();
			void* reset(size_t size);
			void* get();		
		private:
			void* _ptr;
	};

	// data members
	int conn_type;
	int shmid;
	exasmb_message_header header;	// current message (req,rep) header

	// reading aids
	ssize_t	rcved;
	ssize_t	left;
	// Buffers and pointers
	uint32_t *ctrl_buf;
	uint32_t *data_buf;
	uint32_t *_buf;

	auto_malloc_ptr _decoded_message;	

	// private functions
	void decode_message(uint32_t *p);

	uint32_t * decode_request(uint32_t *buf,
							  void *req,
							  xdr_func xdr_ctl,
							  xdr_func xdr_dat);

	//	samba verbs
	void	decode_connect(uint32_t *p);
	void	decode_cleanup(uint32_t *p);
	void	decode_share_lock_lr(uint32_t *p);
	void	decode_share_unlock_lr(uint32_t *p);
	void	decode_share_get_locks(uint32_t *p);
	void	decode_share_add_lock(uint32_t *p);
	void	decode_share_delete_lock(uint32_t *p);
	void	decode_share_change_mode(uint32_t *p);
	void	decode_share_modify_oplock(uint32_t *p);
	void	decode_share_set_delete_on_close(uint32_t *p);
	void	decode_share_get_delete_on_close(uint32_t *p);
	void	decode_br_lock_lr(uint32_t *p);
	void	decode_br_unlock_lr(uint32_t *p);
	void	decode_br_get_locks(uint32_t *p);
	void	decode_br_chk_lock(uint32_t *p);
	void	decode_br_add_lock(uint32_t *p);
	void	decode_br_delete_lock(uint32_t *p);
	void	decode_br_file_close(uint32_t *p);

};

#endif /* _SAMBA_ELM_CONNECTION */

