/*
   This file is intended for internal use by messages_inter_{client,server}.c only.
 */

#ifndef _MESSAGES_INTER_H_
#define _MESSAGES_INTER_H_

/*
   All the tunables (even those related only to client or only to server are located in this file.
 */

#define UDP_SERV_NAME "samba-message"
#define UDP_BUF_SIZE 8192
#define MESSAGE_HIST_SIZE 1000
#define UDP_RETRY_INTERVAL 10
#define UDP_RETRY_INTERVAL_LIMIT 1000
#define MAX_NODES 256
#define SERVER_STARTED_THRESHOLD 30
#define SERVER_START_LIMIT 7
#define SERVER_START_DELAY_LIMIT 15
/* SERVER_START_DELAY_LIMIT should probably be less then SERVER_STARTED_THRESHOLD
   otherwise we'll always consider servers that have slept SERVER_START_DELAY_LIMIT as
   having been successfully started.
   with current values, server restart delay sequence is 0 1 2 4 8 15 15
*/
/* this is to be used for debugging only */
/* #define UDP_DROP_PROB 50 */

/*
Clients communicate with proxy servers using UDP protocol.
Client request consists of a header udp_request_header followed by structure that corresponds
to message type, which in turn may be followed by some other message-specific payload.

Currently, only one message type is defined: UDP_MSG_SEND_MESSAGE, its corresponding structure
is named send_message_request. This structure is followed by the message data supplied
in "buf" parameter of message_send_pid function.

The proxy server replies with udp_reply_header, optionally followed by some message type specific
payload, whose length is supplied in "extra" field.
Currently, no message types use extra field.
*/

struct udp_request_header {
	int type; /* UDP_MSG_xxx*/
	unsigned int seq;
	pid_t source_pid;
};

struct udp_reply_header {
	unsigned int seq;
	NTSTATUS result;
	unsigned int extra;
};

#define UDP_MSG_SEND_MESSAGE 1

struct send_message_request {
	struct process_id src_pid;
	struct process_id dst_pid;
	int msg_type;
	BOOL duplicates_allowed;
	unsigned int timeout;
	size_t len;
};

#define ADDR_IPPORT(addr) inet_ntoa((addr)->sin_addr), ntohs((addr)->sin_port)
#define MESSAGE_SERVER_UDP_CTX_IPPORT(ctx) ADDR_IPPORT(&(ctx)->addr)

#endif
