#ifndef _EXASMB_TYPES_H_
#define _EXASMB_TYPES_H_

#include <stdint.h>

#ifdef __cplusplus
#include <iostream>
#endif

#define EXASMB_COOKIE_INIT			((uint64_t)0)
#define EXASMB_COOKIE_ENDDATA       (~(uint64_t)0)
#define MAX_FETCH_NUM_ENTRIES       256

#define EXASMB_CLEANER_NEEDS_CLEANING ((pid_t)0)
#define EXASMB_CLEANER_IS_CLEANING ((pid_t)1)

// 32nd bit for now
#define VERSION_CLEAR_UNCONDITIONAL (1ULL << 31)

enum exasmb_message_type {
	EXASMB_MESSAGE_REQUEST = 0,
	EXASMB_MESSAGE_REPLY = 1,
};
typedef enum exasmb_message_type exasmb_message_type;


enum exasmb_message {
	EXASMB_CONNECT_NLM = 1,
	EXASMB_CLEANUP,

	EXASMB_SHARE_LOCK_LR,
	EXASMB_SHARE_UNLOCK_LR,
	EXASMB_SHARE_GET_LOCKS,
	EXASMB_SHARE_ADD_LOCK,
	EXASMB_SHARE_DELETE_LOCK,
    EXASMB_SHARE_MODIFY_OPLOCK,
	EXASMB_SHARE_SET_DELETE_ON_CLOSE,
	EXASMB_SHARE_GET_DELETE_ON_CLOSE,

	EXASMB_BR_LOCK_LR,
	EXASMB_BR_UNLOCK_LR,
	EXASMB_BR_GET_LOCKS,
	EXASMB_BR_CHK_LOCK,
	EXASMB_BR_ADD_LOCK,
	EXASMB_BR_DELETE_LOCK,
	EXASMB_BR_FILE_CLOSE,
};
typedef enum exasmb_message exasmb_message;

#ifdef __cplusplus
std::ostream&  operator<<(std::ostream& out, exasmb_message code);
#endif

enum exasmb_message_error {
	/*O.K*/
	EXASMB_OK = 0,
	/* general error - bail out 
	   currently this only marks that the server state is not o.k 
	*/
	EXASMB_ERROR = -1,
	/* 
		transient error, 
		something is not working at the moment should be o.k soon 		
		you may only abort the process if this operation is not a cleanup operation
	*/
	EXASMB_COMM_ERROR = -2,
	/* 
	  smb-dlm protocol error, the state isnot consistent 
	  samba thinks one thing, the dlm thinks differntly, the dlm is the server  
	  so samba should bail out,
	  with a "good" print out to the debug log, as this is likely to be a software
	  bug
	*/
	EXASMB_LRLOCK_ERROR = -3,

	/*  		
		NOTICE there is an incorrect handling of this error in samba code 
		this is a reply to a chk lock query that finds there is no lock
	*/
	EXASMB_NOLOCK_ERROR = -4,

	/* 
	   not used in the protocol no more
	   a cleanup process that fails, just close the connection
	*/
	EXASMB_CLEANUP_ERROR = -5,

	/* 
	   semi transient error, the system is out of memory 
	   decide what to do
	*/ 
	EXASMB_OUT_OF_MEMORY_ERROR = -6,
	/* semi transient error, 
	   the object trying to add a new entry to , reached it's 
	   maximum capcitity and no new data can be added to this object
	   decide what to do
	*/
	EXASMB_OUT_OF_OBJECT_SPACE_ERROR = -7,
	/*
	   something bad in the lower level of the protocol itself
	   missing shared memory,inconsisitent requests and the like of that 
	   fix the state, or bail out :)
	*/
	EXASMB_FATAL_ERROR = -100,
};
typedef enum exasmb_message_error exasmb_message_error;


struct exasmb_message_header {
	uint32_t domain;
	exasmb_message_type message_type;
	exasmb_message message_code;
	exasmb_message_error message_error;
	int shmid;
};
typedef struct exasmb_message_header exasmb_message_header;


struct exasmb_conn_context {
	uint64_t vers;
	uint32_t inst_id;
	const char *inst_name;
};
typedef struct exasmb_conn_context exasmb_conn_context;


/* The lock resource context */
struct exasmb_lr_context {
	uint32_t pid;
	uint16_t tid;
	uint64_t dev;
	uint64_t inode;
    uint32_t file_id;
    uint64_t sequence_id;
	struct {
		uint32_t	len;
		char		buf[64];
	} data;
};
typedef struct exasmb_lr_context exasmb_lr_context;


/* A connection request to ExaStore */
struct exasmb_connect_request {
	exasmb_conn_context conn;
};
typedef struct exasmb_connect_request exasmb_connect_request;

/* Cleanup request for spesific host (pid) */
struct exasmb_cleanup_request {
	uint32_t pid;
};
typedef struct exasmb_cleanup_request exasmb_cleanup_request;

/* A simple request on a LR */
struct exasmb_simple_request {
	exasmb_lr_context lr;
	uint32_t lock_count;
};
typedef struct exasmb_simple_request exasmb_simple_request;


/*  Get entries from LR */
struct exasmb_get_locks_request {
	exasmb_lr_context lr;
	uint64_t cookie;
	uint32_t num_locks;
};
typedef struct exasmb_get_locks_request exasmb_get_locks_request;


struct exasmb_reply_vec_header
{
	uint64_t cookie;
	uint32_t num_entries;
	uint32_t delete_on_close;
};

typedef struct exasmb_reply_vec_header exasmb_reply_vec_header;


struct exasmb_request_vec_header
{
    exasmb_lr_context lr;
	uint32_t num_entries;
};
typedef struct exasmb_request_vec_header exasmb_request_vec_header;


/******************************************************************************/
/*          					Shared locks                                  */
/******************************************************************************/

typedef exasmb_simple_request exasmb_share_lock_lr_request;

typedef exasmb_simple_request exasmb_share_unlock_lr_request;

typedef exasmb_get_locks_request exasmb_share_get_locks_request;


struct exasmb_time_val {
    uint32_t tv_sec;
    uint32_t tv_usec;
};
typedef struct exasmb_time_val exasmb_time_val;


struct exasmb_shared_lock {
	uint32_t pid;
	uint16_t node;
	uint32_t smbid;
	uint16_t op_mid;
	uint16_t op_type;
	uint32_t  share_access;
	uint32_t access_mask;
	exasmb_time_val open_time;
	uint64_t dev;
	uint64_t inode;
	uint32_t share_file_id;
	uint32_t private_options;
	uint32_t uid;
	uint16_t flags;
};
typedef struct exasmb_shared_lock exasmb_shared_lock;


struct exasmb_share_get_locks_reply_body {
    exasmb_shared_lock *entries;
};
typedef struct exasmb_share_get_locks_reply_body exasmb_share_get_locks_reply_body;

struct exasmb_share_get_locks_reply {
	exasmb_reply_vec_header header;
	exasmb_share_get_locks_reply_body body;
};
typedef struct exasmb_share_get_locks_reply exasmb_share_get_locks_reply;


struct exasmb_share_add_lock_request_body {
    exasmb_shared_lock *entries;
};
typedef struct exasmb_share_add_lock_request_body exasmb_share_add_lock_request_body;

struct exasmb_share_add_lock_request {
	exasmb_request_vec_header header;
	exasmb_share_add_lock_request_body body;
};
typedef struct exasmb_share_add_lock_request exasmb_share_add_lock_request;


struct exasmb_share_delete_lock_request_body {
    exasmb_shared_lock *entries;
};
typedef struct exasmb_share_delete_lock_request_body exasmb_share_delete_lock_request_body;

struct exasmb_share_delete_lock_request {
	exasmb_request_vec_header header;
	exasmb_share_delete_lock_request_body body;
};
typedef struct exasmb_share_delete_lock_request exasmb_share_delete_lock_request;


struct exasmb_share_delete_lock_reply {
		uint32_t num_left;
        uint16_t locks_were_deleted;

        /* A copy of the last lock that was deleted */
	    exasmb_shared_lock lock_deleted;
		uint32_t delete_on_close;
};
typedef struct exasmb_share_delete_lock_reply exasmb_share_delete_lock_reply;


struct exasmb_share_delete_by_pid
{
    uint32_t pid;
    uint32_t smbid;
};
typedef struct exasmb_share_delete_by_pid exasmb_share_delete_by_pid;


struct exasmb_share_delete_lock_by_pid_request_body {
    exasmb_share_delete_by_pid *entries;
};
typedef struct exasmb_share_delete_lock_by_pid_request_body exasmb_share_delete_lock_by_pid_request_body;

struct exasmb_share_delete_lock_by_pid_request {
	exasmb_request_vec_header header;
	exasmb_share_delete_lock_by_pid_request_body body;
};
typedef struct exasmb_share_delete_lock_by_pid_request exasmb_share_delete_lock_by_pid_request;

struct exasmb_share_set_delete_on_close_request {
	exasmb_lr_context lr;
	uint32_t delete_on_close;
};
typedef struct exasmb_share_set_delete_on_close_request exasmb_share_set_delete_on_close_request;


struct exasmb_share_modify_oplock_request {
	exasmb_lr_context lr;
	uint32_t set_port;
	exasmb_shared_lock entry;
};
typedef struct exasmb_share_modify_oplock_request exasmb_share_modify_oplock_request;


struct exasmb_share_get_delete_on_close_request {
	exasmb_lr_context lr;
};
typedef struct exasmb_share_get_delete_on_close_request exasmb_share_get_delete_on_close_request;

struct exasmb_share_get_delete_on_close_reply {
	uint32_t delete_on_close;
};
typedef struct exasmb_share_get_delete_on_close_reply exasmb_share_get_delete_on_close_reply;


/******************************************************************************/
/*                              Byte Range locks                              */
/******************************************************************************/

typedef exasmb_simple_request exasmb_br_lock_lr_request;

typedef exasmb_simple_request exasmb_br_unlock_lr_request;

typedef exasmb_get_locks_request exasmb_br_get_locks_request;

struct exasmb_br_lock {
	uint16_t smbpid;
    uint16_t tid;
    uint32_t pid;
	uint32_t node;
    uint32_t smbid;
	uint32_t fnum;
	uint64_t start;
	uint64_t size;
	uint32_t lock_type;
};
typedef struct exasmb_br_lock exasmb_br_lock;


struct exasmb_br_get_locks_reply_body {
    exasmb_br_lock *entries;
};
typedef struct exasmb_br_get_locks_reply_body exasmb_br_get_locks_reply_body;

struct exasmb_br_get_locks_reply {
	exasmb_reply_vec_header header;
	exasmb_br_get_locks_reply_body body;
};
typedef struct exasmb_br_get_locks_reply exasmb_br_get_locks_reply;


struct exasmb_br_chk_lock_request_body {
    exasmb_br_lock *entries;
};
typedef struct exasmb_br_chk_lock_request_body exasmb_br_chk_lock_request_body;

struct exasmb_br_chk_lock_request {
	exasmb_request_vec_header header;
	exasmb_br_chk_lock_request_body body;
};
typedef struct exasmb_br_chk_lock_request exasmb_br_chk_lock_request;

struct exasmb_br_add_lock_request_body {
    exasmb_br_lock *entries;
};
typedef struct exasmb_br_add_lock_request_body exasmb_br_add_lock_request_body;

struct exasmb_br_add_lock_request {
	exasmb_request_vec_header header;
	exasmb_br_add_lock_request_body body;
};
typedef struct exasmb_br_add_lock_request exasmb_br_add_lock_request;


struct exasmb_br_delete_lock_request_body {
    exasmb_br_lock *entries;
};
typedef struct exasmb_br_delete_lock_request_body exasmb_br_delete_lock_request_body;

struct exasmb_br_delete_lock_request {
	exasmb_request_vec_header header;
	exasmb_br_delete_lock_request_body body;
};
typedef struct exasmb_br_delete_lock_request exasmb_br_delete_lock_request;


struct exasmb_file_close_request {
	exasmb_lr_context lr;
	uint32_t fnum;
    uint32_t share_file_id;
};
typedef struct exasmb_file_close_request exasmb_file_close_request;


/******************************************************************************/
/*                              Block locks                                   */
/******************************************************************************/

typedef exasmb_simple_request exasmb_blk_lock_lr_request;

typedef exasmb_simple_request exasmb_blk_unlock_lr_request;

typedef exasmb_get_locks_request exasmb_blk_get_locks_request;

struct exasmb_blk_lock {
    uint32_t  com_type;
    uint32_t expire_time;
    uint32_t  lock_num;
	uint32_t mid;
    uint32_t fnum;
    
    /* inbuf */
    uint32_t length;
    char     *inbuf;
};
typedef struct exasmb_blk_lock exasmb_blk_lock;


struct exasmb_blk_get_locks_reply_body {
    exasmb_blk_lock *entries;
};
typedef struct exasmb_blk_get_locks_reply_body exasmb_blk_get_locks_reply_body;

struct exasmb_blk_get_locks_reply {
	exasmb_reply_vec_header header;
	exasmb_blk_get_locks_reply_body body;
};
typedef struct exasmb_blk_get_locks_reply exasmb_blk_get_locks_reply;


struct exasmb_blk_add_lock_request_body
{
    exasmb_blk_lock *entries;
};
typedef struct exasmb_blk_add_lock_request_body exasmb_blk_add_lock_request_body;

struct exasmb_blk_add_lock_request
{
	exasmb_request_vec_header header;
	exasmb_blk_add_lock_request_body body;
};
typedef struct exasmb_blk_add_lock_request exasmb_blk_add_lock_request;


struct exasmb_blk_delete_lock_by_fid_request
{	
	exasmb_lr_context lr;
	uint32_t fnum;
};
typedef struct exasmb_blk_delete_lock_by_fid_request exasmb_blk_delete_lock_by_fid_request;


struct exasmb_blk_delete_lock_by_mid_request
{
    exasmb_lr_context lr;
    uint32_t mid;
};
typedef struct exasmb_blk_delete_lock_by_mid_request exasmb_blk_delete_lock_by_mid_request;


struct exasmb_blk_check_timeout_request {
    exasmb_lr_context lr;
    uint32_t expire_time;
};
typedef struct exasmb_blk_check_timeout_request exasmb_blk_check_timeout_request;


struct exasmb_blk_check_timeout_reply {
    uint16_t queue_empty;
    uint32_t time;
};
typedef struct exasmb_blk_check_timeout_reply exasmb_blk_check_timeout_reply;

struct exasmb_blk_delete_and_next_request {
    exasmb_lr_context lr;
    uint64_t cookie;
};
typedef struct exasmb_blk_delete_and_next_request exasmb_blk_delete_and_next_request;


struct exasmb_blk_delete_and_next_reply {
    uint64_t cookie;
};
typedef struct exasmb_blk_delete_and_next_reply exasmb_blk_delete_and_next_reply;


#endif // #ifndef _EXASMB_TYPES_H_
