/* 
 * CAP VFS module for Samba 3.x Version 0.3
 *
 * Copyright (C) Tim Potter, 1999-2000
 * Copyright (C) Alexander Bokovoy, 2002-2003
 * Copyright (C) Stefan (metze) Metzmacher, 2003
 * Copyright (C) TAKAHASHI Motonobu (monyo), 2003
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.                                     
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * modified for alternate data stream
 * Copyright (C) Didier Gautheron 2004
 * 
 * this module should compile with old 3.0 API and 2004-07 svn API
 */


#include "includes.h"

#undef DBGC_CLASS
#define DBGC_CLASS DBGC_VFS

#define ADS_FOLDER	".AppleDouble"
#define ADOUBLEMODE	0777

/* FIXME found a better test */
#ifdef 	SMB_VFS_OP
#define ADS_NEW_MODULE

/* for current svn tree */
#define ADS_TALLOC_INIT(a) talloc_init(a)


/* ------------------- */
#else

#define ADS_TALLOC_INIT(a) talloc_init()
#endif


static int getMode(mode_t mode)
{
    static mode_t local_mode;
    if (mode)
    {
        local_mode = mode;
        return 0;
    }
    else
        return local_mode;
}

/* ------------------------- 
 * format
 * .AppleDouble/filename/stream name
 * 
 * return the *LAST* '/' in path
 */
static int ads_get_path_ptr(char *path)
{
	int i   = 0;
	int ptr = 0;
	
	for (i = 0; path[i]; i ++) { 

		if (path[i] == '/')
			ptr = i;
	}
	
	return ptr;
}

/* ------------------------------
 * return the *FIRST* ':' in path
*/
static int ads_get_stream_ptr(const char *path)
{
	int i   = 0;
	int ptr = 0;
	
	for (i = 0; path[i]; i ++) {
		if (path[i] == ':') {
            if (path[i+1] && path[i+2])
            {
                if (path[i+1] == '2' && (path[i+2] == 'f' || path[i+2] == 'e'))
                    continue;
            }
			ptr = i;
			break;
		}
	}
	return ptr;
}

/* ---------------- 
 * fname is only a filename
*/

static char *ads_canonical_dir(TALLOC_CTX *ctx, const char *path, const char *fname, int isdir)
{
    if (isdir) {
        if(path)
            return talloc_asprintf(ctx, "%s/%s/%s/.Parent", path, fname, ADS_FOLDER);
        else
            return talloc_asprintf(ctx, "%s/%s/.Parent", fname, ADS_FOLDER);
    }
    if(path)
        return talloc_asprintf(ctx, "%s/%s/%s", path, ADS_FOLDER, fname);
    else
        return talloc_asprintf(ctx, "%s/%s", ADS_FOLDER, fname);

}

/* ---------------- 
 * return directory pathname for an alternate data stream
 * fname is *NOT* an altername name (ie foo:bar)
*/
static char *ads_dir(TALLOC_CTX *ctx, const char *path, const char *fname, int isdir)
{
    int ptr0 = 0;
    int ptr1 = 0;
    char *temp;
    DEBUG(3, ("path: %s, fname: %s\n", path, fname));

#if 0
    if (fname[0] == '.') ptr0 ++;
    if (fname[1] == '/') ptr0 ++;
#endif    
//    temp = talloc_asprintf(ctx, "%s/%s", path, &fname[ptr0]);
    temp = talloc_asprintf(ctx, "%s", &fname[ptr0]);
    ptr1 = ads_get_path_ptr(temp);
    if (ptr1 == 0 ) {
        return ads_canonical_dir(ctx, NULL, &temp[ptr1], isdir);
    }
    else{
        temp[ptr1] = '\0';
        return ads_canonical_dir(ctx, temp, &temp[ptr1 + 1], isdir);
    }
}

/* Shlomi: we need a way to support OS X SMB-type resource forks
 * The format is:
 *      data fork:  FILE1.txt
 *      resource fork: ._FILE1.txt
 * What we need to do is to convert the ._FILE1.txt to something like
 * ._FILE1.txt --> FILE1.txt:AFP_AfpInfo
 * This way the resource forks will also be available to MAC clients
 * who work through normal AFP mounts.
 */


/* ----------------------------------
 * build the pathname for stream, create folder if (mode & O_CREAT)
 * return -1 on error
 * 0 it's not a stream
 * 1 it's a stream
 *
 * main_path : file fullpathname with :$DATA removed
 * ads_path: unix pathname 
 * if it's not an ADS then main_path == ads_path
 *
 */

/*
 the following is used to let dlm related code in samba to open ads without resorting the vfs chain,
 as it causes troubles (e.g. undesired execution of antivirus check)
 */
extern int (*ads_build_paths_p)(TALLOC_CTX *ctx, const char *path, const char *fname,
                           char **ads_path, char **main_path, SMB_STRUCT_STAT **main_info, int flags);

static int ads_build_paths(TALLOC_CTX *ctx, const char *path, const char *fname,
                           char **ads_path, char **main_path, SMB_STRUCT_STAT **main_info, int flags)
{
    int ret = 0;
    int ptr0 = 0;
    int ptr1 = 0;
    int ptr2 = 0;
    int ptr3 = 0;
#if 0
    int smb_ptr = 0;
    char* smb_ads = NULL;
    pstring mac_smb_path;
#endif
    int mainpath_exists = -1;
    char *dname = 0;
    char *name  = 0;
    SMB_STRUCT_STAT ads_info;
    SMB_STRUCT_STAT local_stat;
    SMB_STRUCT_STAT* local_main_info;
    int use_stat = main_info ? 1 : 0;
    int main_path_stat_result = -1;

    if (!ctx || !path || !fname || !ads_path || !main_path)
        return -1;
#if 1
    DEBUG(3, ("PATH: %s[%s]\n", path, fname));
#endif
    if (strstr(path, ADS_FOLDER) || strstr(fname, ADS_FOLDER)) {
        DEBUG(3, ("ADS: path %s[%s] already contains %s\n", path, fname, ADS_FOLDER));
        return -1;
    }

#if 0
    if (fname[0] == '.') ptr0 ++;
    if (fname[1] == '/') ptr0 ++;
#endif

#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_mac_smb feature considered obsolete (now always False) */

    if (lp_allow_mac_smb() && (smb_ads=strstr (fname, "._")) != NULL) {
        pstring temp_smb_path;

        smb_ptr = smb_ads - fname;
        *smb_ads = '\0';
        pstrcpy (temp_smb_path,fname);
        *smb_ads = '.';

        pstrcat (temp_smb_path, fname+smb_ptr+2);
        pstrcat (temp_smb_path, ":AFP_AfpInfo");
        *main_path = talloc_asprintf(ctx, "%s", temp_smb_path);

        
        /*if (fname[0] == '.' && fname[1] == '/')
            smb_ptr += 2;*/
        /*
        if (smb_ads == fname) {
//        if (fname[smb_ptr] == '.' && fname[smb_ptr+1] == '_') { 
            // This is a MAC SMB type resource fork ! Now we will build the correct path:
              // FILE:AFP_AfpInfo 
            pstrcpy (mac_smb_path,fname+smb_ptr+2); 
            pstrcat (mac_smb_path, ":AFP_AfpInfo");
            *main_path = talloc_asprintf(ctx, "%s", mac_smb_path);
        }*/
    } else {
#endif
        *main_path = talloc_asprintf(ctx, "%s", &fname[ptr0]);
#if 0
    }
#endif
    *ads_path = NULL;
    
    /* get pointer to last '/' */
    ptr1 = ads_get_path_ptr(*main_path);
    ptr2 = ads_get_stream_ptr(*main_path +ptr1 +1);
    /* FIXME
     * what about ::$DATA or :name:$DATA
     * Shlomi:  :name is handled in the else - the operation fails. *
    */

	/* ptr2 == 0 means no : so no ads, but in the case of .:ads (ads on the 
	 * local direcotry in root of share) 
	 * ptr2 will be 0 because it's the first position where a : was found.*/
	if (ptr2 || (!ptr2 && (*main_path)[0] && (*main_path)[1] == ':')) {
        /* it's an alternate stream */
        ptr2 += ptr1 +1;
        (*main_path)[ptr2] = 0;
        ptr3 = ads_get_stream_ptr(*main_path +ptr2 +1);
        if (ptr3) {
            ptr3 += ptr2 +1;
            /* check it's $DATA */
            if (!StrCaseCmp("$DATA", &(*main_path)[ptr3+1])) {
                (*main_path)[ptr3] = 0;
            }
        } else if (!StrCaseCmp(":$DATA", &(*main_path)[ptr2+1])) {
			/* Shlomi: fix bug #20415 - in case the file looks like FILE::$DATA it is equivalent to FILE. However fname
			   still contains the ::$DATA string so we need to use the main_path string which is without it. */
			DEBUG(3, ("ADS: Found very special case of type FILE::$DATA for: %s\n", fname));
			*ads_path = *main_path;
			return ret;
		}

        DEBUG(3, ("ADS: MAIN DATA %s\n", *main_path));


        /* in case of ads_stat or others that don't care about the stat result, use local struct */
        if (!main_info || !*main_info)
            local_main_info = &local_stat;
        else
            local_main_info = *main_info;

	    if ((main_path_stat_result = sys_lstat(*main_path, local_main_info)) < 0) {
            /* Shlomi : it seems that Windows 2000 server allows creation of 
               ADS files even without their "main file", so we should allow this too.
            */ 
            mainpath_exists = 0;
            /* return -1; */
        } else
            mainpath_exists = 1;
        (*main_path)[ptr2] = ':';
        dname = talloc_strdup(ctx, *main_path);
        if (ptr1)
            dname[ptr1] = '\0';

        name = *main_path;
        name[ptr2] = '\0';

        if (!main_path_stat_result && S_ISDIR(local_main_info->st_mode)) {
            if (ptr1) 
                *ads_path = talloc_asprintf(ctx, "%s/%s/%s/.Parent/%s", dname, &name[ptr1 + 1], ADS_FOLDER, &name[ptr2 + 1]);
            else
                *ads_path = talloc_asprintf(ctx, "%s/%s/.Parent/%s", name, ADS_FOLDER, &name[ptr2 + 1]);
            
        } else {
            if (ptr1)
                *ads_path = talloc_asprintf(ctx, "%s/%s/%s/%s", dname, ADS_FOLDER, &name[ptr1 + 1], &name[ptr2 + 1]);
            else
                *ads_path = talloc_asprintf(ctx, "%s/%s/%s",ADS_FOLDER, name, &name[ptr2 + 1]);
        }

        /* XXX are we always the right user ?*/
        DEBUG(3,("ads_path is %s\n", *ads_path? *ads_path:"NULL"));

        /* Shlomi: small optimization: if already we realized that the main_path part of the file does not
         * exist, there's no point attempting anotehr stat on the ads part, so check main_path_stat_result */
        if (main_path_stat_result || sys_lstat(*ads_path, &ads_info) < 0) {
            int st_ret;
            if (errno == ENOENT && (flags & O_CREAT)) {
                char *ads_base;
                if (ptr1)
                    ads_base = talloc_asprintf(ctx, "%s",ads_canonical_dir(ctx, dname, &name[ptr1 + 1], S_ISDIR(local_main_info->st_mode)));
                else
                    ads_base = talloc_asprintf(ctx, "%s",ads_canonical_dir(ctx, NULL, &name[ptr1], S_ISDIR(local_main_info->st_mode)));
    
                DEBUG(3,("ads_base is %s\n", ads_base));
                mode_t mode;
                if (main_path_stat_result)
                    mode = getMode(0);
                else
                    mode = local_main_info->st_mode;
                DEBUG(3,("main_path_stat_result is %d, st_mode is %d and mode is %d\n", 
                         main_path_stat_result, local_main_info->st_mode, mode));
                
                /* if main_path does not exist, is there a point in attempting to create ads_base ?
                 * yes, because it is possible that .AppleDouble already exists, so the mkdir will succeed */
                st_ret = mkdir(ads_base, 0777);

                if (st_ret < 0) {
                    if (errno == ENOENT) {
                        char *ads_double;
                        if (S_ISDIR(local_main_info->st_mode)) {
                            if (ptr1) 
                                ads_double = talloc_asprintf(ctx, "%s/%s/%s", dname, &name[ptr1 + 1], ADS_FOLDER);
                            else
                                ads_double = talloc_asprintf(ctx, "%s/%s", name, ADS_FOLDER);
                            
                        } else {
                            if (ptr1)
                                ads_double = talloc_asprintf(ctx, "%s/%s", dname, ADS_FOLDER);
                            else
                                ads_double = talloc_asprintf(ctx, "%s", ADS_FOLDER);
                        }
                        if (mkdir(ads_double, 0777) < 0)
                            return -1;
                        if ((st_ret = mkdir(ads_base, 0777)) < 0)
                            return -1;

                        /* we just created .AppleDouble/file/ update mode with dir search 
                             * XXX what about acl?
                            */
                        if ((mode & (S_IRUSR | S_IWUSR )))
                            mode |= S_IXUSR;
                        if ((mode & (S_IRGRP | S_IWGRP )))
                            mode |= S_IXGRP;
                        if ((mode & (S_IROTH | S_IWOTH )))
                            mode |= S_IXOTH;
                        chmod(ads_base, mode);
                        /* here we already succeeded in creating all the ADS stuff, now
                           we need to make sure the main_path exists too, so that the 
                           ads will be visible to clients.
                        */   


                    } else
                        errno = ENOENT;
                }
                if (mainpath_exists == 0) {
                    DEBUG(1, ("ADS: creating base file %s with mode %d\n", *main_path, getMode(0)));
                    if ((ret = open(*main_path, flags, getMode(0)) != -1))
                        close(ret);
                }
            } else
                return -1;
        }
        ret = 1;
    } else {
        *ads_path = *main_path;
        if (use_stat && sys_lstat(*main_path, *main_info) < 0) {
            *main_info = NULL;
        }
    }
#if 1
    DEBUG(3, ("ADS: DEBUG:[%s] [%s]\n", *main_path, *ads_path)); 
#endif
    return ret;
}


static int ads_ntimes(vfs_handle_struct *handle, const char *path, const struct timespec ts[2])
{
    char *ads_path = NULL;
    char *main_path = NULL;
    int ads_ret = -1;
    int ret = 0;
    TALLOC_CTX *ctx;

    DEBUG(3,("ads: ads_ntimes %s\n", path));
    /* if stream error ?, change only the stream */
    if (!(ctx = ADS_TALLOC_INIT("ads_ntimes")))
        return -1;

    if ((ads_ret = ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, NULL, 0)) >= 0) {
        DEBUG(3,("ads: ads_ntimes main_path %s ads_path %s\n", main_path, ads_path));
        if (ads_ret)
            ret = SMB_VFS_NEXT_NTIMES(handle, ads_path, ts);
        else
            ret = SMB_VFS_NEXT_NTIMES(handle, main_path, ts);
    }
	else {
        ret = SMB_VFS_NEXT_NTIMES(handle, path, ts);
	}
    if (ret == -1) {
        DEBUG(2, ("ads_ntimes: failed trying to set modtime %lu for %s, errno %d\n",
                  ts ? ts[1].tv_sec : 0,path,errno));
        talloc_destroy(ctx);
        return -1;
    }
    /* bug 16528? */
    if (lp_force_root_utime_change()) {
        if ((ret = SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL)) == -1) {
            DEBUG(2, ("ads_utime: failed trying to set modtime %lu for %s, errno %d\n",
                      ts ? ts[1].tv_sec : 0, handle->conn->connectpath,errno));
            talloc_destroy(ctx);
            return -1;
    
        }
    }
    talloc_destroy(ctx);
	return ret;
}

/* ------------------------- */
static int ads_mkdir(vfs_handle_struct *handle, const char *path, mode_t mode)
{
    static int res=0;
    if ((res = SMB_VFS_NEXT_MKDIR(handle, path, mode)) == 0)
        if (lp_force_root_utime_change())
        {
            SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
        }
    return res;
}

/* ------------------------- */
static int unlink_file(const char *path)
{
    int ret = 0;

    become_root();
    ret = unlink(path);
    unbecome_root();
    return ret;
}

/* ------------------------- */
static int unlink_folder(const char *path)
{
    int ret = 0;

    become_root();
    ret = rmdir(path);
    unbecome_root();
    return ret;
}

/* ------------------------- 
   remove all files in an AppleDouble folder
*/
static void rrmdir(TALLOC_CTX *ctx, char *path)
{
        int n;
        char *dpath;
        struct dirent **namelist;
 
        if (!path) return;
 
        n = scandir(path, &namelist, 0, alphasort);
        if (n < 0) {
            /* if the contents of .AppleDouble is not directories, but files, this might fail
             * we'll attempt to delete the path as if it was a file and not a directory */
            unlink_file(path);
            return;
        } 
        while (n --) {
        	if (strcmp(namelist[n]->d_name, ".") == 0 || strcmp(namelist[n]->d_name, "..") == 0) {
                	free(namelist[n]);
                        continue;
                }
                if ((dpath = talloc_asprintf(ctx, "%s/%s",path, namelist[n]->d_name))) {
                	unlink_file(dpath);
                }
                free(namelist[n]);
        }
        free(namelist);
        unlink_folder(path);
}

/* --------------------------- */
static void rrm_adsdir(TALLOC_CTX *ctx, char *path)
{
        int n;
        char *dpath;
        struct dirent **namelist;
 
        if (!path) return;
 
        n = scandir(path, &namelist, 0, alphasort);
        if (n < 0) {
            return;
        } 
        while (n --) {
        	if (strcmp(namelist[n]->d_name, ".") == 0 || strcmp(namelist[n]->d_name, "..") == 0) {
                	free(namelist[n]);
                        continue;
                }
                if ((dpath = talloc_asprintf(ctx, "%s/%s",path, namelist[n]->d_name))) {
                	rrmdir(ctx, dpath);
                }
                free(namelist[n]);
        }
        free(namelist);
        unlink_folder(path);
}

/* ------------------------- 
 * XXX 
 * if in smb.conf there's :
 * delete veto files = True
 * veto files = /.AppleD* /
*/
static int ads_rmdir( vfs_handle_struct *handle, const char *path)
{
        BOOL add = False;
        TALLOC_CTX *ctx = 0;
        char *dpath;
        int  ret = 0;
 
        if (!path) goto exit_rmdir;

	/* .AppleD* */
        strstr(path, ADS_FOLDER) ? (add = False) : (add = True);
 
        if (!(ctx = ADS_TALLOC_INIT("ads_rmdir")))
                goto exit_rmdir;
 
        if (!(dpath = talloc_asprintf(ctx, "%s%s", path, add ? "/"ADS_FOLDER : "")))
		goto exit_rmdir;
		
	/* remove folder .AppleDouble */
        rrm_adsdir(ctx, dpath);
 
exit_rmdir:
    	ret = SMB_VFS_NEXT_RMDIR(handle, path);
        talloc_destroy(ctx);

        if (ret == 0 && lp_force_root_utime_change()) {
		
            SMB_VFS_NEXT_NTIMES (handle, handle->conn->connectpath, NULL);
		}
        return ret;
}

/******************************************************************
* Deny open if both nfs and ads with the same name exists
* Follow the following policy for nfs colon names:
* 1. nfs file exists, ads doesn't exists - open nfs
* 2. nfs file exists, ads file exists - return error
* 3. nfs file doesn't exists, ads doesn't exists - open ads
* 4. nfs file doesn't exists, ads exists = open ads
* Note that all the above is supported only if 'allow nfs colon' is 
* activated and the path name contains ':'
******************************************************************/

static int ads_open(vfs_handle_struct *handle, const char *fname, files_struct *fsp, int flags, mode_t mode)
{
	int ret = -1;
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
    SMB_STRUCT_STAT st_stat;
    int nfs_exists = 0;
    int ads_syntax = 0;
    int ads_exists = 1;

	
	DEBUG(3,("ads: ads_open for %s %x\n", fname, flags));
	if (!(ctx = ADS_TALLOC_INIT("ads_open"))) {
        DEBUG(0,("Failed to talloc ads context\n"));
        return -1;
    }
    /* convert to */
    getMode(mode);
    

    if (strchr(fname, ':')) {
        ads_syntax = 1;
    }
    /* first check if the linux path with the same name already exists */    
#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */
    if (ads_syntax && lp_allow_nfs_colon(SNUM(handle->conn)) && SMB_VFS_NEXT_STAT(handle, fname, &st_stat) == 0){
        nfs_exists = 1;
    }
#endif

    /* Try to open samba alternate data stream */
    if (ads_build_paths(ctx, handle->conn->origpath, fname, &ads_path, &main_path, NULL, flags) >= 0) {

        // Dina: I need additional stat to check if the ads also exists
        // and ads_path is not as main_path (means regular file despite ads syntax)
        DEBUG(3,("ads_path(%s), main_path(%s), nfs_exists(%d)\n", ads_path, main_path, nfs_exists));
        if ( (ads_path == main_path) ||  (nfs_exists && SMB_VFS_NEXT_STAT(handle, ads_path, &st_stat) == -1)) {
            ads_exists = 0;
        }
        // if both file exist - return error
        if (ads_exists && nfs_exists ) {
            errno = EACCES;
            talloc_destroy(ctx);
            DEBUG(0,("Failed: nfs file and ads with the same name(%s)\n", fname));
            return -1;
        }
        if (ads_exists) {
            DEBUG(3,("Next open on %s\n", ads_path));
            ret = SMB_VFS_NEXT_OPEN(handle, ads_path, fsp, flags, mode);
        }else {
            DEBUG(3,("Next open on %s\n", main_path));
			/* Shlomi: fix bug #20415 - in case the file looks like FILE::$DATA it is equivalent to FILE. However fname
			   still contains the ::$DATA string so we need to use the main_path string which is without it. */
            ret = SMB_VFS_NEXT_OPEN(handle, main_path, fsp, flags, mode);
        }
     } else if (nfs_exists) {
        DEBUG(3,("Next open on %s\n", fname));
        ret = SMB_VFS_NEXT_OPEN(handle, fname, fsp, flags, mode);
     }

    /* Shlomi: if a new file is created, update the share's path so that OS 9 will refresh the
       window. This might also apply to OS 10.4
    */
    if (ret != -1 && (flags & (O_CREAT | O_EXCL)) && lp_force_root_utime_change()) {
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
	talloc_destroy(ctx);
    return ret;

} 

static ssize_t ads_write_data(vfs_handle_struct *handle, files_struct *fsp, int fd, const char *buffer,
                              size_t N)
{
    int ret = 0;
    char *ads_path = 0;
    char *main_path = 0;
    
    if (fsp->ads_type == ADS_UNKNOWN) {
        TALLOC_CTX *ctx;

        if (!(ctx = ADS_TALLOC_INIT("ads_write_data")))
            return -1;

        if ( ads_build_paths(ctx, fsp->conn->origpath, fsp->fsp_name, &ads_path, &main_path, NULL, O_WRONLY) >= 0 
             && ads_path && ads_path != main_path) {
            fsp->ads_type = IS_ADS_UPDATE_UTIME;
        } else
            fsp->ads_type = ADS_NONE;

        talloc_destroy(ctx);
    }

    ret = SMB_VFS_NEXT_WRITE(handle, fsp, fd, buffer, N);

    if (ret != -1 && fsp->ads_type == IS_ADS_UPDATE_UTIME) {
        SMB_VFS_NEXT_NTIMES(handle, main_path, NULL);
        fsp->ads_type = IS_ADS_UPDATE_ON_CLOSE;
    }

    
    return ret;
}

static ssize_t ads_pwrite_data(vfs_handle_struct *handle, files_struct *fsp, int fd, const char *buffer,size_t N, SMB_OFF_T offset)
{
    int ret = 0;
    char *ads_path = 0;
    char *main_path = 0;


    if (fsp->ads_type == ADS_UNKNOWN) {
        TALLOC_CTX *ctx;

        if (!(ctx = ADS_TALLOC_INIT("ads_pwrite_data")))
            return -1;

        if ( ads_build_paths(ctx, fsp->conn->origpath, fsp->fsp_name, &ads_path, &main_path, NULL, O_WRONLY) >= 0 
             && ads_path && ads_path != main_path) {
            fsp->ads_type = IS_ADS_UPDATE_UTIME;
        } else
            fsp->ads_type = ADS_NONE;
        talloc_destroy(ctx);
    }

    ret = SMB_VFS_NEXT_PWRITE(handle, fsp, fd, buffer, N, offset);

    if (ret != -1 && fsp->ads_type == IS_ADS_UPDATE_UTIME) {
        SMB_VFS_NEXT_NTIMES(handle, main_path, NULL);
        fsp->ads_type = IS_ADS_UPDATE_ON_CLOSE;
    }

    return ret;
} 

static int ads_close(struct vfs_handle_struct *handle, files_struct *fsp, int fd)
{
    int ret = 0;
    char *ads_path = 0;
	char *main_path = 0;
    TALLOC_CTX *ctx;

    if (!(ctx = ADS_TALLOC_INIT("ads_close")))
		return -1;
   
    if(fsp->ads_type == IS_ADS_UPDATE_ON_CLOSE)
    {
        DEBUG(3,("here\n"));
        if (ads_build_paths(ctx, fsp->conn->origpath, fsp->fsp_name, &ads_path, &main_path, NULL, O_WRONLY) < 0) {
            talloc_destroy(ctx);
            return -1;
        }
    }
    ret = SMB_VFS_NEXT_CLOSE(handle, fsp, fd);
    
    if (ret != -1 && fsp->ads_type == IS_ADS_UPDATE_ON_CLOSE) {
        SMB_VFS_NEXT_NTIMES(handle, main_path, NULL);
	}
    talloc_destroy(ctx);
    return ret;
}

static int isDir(SMB_STRUCT_STAT *st)
{
     if (st == NULL) {
         return 0;
     }
     return S_ISDIR(st->st_mode);
}

static int ads_rename(vfs_handle_struct *handle, const char *old, const char *new_name)
{
	int ret = 0;
	TALLOC_CTX *ctx;
	char *ads_path = 0;
	char *main_path = 0;
	SMB_STRUCT_STAT st;
    SMB_STRUCT_STAT ads_st;
	SMB_STRUCT_STAT *main_info = &st;
	
	DEBUG(3,("ads: ads_rename %s --> %sx\n", old, new_name));
	
	if (!(ctx = ADS_TALLOC_INIT("ads_rename")))
	    return -1;

	if (ads_build_paths(ctx, handle->conn->origpath, old, &ads_path, &main_path, &main_info, 0) < 0) {
#if 0
		/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */
        if (lp_allow_nfs_colon(SNUM(handle->conn)) && (strchr(old,':') != NULL)){
            /*try to rename linux path as at is*/
            DEBUG(3,("The ads doesn't exists, try linux path %s\n", old));
            ret = SMB_VFS_NEXT_RENAME(handle, old, new_name);
        }else
#endif
            ret = -1;
	    talloc_destroy(ctx);
	    return ret;
	}

	DEBUG(3,("after build path: ads_path - %s, main_path - %s\n", ads_path, main_path));
	if (ads_path != main_path) {
	    /* you can't rename an ads ! */
	    talloc_destroy(ctx);
	    errno = EINVAL;
	    return -1;
	}

    
	ret = SMB_VFS_NEXT_RENAME(handle, old, new_name);
    if (ret == 0 && lp_force_root_utime_change()) {
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
    
	if (!ret && !isDir(main_info)) {
	    int  ptr1;
	    int  ptr2;
	    
	    char *ads_old  = ads_dir(ctx, handle->conn->origpath, old, 0);
	    char *ads_new  = ads_dir(ctx, handle->conn->origpath, new_name, 0);

	    /* is dest folder .Adouble there ? */
	    ptr1 = ads_get_path_ptr(ads_new);
	    ptr2 = ads_get_path_ptr(ads_old);

	    ads_new[ptr1] = '\0';
	    ads_old[ptr2] = '\0';
	    if (strcmp(ads_new, ads_old)) {
	    	mkdir(ads_new, 0777);
	    }

	    ads_new[ptr1] = '/';
	    ads_old[ptr2] = '/';
	    
        DEBUG(3,("ads: calling to next_rename with %s --> %sx\n", ads_old, ads_new));
        if (sys_stat(ads_old, &ads_st) == 0)
            SMB_VFS_NEXT_RENAME(handle, ads_old, ads_new);
	}

	talloc_destroy(ctx);
	return ret;
}

/* ------------------------- 
 * For an ADS what do we need to return , ADS ? main DATA?
*/
static int ads_stat(vfs_handle_struct *handle, const char *fname, SMB_STRUCT_STAT *sbuf)
{
	int ret = -1;
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;

	DEBUG(3,("ads: ads_stat for %s\n", fname));

	if (!(ctx = ADS_TALLOC_INIT("ads_stat")))
	    return -1;
	/* which inode ?
	*/
	if (ads_build_paths(ctx, handle->conn->origpath, fname, &ads_path, &main_path, NULL, 0) >= 0) {
        DEBUG(3,("ads: Calling next stat for %s\n", ads_path));
    	ret = SMB_VFS_NEXT_STAT(handle, ads_path, sbuf);
        DEBUG(3,("ads: the ret is %d\n",ret));
    }

#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */

    /* only if the path contains ads syntax (:) and no matching data stream exists, try the linux path as it is*/
    if (ret != 0 && ads_path && ads_path != main_path && lp_allow_nfs_colon(SNUM(handle->conn))) {
        DEBUG(3,("ads: the ads doesn't exists, try linux path %s\n", fname));
        ret = SMB_VFS_NEXT_STAT(handle, fname, sbuf);
    }
#endif
    talloc_destroy(ctx);
    return ret;
}

/* ------------------------- */
static int ads_lstat(vfs_handle_struct *handle, const char *path, SMB_STRUCT_STAT *sbuf)
{
	int ret = -1;
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
	
	if (!(ctx = ADS_TALLOC_INIT("ads_lstat")))
		return -1;
	/* which inode ?
	*/
	if (ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, NULL, 0) >= 0) {
        DEBUG(3,("ads: Calling next lstat for %s\n", ads_path));
        ret = SMB_VFS_NEXT_LSTAT(handle, ads_path, sbuf);
	}

#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */
	
    /* only if the path contains ads syntax (:) and no matching data stread exists, try the linux path as it is*/
    if (ret != 0 && ads_path && ads_path != main_path && lp_allow_nfs_colon(SNUM(handle->conn))) {
        DEBUG(3,("ads: the ads doesn't exists, try linux path %s\n", path));
        ret =  SMB_VFS_NEXT_LSTAT(handle, path, sbuf);
    }                                               
#endif

	talloc_destroy(ctx);
	return ret;
}

/* ------------------------- */
static int ads_unlink(vfs_handle_struct *handle, const char *path)
{
	int ret = -1;

	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
	SMB_STRUCT_STAT st;
	SMB_STRUCT_STAT *main_info = &st;
	
	DEBUG(3,("ads: ads_unlink %s\n", path));
	if (!(ctx = ADS_TALLOC_INIT("ads_unlink")))
		return -1;

	if (ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, &main_info, 0) >= 0) {
        
#if 0
		/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */

        //if there is also a linux file with the same name a:b, don't allow unlink.
        // Stat only for files with ':'
		SMB_STRUCT_STAT st_stat;
        if (lp_allow_nfs_colon(SNUM(handle->conn)) &&  ads_path != main_path && strchr(path, ':')
            &&  SMB_VFS_NEXT_STAT(handle, ads_path, &st_stat) == 0
            &&  SMB_VFS_NEXT_STAT(handle, path, &st_stat) == 0 ){
                DEBUG(1,("Both %s file and ads %s exists on the fs - UNLINK NOT ALLOWED\n", path, path));
                errno = EACCES;
                talloc_destroy(ctx);
                return -1;
        }
#endif
	    ret = SMB_VFS_NEXT_UNLINK(handle, ads_path);
	}

    if (ret == 0 && lp_force_root_utime_change())
	{
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	/*
	   if data stream
	      for each stream
	          unlink
	*/
	}
	if (ret == 0 && ads_path == main_path) {
	    char *ads_base  = ads_dir(ctx, handle->conn->origpath, path, isDir(main_info));
	    struct dirent *dent = 0;
	    DIR *dir = opendir(ads_base);
	    
	    if (dir) {
	        char *dpath;
	        
	        while (NULL != (dent = readdir(dir))) {
		    if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0)
			continue;
		    if (!(dpath = talloc_asprintf(ctx, "%s/%s", ads_base, dent->d_name)))
			continue;
		    /* XXX need to be root ?  */
		    SMB_VFS_NEXT_UNLINK(handle, dpath);		    
		}
	        closedir(dir);
	        rmdir(ads_base);
	    }
	}

#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */
	
    /* only if the path contains ads syntax (:) and no matching data stread exists, try the linux path as it is*/
    if (ret != 0 && ads_path && ads_path != main_path && lp_allow_nfs_colon(SNUM(handle->conn))) {
        DEBUG(3,("ads: the ads doesn't exists, try linux path %s\n", path));
        ret = SMB_VFS_NEXT_UNLINK(handle, path);
    }
	
#endif

	talloc_destroy(ctx);
	return ret;
}

/* ------------------------- */
static int ads_chmod(vfs_handle_struct *handle, const char *path, mode_t mode)
{
	int ret = -1;
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
	SMB_STRUCT_STAT st;
	SMB_STRUCT_STAT *main_info = &st;

	DEBUG(3,("ads: ads_chmod %s\n", path));
	/* if stream 
	       error ?, change only the stream
	*/
	if (!(ctx = ADS_TALLOC_INIT("ads_chmod")))
		return -1;

	if (ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, &main_info, 0) >= 0) {
        ret = SMB_VFS_NEXT_CHMOD(handle, ads_path, mode);
    }           
    
    if (ret == 0 && lp_force_root_utime_change() ) {
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
	/*
	   if data stream
	      for each stream
	          chmod
	*/
	if (ret == 0 && ads_path == main_path) {
	    char *ads_base  = ads_dir(ctx, handle->conn->origpath, path, isDir(main_info));
	    struct dirent *dent = 0;
	    DIR *dir = opendir(ads_base);
	    
	    if (dir) {
	        char *dpath;
	        
	        while (NULL != (dent = readdir(dir))) {
		    if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0)
			continue;
		    if (!(dpath = talloc_asprintf(ctx, "%s/%s", ads_base, dent->d_name)))
			continue;
		    /* XXX need to be root ? */
		    SMB_VFS_NEXT_CHMOD(handle, dpath, mode);
		}
	        closedir(dir);
	        /* XXX need to change ads_base too*/
	    }
	}
    
#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */

   /* only if the path contains ads syntax (:) and no matching data stread exists, try the linux path as it is*/
    if (ret != 0 && ads_path && ads_path != main_path && lp_allow_nfs_colon(SNUM(handle->conn))) {
        DEBUG(3,("ads: the ads doesn't exists, try linux path %s\n", path));
        ret = SMB_VFS_NEXT_CHMOD(handle, path, mode);
    }

#endif

	talloc_destroy(ctx);
	return ret;
}

/* ------------------------- */
static int ads_chown(vfs_handle_struct *handle, const char *path, uid_t uid, gid_t gid)
{
	int ret = -1;
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
	SMB_STRUCT_STAT st;
	SMB_STRUCT_STAT *main_info = &st;

	DEBUG(3,("ads: ads_chown %s\n", path));
	/* if stream 
	       error ?, change only the stream
	*/
	if (!(ctx = ADS_TALLOC_INIT("ads_chown")))
		return -1;

	if (ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, &main_info, 0) >= 0) {
        ret = SMB_VFS_NEXT_CHOWN(handle, ads_path, uid, gid);
	}
    
    if (ret == 0 && lp_force_root_utime_change() ) {
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
	/* if data stream
	      for each stream
	          chmod
	*/
	if (ret == 0 && ads_path == main_path) {
	    char *ads_base  = ads_dir(ctx, handle->conn->origpath, path, isDir(main_info));
	    struct dirent *dent = 0;
	    DIR *dir = opendir(ads_base);
	    
	    if (dir) {
	        char *dpath;
	        
	        while (NULL != (dent = readdir(dir))) {
		    if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0)
			continue;
		    if (!(dpath = talloc_asprintf(ctx, "%s/%s", ads_base, dent->d_name)))
			continue;
		    /* XXX need to be root ?, what do we do in case of error? */
		    SMB_VFS_NEXT_CHOWN(handle, dpath, uid, gid);
		}
	        closedir(dir);
	        SMB_VFS_NEXT_CHOWN(handle, ads_path, uid, gid);
	    }
	}

#if 0
	/* ExaNet 2008 / Samba upgrade project / lp_allow_nfs_colon feature considered obsolete (now always False) */

   /* only if the path contains ads syntax (:) and no matching data stread exists, try the linux path as it is*/
    if (ret != 0 && ads_path && ads_path != main_path && lp_allow_nfs_colon(SNUM(handle->conn))) {
        DEBUG(3,("ads: the ads doesn't exists, try linux path %s\n", path));
        ret = SMB_VFS_NEXT_CHOWN(handle, path, uid, gid);    
    }
	
#endif

	talloc_destroy(ctx);
	return ret;
}


static BOOL ads_symlink(vfs_handle_struct *handle, const char *oldpath, const char *newpath)
{
    int ret=0;
    if (( (ret = SMB_VFS_NEXT_SYMLINK(handle, oldpath, newpath)) == 0 )
        && lp_force_root_utime_change())
	{
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
	return ret;
}

static int ads_link( vfs_handle_struct *handle, const char *oldpath, const char *newpath)
{
    int ret=0;
    if (((ret = SMB_VFS_NEXT_LINK(handle, oldpath, newpath)) == 0) 
        && lp_force_root_utime_change() )
	{
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
	return ret;
}

static int ads_mknod(vfs_handle_struct *handle, const char *path, mode_t mode, SMB_DEV_T dev)
{
    int ret=0;
    if (((ret = SMB_VFS_NEXT_MKNOD(handle, path, mode, dev)) == 0)
        && lp_force_root_utime_change() )
	{
        SMB_VFS_NEXT_NTIMES(handle, handle->conn->connectpath, NULL);
	}
    return ret;
}

/* ---------------------------------- 
 * enumerate 
*/
static ssize_t ads_listads(vfs_handle_struct *handle, const char *path, char *list, size_t size)
{
	char *ads_path = 0;
	char *main_path = 0;
	TALLOC_CTX *ctx;
	size_t     len, total = 0;
	SMB_STRUCT_STAT st;
	SMB_STRUCT_STAT *main_info = &st;
	

	if (!list || !path) {
	        /* aka we have ads functionnality */
		return 0;
	}

	DEBUG(3,("ads: ads_listads %s\n", path));

	if (!(ctx = ADS_TALLOC_INIT("ads_listads")))
		return -1;

	if (ads_build_paths(ctx, handle->conn->origpath, path, &ads_path, &main_path, &main_info, 0) < 0) {
	    talloc_destroy(ctx);
	    return -1;
	}
    DEBUG(3,("ads_path: %s, main_path: %s\n", ads_path, main_path));

	/*
	   if data stream
	      for each stream
	*/
	if (ads_path == main_path) {
	    char *ads_base  = ads_dir(ctx, handle->conn->origpath, path, isDir(main_info));
	    struct dirent *dent = 0;
	    DIR *dir = opendir(ads_base);
	    DEBUG(3, ("ads_base: %s\n", ads_base));
	    /* XXX need to be root ? */
	    if (dir) {
	        
	        while (NULL != (dent = readdir(dir))) {
                DEBUG(10, ("ent_name: %s\n", dent->d_name));
    		    if (strcmp(dent->d_name, ".") == 0 || strcmp(dent->d_name, "..") == 0)
    			continue;
    		    len = strlen(dent->d_name) +8 ;
    		    total += len;
    		    if (total >= size) {
    		        talloc_destroy(ctx);
    		        errno = ERANGE;
    		        return -1;
    		    }
                snprintf (list, len, ":%s:$DATA", dent->d_name);
    		    list += len;
            }
	        closedir(dir);
	    }else{
            DEBUG(3, ("No such dir: %s\n", ads_base));
        }
	}

	talloc_destroy(ctx);
	return total;
}

/* ------------------------------------
 * VFS operations structure */

#ifndef SMB_VFS_OP
#define SMB_VFS_OP(x) ((void *) x)  
#endif

static vfs_op_tuple ads_op_tuples[] = {

    /* Directory operations */

    {SMB_VFS_OP(ads_mkdir),				SMB_VFS_OP_MKDIR,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_rmdir),				SMB_VFS_OP_RMDIR,		SMB_VFS_LAYER_TRANSPARENT},

	/* File operations */

	{SMB_VFS_OP(ads_open),				SMB_VFS_OP_OPEN,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(ads_close),                   SMB_VFS_OP_CLOSE,               SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_rename),			SMB_VFS_OP_RENAME,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_stat),				SMB_VFS_OP_STAT,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_lstat),				SMB_VFS_OP_LSTAT,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_unlink),			SMB_VFS_OP_UNLINK,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_chmod),				SMB_VFS_OP_CHMOD,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_chown),				SMB_VFS_OP_CHOWN,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(ads_ntimes),			SMB_VFS_OP_NTIMES,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_symlink),			SMB_VFS_OP_SYMLINK,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(ads_link),				SMB_VFS_OP_LINK,		SMB_VFS_LAYER_TRANSPARENT},
	{SMB_VFS_OP(ads_mknod),				SMB_VFS_OP_MKNOD,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(ads_write_data),		SMB_VFS_OP_WRITE,		SMB_VFS_LAYER_TRANSPARENT},
    {SMB_VFS_OP(ads_pwrite_data),		SMB_VFS_OP_PWRITE,		SMB_VFS_LAYER_TRANSPARENT},

	/* ADS operations */
	{SMB_VFS_OP(ads_listads),			SMB_VFS_OP_LISTADS,			SMB_VFS_LAYER_OPAQUE},

	{NULL,						SMB_VFS_OP_NOOP,			SMB_VFS_LAYER_NOOP}
};

#ifdef ADS_NEW_MODULE

/*
 * ExaNet 2008 / Samba upgrade project / ADS (Alternate Data Streams)
 * Each Samba module a VFS module should have a init_module function if it's a shared module,
 * or smthing like vfs_example_init if it's staticly linked to samba
 */
NTSTATUS init_module(void)
{
        DEBUG(3, ("ADS: vfs_ads_init\n"));

	if(!ads_build_paths_p)
		ads_build_paths_p = ads_build_paths;

		return smb_register_vfs(SMB_VFS_INTERFACE_VERSION, "ads", ads_op_tuples);
}
         
#else
/* VFS initialisation function.  Return vfs_op_tuple array back to SAMBA. */
vfs_op_tuple *vfs_init(int *vfs_version, struct vfs_ops *def_vfs_ops,struct smb_vfs_handle_struct *vfs_handle)
{
	    *vfs_version = SMB_VFS_INTERFACE_VERSION;
        
	if(!ads_build_paths_p)
		ads_build_paths_p = ads_build_paths;

        DEBUG(3, ("ADS: vfs module loaded\n"));
        return ads_op_tuples;
}
                                             
/* VFS finalization function. */
void vfs_done(connection_struct *conn)
{
	DEBUG(3, ("ADS: vfs module unloaded\n"));

	if(ads_build_paths_p == ads_build_paths)
		ads_build_paths_p = NULL;

}

#endif
