#ifndef EXASMB_DLM_CLIENT_H
#define EXASMB_DLM_CLIENT_H

#include "smb/exasmb_types.h"
#include "exasmb_client_backoff.h"
#include <stdint.h>
#include <sys/types.h>
#include <time.h>

enum {
	CONN_TYPE_DIRECT,
	CONN_TYPE_REMOTE,
};

typedef uint32_t *(*xdr_func)(void *, void *);
const uint32_t INVALID_DOMAIN = ~((uint32_t) 0);

struct exasmb_dlm_client : public exasmb_client_backoff {
	int conn_type;
	int	sockfd;
	uint32_t *ctrl_buf;
	uint32_t *data_buf;
	uint32_t *ctrl_buf_p;
	int shmid;
	int data_size;
	ssize_t	rcved;

	exasmb_dlm_client(int a_domain) throw();

	virtual ~exasmb_dlm_client();

	exasmb_message_error
	call(exasmb_message msg, void *request, xdr_func encode_ctl,
			xdr_func encode_dat);

	exasmb_message_error
	call(exasmb_message msg, void *request, xdr_func encode_ctl,
			xdr_func encode_dat, void *reply,
			xdr_func decode_hdr, xdr_func decode_dat);
	
/*	inline*/ exasmb_message_error
	call(exasmb_message msg, void *request, xdr_func encode);/*
			{ return call(msg, request, encode, (xdr_func) NULL); }*/

/*	inline*/ exasmb_message_error
	call(exasmb_message msg, void *request, xdr_func encode,
							 void *reply, xdr_func decode);/*
			{ return call(msg, request, encode, (xdr_func) NULL,
								reply, decode, (xdr_func) NULL); }*/

/*	inline*/ exasmb_message_error
	call(exasmb_message msg, void *request, xdr_func encode,
							 void *reply, xdr_func decode_hdr,
							 xdr_func decode_dat);/*
			{ return call(msg, request, encode, (xdr_func) NULL,
								reply, decode_hdr, decode_dat); }*/

	void *get_req_buf(size_t size);
	void release_req_buf(void *buf);

	void disconnect(bool sleep = true);

	int domain;
protected:
	virtual const char *socket_name() const;
	mutable char sock_name[64];
private:
	exasmb_message_error send(int len);
	exasmb_message_error recv();
	
    void connect();
	
	int read(void *buf, size_t count);
	int writev(const struct iovec *vector, int count);
};


#endif // #ifndef EXASMB_DLM_CLIENT_H

