#include <netinet/in.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>

// declare dlm_client non-const
#define	EXASMB_DLMCLIENT_SOURCE

#undef XNT_USING_STD_CERR	// dumping errors to std::stream + write(2)
#include "xnt_error_logging.h"

#include "exasmb_dlmclient.h"
#include "exasmb_dlm_client.h"

const int MAX_BUFF	= 1024*1024;

#if 1
// Note: currently DLM client's implementation has no proper error handling
// (only EXASMB_OK/EXASMB_COMM_ERROR are acceptable, otherwise - PANIC;
// see exasmb_if_br_check_error/exasmb_if_share_check_error for details).
// To avoid panics (for example on failover, bug23540), disabling backoff abort algorithm for DLM...

// Note - operator comma returns last operand while executes all operandes
#define BACKOFF_RESULT(regular_error) (backoff(),regular_error)

#else
// Remap 'BACKOFF_ABORTED' event to some existent EXASMB error
#define BACKOFF_ABORTED_ERROR EXASMB_ERROR
#define BACKOFF_RESULT(regular_error) backoff_result(regular_error, BACKOFF_ABORTED_ERROR)
#endif

static void (*arm_alarm_ptr)(void (*callback)(void))  = NULL;
static void (*disarm_alarm_ptr)(bool) = NULL;
static void (*trigger_timeout_ptr)(int) = NULL;
static int trigger_period  = -1;

/* this is a hack to avoid compile issues */
void exasmb_dlm_set_alarm_funcs(void (*in_arm_alarm_ptr)(void (*callback)(void)),
				void (*in_disarm_alarm_ptr)(bool),
				void (*in_trigger_timeout)(int),int in_trigger_period)
{
	arm_alarm_ptr = in_arm_alarm_ptr;
	disarm_alarm_ptr = in_disarm_alarm_ptr;
	trigger_timeout_ptr = in_trigger_timeout;
	trigger_period = in_trigger_period;
	
}





exasmb_dlm_client* global_client = NULL;
void close_socket_function(void)
{
	global_client->disconnect(false);
}
void arm_alarm(exasmb_dlm_client* client)
{
	
	if (arm_alarm_ptr) {
		global_client = client;
		(*arm_alarm_ptr)(close_socket_function);
	}
}



void disarm_alarm(int location)
{
	if (disarm_alarm_ptr) {
		(*disarm_alarm_ptr)(location);
	}
}

/***/
exasmb_message_error
exasmb_dlm_client::call(exasmb_message msg, void *request, xdr_func encode)
{
	return call(msg, request, encode, (xdr_func) NULL);
}
exasmb_message_error
exasmb_dlm_client::call(
	exasmb_message msg,
	void *request, xdr_func encode,
	void *reply, xdr_func decode)
{
	return call(msg, request, encode, (xdr_func) NULL,
					reply, decode, (xdr_func) NULL);
}
exasmb_message_error
exasmb_dlm_client::call(
	exasmb_message msg, 
	void *request, xdr_func encode,
	void *reply, xdr_func decode_hdr, xdr_func decode_dat)
{
	return call(msg, request, encode, (xdr_func) NULL,
						reply, decode_hdr, decode_dat);
}
/***/


exasmb_dlm_client::exasmb_dlm_client(int a_domain) throw() :
	conn_type(CONN_TYPE_DIRECT),
	sockfd(-1),
	ctrl_buf(NULL),
	data_buf(NULL),
	shmid(-1),
	data_size(0),
	domain(a_domain)
{
	sock_name[0] = '\0';
	if (conn_type == CONN_TYPE_DIRECT) {
		ctrl_buf = new uint32_t[1024/sizeof(uint32_t)];
		
		get_req_buf(8192);
		XNT_ERROR("pid " << getpid() << " shmid " << shmid);
		return;
	} else {
		ctrl_buf = new uint32_t[MAX_BUFF/sizeof(uint32_t)];
	}
}

exasmb_dlm_client::~exasmb_dlm_client()
{
	disconnect();
	if (shmid != -1) {
		if (data_buf) {
			if (shmdt(data_buf) < 0)
				XNT_ERROR("shmdt failed: " << strerror(errno));
			data_buf = NULL;
		}
		shmid = -1;
	}
	delete[] ctrl_buf;
	ctrl_buf = NULL;
}

exasmb_message_error
exasmb_dlm_client::call(
	exasmb_message msg,
	void *request,
	xdr_func encode_ctl,
	xdr_func encode_dat)
{
	typedef struct typical_vec_request {
		exasmb_request_vec_header header;
		struct typical_body {
			void	*entries;
		} body;
	} typical_reply;
	exasmb_message_error r;
	exasmb_message_header header = {
		domain, EXASMB_MESSAGE_REQUEST, msg, EXASMB_ERROR, shmid
	};
	uint32_t *p = enc_exasmb_message_header(&header, ctrl_buf);
	p = encode_ctl(request, p);
	if (encode_dat) {
		switch (conn_type) {
			case CONN_TYPE_DIRECT:
				if (((typical_vec_request *) request)->body.entries != data_buf)
					return EXASMB_ERROR;
				break;
			case CONN_TYPE_REMOTE:
				p = encode_dat(request, p);
				break;
			}
	}
	XNT_WARN("sending: " << header.message_code << " : " <<
			 header.message_type << " : " << header.message_error);
	if ((r = send((char *) p - (char *) ctrl_buf)) != EXASMB_OK)
		return r;
	if ((r = recv()) != EXASMB_OK)
		return r;
	ctrl_buf_p = dec_exasmb_message_header(&header, ctrl_buf);
	XNT_WARN("decoded: " << header.message_code << " : " <<
			 header.message_type << " : " << header.message_error << " " << r);
	switch (conn_type) {
	case CONN_TYPE_DIRECT:
		if (header.shmid != -1 && header.shmid != shmid) {
			XNT_DEBUG("new shmid: " << header.shmid);
			if (shmid != -1) {
				if (shmdt(data_buf) < 0)
					XNT_ERROR("shmdt returned error: " << strerror(errno));
				if (shmctl(shmid, IPC_RMID, NULL) < 0)
					XNT_ERROR("shmctl returned error: " << strerror(errno));
				shmid = -1;
			}
			data_buf = (uint32_t *) shmat(header.shmid, NULL, 0);
			if (data_buf != (uint32_t *)-1)
				shmid = header.shmid;
			else {
				shmid = -1;
				XNT_ERROR("shmat failed, errno = " << errno);
				abort();
			}
		}
		break;
	case CONN_TYPE_REMOTE:
		break;
	}
	XNT_DEBUG("client " << this << " done message " << msg << " result " << 
		header.message_error);
	
    if ( header.message_error == EXASMB_COMM_ERROR) {
		return BACKOFF_RESULT(EXASMB_COMM_ERROR);
    }
	reset_backoff_timeout();
    return header.message_error;
}

exasmb_message_error
exasmb_dlm_client::call(
	exasmb_message msg,
	void *request,
	xdr_func encode_ctl,
	xdr_func encode_dat,
	void *reply,
	xdr_func decode_hdr,
	xdr_func decode_dat)
{
	typedef struct typical_vec_reply {
		exasmb_reply_vec_header header;
		struct typical_body {
			void	*entries;
		} body;
	} typical_vec_reply;
	
	exasmb_message_error r = call(msg, request, encode_ctl, encode_dat);
	if ((r == EXASMB_OK) && decode_hdr) {
		ctrl_buf_p = decode_hdr(reply, ctrl_buf_p);
		if (decode_dat) {
			switch (conn_type) {
			case CONN_TYPE_DIRECT:
				((typical_vec_reply *) reply)->body.entries = data_buf;
				break;
			default:
				ctrl_buf_p = decode_dat(reply, ctrl_buf_p);
			}
		}
	}
	return r;
}

void *
exasmb_dlm_client::get_req_buf(size_t size)
{
	if (!data_buf || data_size < size) {
		switch (conn_type) {
		case CONN_TYPE_DIRECT:
			if (size > MAX_BUFF)
				return NULL;
			if (data_buf) {
				if (shmdt(data_buf) < 0)
					XNT_ERROR("shmdt failed: " << data_buf << ": " << strerror(errno));
				data_buf = NULL;
			}
			shmid = shmget(IPC_PRIVATE, size,
						   IPC_CREAT|SHM_R|SHM_W|S_IROTH|S_IWOTH);
			if (shmid == -1) {
					XNT_ERROR("shmget failed: " << strerror(errno));
					return NULL;
			}
			data_buf = (uint32_t *) shmat(shmid, NULL, 0);
			if (data_buf == (uint32_t *) -1) {
					XNT_ERROR("shmat failed: " << strerror(errno));
					// shmctl will free this segment
					data_buf = NULL;
			}
			if (shmctl(shmid, IPC_RMID, NULL) < 0) {
					XNT_ERROR("shmctl failed: " << strerror(errno));
					// we'll continue, not much we can do anyway
			}
			// shmat failed, and shmctl freed this segment, forget it
			if (!data_buf)
				shmid = -1;
			else {
				struct shmid_ds ds;
				if (shmctl(shmid, IPC_STAT, &ds) < 0) {
					XNT_WARN("shmctl failed: " << strerror(errno));
					data_size = size;
				} else {
					data_size = ds.shm_segsz;
				}
			}
			break;
		default:
			if (size > MAX_BUFF - 1024)
				return NULL;
			delete[] data_buf;
			data_size = size;
			int mod = size % sizeof(uint32_t);
			if (mod)
				data_size += sizeof(uint32_t) - mod;
			data_buf = new uint32_t[data_size / sizeof(uint32_t)];
			break;
		}
	}
	return data_buf;
}

void
exasmb_dlm_client::release_req_buf(void *buf)
{
	if (buf != data_buf) {
		XNT_ERROR("bad buf ptr: " << data_buf << buf);
		return;
	}
	if (conn_type == CONN_TYPE_DIRECT) {
		// nothing really to do
	} else {
		delete[] data_buf;
		data_buf = NULL;
		data_size = 0;
	}
}

exasmb_message_error
exasmb_dlm_client::send(int len)
{
	arm_alarm(this);

	uint32_t msg_len = htonl(len);
	struct iovec iov[2] = {
		{&msg_len, sizeof(msg_len)},
		{ctrl_buf, len}
	};
	ssize_t sent = writev(iov, 2);
	if (sent < len + sizeof(msg_len)) {
        disarm_alarm(false);
		return BACKOFF_RESULT(EXASMB_COMM_ERROR);
    }
    disarm_alarm(false);
	
	return EXASMB_OK;
}

exasmb_message_error
exasmb_dlm_client::recv()
{
	uint32_t left = 0;
	arm_alarm(this);
	ssize_t n = read(&left, sizeof(left));
	left = ntohl(left);
	XNT_WARN("n: " << n << " left: " << left);
	if (n < 0 || n < sizeof(left)) {
		disarm_alarm(false);
		return BACKOFF_RESULT(EXASMB_COMM_ERROR);
	}
	for (rcved = 0; left; left -= n, rcved += n) {
		n = read(&ctrl_buf[rcved], left);
		if (n < 0) {
			disarm_alarm(false);
			XNT_WARN("read data failed; n: " << n << " : " << strerror(errno));			
			return BACKOFF_RESULT(EXASMB_COMM_ERROR);
		}
	}
	disarm_alarm(true);
	XNT_WARN("rcved: " << rcved);
	return EXASMB_OK;
}

static inline int
exasmb_sun_len(const struct sockaddr_un& addr)
{
	int len = SUN_LEN(&addr);
	if (!addr.sun_path[0]) {
		len += strlen(&addr.sun_path[1]) + 2;
	}
	return len;
}

const char *exasmb_dlm_client::socket_name() const
{
	if (!sock_name[0])
		sprintf(sock_name, "smbdlm.%d", domain);
	return sock_name;
}

void exasmb_dlm_client::connect()
{
	char buf[128];
	int old_errno = 0;
	struct sockaddr_un addr;
	int loop_count =  trigger_period;
	while (true) {
		if ((sockfd = ::socket(PF_UNIX, SOCK_STREAM, 0)) > 0) {
			::memset(&addr, 0, sizeof(addr));
			addr.sun_family = AF_LOCAL;
			::strcpy(&addr.sun_path[1], socket_name());
			if (::connect(sockfd, (struct sockaddr *) &addr,
						  exasmb_sun_len(addr)) == 0)
				break;
			else  {
				old_errno = errno;
				sprintf(buf, "::connect() failed: %m - (this=%p,uid=%u,gid=%u)",
						this, geteuid(), getegid());				
			}
		} else {
			old_errno = errno;
			sprintf(buf, "::socket(PF_UNIX, SOCK_STREAM, 0) failed");
			goto error_;
		}
error_:
		disarm_alarm(false);
		
		fprintf(stderr, "%s\n", buf);
		
		disconnect(false);
		sleep(1); //must use sleep, using alarm on top , 
		if (loop_count >0) {
		  loop_count--;
		}
		if (loop_count == 0 ) {
		  if (trigger_timeout_ptr){ 
			  arm_alarm(this); //must first make it think it's triggered
		    (*trigger_timeout_ptr)(0);
		  }
		}
		arm_alarm(this);				
	}
}

void exasmb_dlm_client::disconnect(bool sleep_in)
{
	if (sockfd != -1) {
		close(sockfd);
		sockfd = -1;
		if(sleep_in){
			struct timespec sleep_timer;
			sleep_timer.tv_sec = 0;
			sleep_timer.tv_nsec = (111111 + (rand() % 222222) ) *1000;
			nanosleep(&sleep_timer,NULL);
		}
	}
}

int exasmb_dlm_client::read(void *buf, size_t count)
{
	if (sockfd == -1)
		return -1;
	int n = ::read(sockfd, buf, count);
	if (n <= 0) {
	  disarm_alarm(false);
		disconnect();
	}
	return n;
}

int exasmb_dlm_client::writev(const struct iovec *vector, int count)
{
	if (sockfd == -1)
		connect();
	int n = ::writev(sockfd, vector, count);
	if (n < 0 || n < count) {
	  disarm_alarm(false);
		disconnect();
	}
	return n;
}


dlm_client
exasmb_dlm_client_create(int domain)
{
	return new exasmb_dlm_client(domain);
}

exasmb_message_error
exasmb_dlm_client_destroy(dlm_client cln)
{
	delete cln;
	return EXASMB_OK;
}

exasmb_message_error
exasmb_cleanup(dlm_client cln, exasmb_cleanup_request *req)
{
	return cln->call(EXASMB_CLEANUP, req,
					 (xdr_func) enc_exasmb_cleanup_request);
}

void *
exasmb_dlm_client_get_req_buf(dlm_client cln, size_t size)
{
	return cln->get_req_buf(size);
}

void
exasmb_dlm_client_release_req_buf(dlm_client cln, void *buf)
{
	cln->release_req_buf(buf);
}
static pid_t connect_pid;

exasmb_message_error
exasmb_connect(
	dlm_client cln,
	exasmb_connect_request *req)
{
	exasmb_message_error r = EXASMB_ERROR;
	if (connect_pid == 0)
		connect_pid = getpid();
	else if (connect_pid != getpid())
		abort();
	r = cln->call(EXASMB_CONNECT_NLM, req,
				 (xdr_func) enc_exasmb_connect_request);
	return r;
}

/*****************************************************************************/
/*          					Shared locks                                 */
/*****************************************************************************/

exasmb_message_error
exasmb_share_lock_lr(
	dlm_client cln,
	exasmb_share_lock_lr_request *req)
{
	exasmb_lr_context *ctx = &req->lr;
	(void) ctx;
	XNT_ERROR("request: pid 0x" << STD::hex << ctx->pid <<
				" tid 0x" << STD::hex << ctx->tid <<
				" dev 0x" << STD::hex << ctx->dev <<
				" inode 0x" << STD::hex << ctx->inode);
	exasmb_message_error res =
		cln->call(EXASMB_SHARE_LOCK_LR, req,
					(xdr_func) enc_exasmb_simple_request);
	XNT_ERROR("result: "  << res << " pid 0x" << STD::hex << ctx->pid);
	return res;
}

exasmb_message_error
exasmb_share_unlock_lr(
	dlm_client cln,
	exasmb_share_lock_lr_request *req)
{
	exasmb_lr_context *ctx = &req->lr;
	(void) ctx;
	XNT_ERROR("request: pid 0x" << STD::hex << ctx->pid <<
				" tid 0x" << STD::hex << ctx->tid <<
				" dev 0x" << STD::hex << ctx->dev <<
				" inode 0x" << STD::hex << ctx->inode);
	exasmb_message_error res =
		cln->call(EXASMB_SHARE_UNLOCK_LR, req,
					(xdr_func) enc_exasmb_simple_request);
	XNT_ERROR("result: "  << res << " pid 0x" << STD::hex << ctx->pid);
	return res;
}

exasmb_message_error
exasmb_share_get_locks(
	dlm_client cln,
	exasmb_share_get_locks_request *req,
	exasmb_share_get_locks_reply *rep)
{
	return cln->call(EXASMB_SHARE_GET_LOCKS, req,
					(xdr_func) enc_exasmb_get_locks_request,
					rep, (xdr_func) dec_exasmb_share_get_locks_reply_header,
					(xdr_func) dec_exasmb_share_get_locks_reply_body);
}

exasmb_message_error
exasmb_share_add_lock(
	dlm_client cln,
	exasmb_share_add_lock_request *req)
{
	return cln->call(EXASMB_SHARE_ADD_LOCK, req,
					(xdr_func) enc_exasmb_share_add_lock_request_header,
					(xdr_func) enc_exasmb_share_add_lock_request_body);
}



exasmb_message_error
exasmb_share_delete_lock(
	dlm_client cln,
	exasmb_share_delete_lock_request *req,
	exasmb_share_delete_lock_reply *rep)
{
	return cln->call(EXASMB_SHARE_DELETE_LOCK, req,
					(xdr_func) enc_exasmb_share_delete_lock_request_header,
					(xdr_func) enc_exasmb_share_delete_lock_request_body,
					rep,
					(xdr_func) dec_exasmb_share_delete_lock_reply,
					NULL);
}

exasmb_message_error
exasmb_share_set_delete_on_close(
	dlm_client cln,
	exasmb_share_set_delete_on_close_request *req)
{
	return cln->call(EXASMB_SHARE_SET_DELETE_ON_CLOSE, req,
					 (xdr_func) enc_exasmb_share_set_delete_on_close_request);
}

exasmb_message_error
exasmb_share_modify_oplock(
	dlm_client cln,
	exasmb_share_modify_oplock_request *req)
{
	return cln->call(EXASMB_SHARE_MODIFY_OPLOCK, req,
					 (xdr_func) enc_exasmb_share_modify_oplock_request);
}

exasmb_message_error
exasmb_share_get_delete_on_close(
	dlm_client cln, 
	exasmb_share_get_delete_on_close_request *req,
	exasmb_share_get_delete_on_close_reply *rep)
{
	return cln->call(EXASMB_SHARE_GET_DELETE_ON_CLOSE, req,
					 (xdr_func) enc_exasmb_share_get_delete_on_close_request,
					 rep,
					 (xdr_func) dec_exasmb_share_get_delete_on_close_reply);
}

exasmb_message_error
exasmb_br_lock_lr(
	dlm_client cln,
	exasmb_br_lock_lr_request *req)
{
	return cln->call(EXASMB_BR_LOCK_LR, req,
					(xdr_func) enc_exasmb_br_lock_lr_request);
}

exasmb_message_error
exasmb_br_unlock_lr(
	dlm_client cln,
	exasmb_br_unlock_lr_request *req)
{
	return cln->call(EXASMB_BR_UNLOCK_LR, req,
					(xdr_func) enc_exasmb_br_unlock_lr_request);
}

exasmb_message_error
exasmb_br_get_locks(
	dlm_client cln,
	exasmb_br_get_locks_request *req,
	exasmb_br_get_locks_reply *rep)
{
	return cln->call(EXASMB_BR_GET_LOCKS, req,
					(xdr_func) enc_exasmb_br_get_locks_request,
					rep,
					(xdr_func) dec_exasmb_br_get_locks_reply_header,
					(xdr_func) dec_exasmb_br_get_locks_reply_body);
}

exasmb_message_error
exasmb_br_chk_lock(
	dlm_client cln,
	exasmb_br_chk_lock_request *req)
{
	return cln->call(EXASMB_BR_CHK_LOCK, req,
					(xdr_func) enc_exasmb_br_unlock_lr_request);
}

exasmb_message_error
exasmb_br_add_lock(
	dlm_client cln,
	exasmb_br_add_lock_request *req)
{
	return cln->call(EXASMB_BR_ADD_LOCK, req,
					(xdr_func) enc_exasmb_share_add_lock_request_header,
					(xdr_func) enc_exasmb_share_add_lock_request_body);
}

exasmb_message_error
exasmb_br_delete_lock(
	dlm_client cln,
	exasmb_br_delete_lock_request *req)
{
	return cln->call(EXASMB_BR_DELETE_LOCK, req,
					(xdr_func) enc_exasmb_br_delete_lock_request_header,
					(xdr_func) enc_exasmb_br_delete_lock_request_body);
}


/*****************************************************************************/
/*                              Files close                                  */
/*****************************************************************************/

exasmb_message_error
exasmb_file_close(
	dlm_client cln,
	exasmb_file_close_request *req)
{
	exasmb_message_error r = cln->call(EXASMB_BR_FILE_CLOSE, req,
					 (xdr_func) enc_exasmb_file_close_request);
	(void) r;
	return EXASMB_OK;
}

