#ifndef EXASMB_VFSFS_CLIENT_H
#define EXASMB_VFSFS_CLIENT_H

#include "exasmb_vfsfs_message.h"
#include "exasmb_client_backoff.h"
#include <stdint.h>
#include <sys/types.h>


enum {
	CONN_TYPE_DIRECT,
	CONN_TYPE_REMOTE,
};


typedef uint32_t *(*xdr_encode)(const smb_acl_request *, uint32_t *);
typedef const uint32_t *(*xdr_decode)(smb_acl_message *, const uint32_t *);

struct exasmb_vfsfs_client : public exasmb_client_backoff {
	int conn_type;
	int	sockfd;
	uint32_t *ctrl_buf;
	int data_size;
	ssize_t	rcved;

	exasmb_vfsfs_client(int a_domain) throw();

	virtual ~exasmb_vfsfs_client();

	smb_acl_error
	call(const smb_acl_request *request, xdr_encode encode,
         smb_acl_message *reply, xdr_decode decode, int fd = -1, bool overide_fd_sent = false);
	
protected:
	virtual const char *socket_name() const;
private:
    
	exasmb_vfsfs_client();
	smb_acl_error send(int len);
	smb_acl_error recv();
	void connect();
	void disconnect();
	int read(void *buf, size_t count);
	int writev(const struct iovec *vector, int count);
	int	domain;
    int ancillary_fd;
    bool ancillary_fd_sent;
};


#endif // #ifndef EXASMB_VFSFS_CLIENT_H


