#include <sys/types.h>
#include <netinet/in.h>
#include <string.h>
#include <byteswap.h>
#include <iostream>
#include "exasmb_xdr.h"
/******************************************************************************/
/*                          Basic types decode/encode functions			 	  */
/******************************************************************************/

#define xdr_sizeof(x) (sizeof(x) > sizeof(uint32_t) ? \
	sizeof(x) : sizeof(uint32_t))

static inline uint32_t *
dec16(uint16_t *val, uint32_t *p)
{
    *val = (uint16_t) ntohl(*p);

    return p + 1;
}

static inline uint32_t *
enc16(const uint16_t *val, uint32_t *p)
{
    *p = htonl((uint32_t) *val);

    return p + 1;
}

static inline uint32_t *
dec32(uint32_t *val, uint32_t *p)
{
    *val = ntohl(*p);

    return p + 1;
}

static inline uint32_t *
enc32(const uint32_t *val, uint32_t *p)
{
    *p = htonl(*val);

    return p + 1;
}

static inline uint32_t *
dec64(uint64_t *val, uint32_t *p)
{
#ifdef WORDS_BIGENDIAN
    *val = *(uint64_t *) p;
#else
    uint32_t *val32 = (uint32_t *)val;
    *val32++ = ntohl(p[1]);
    *val32 = ntohl(p[0]);
#endif
    return p + 2;
}

static inline uint32_t *
enc64(const uint64_t *val, uint32_t *p)
{
#ifdef WORDS_BIGENDIAN
    *(uint64_t *) p = *val;
#else
    uint32_t *val32 = (uint32_t *)val;
    p[1] = htonl(*val32); val32++;
    p[0] = htonl(*val32);
#endif
    return p + 2;
}

/*****************************************************************************/

inline uint32_t *
enc_exasmb_message_type(exasmb_message_type *val, uint32_t *p)
{
    *p++ = htonl((uint32_t) *val);

    return p;
}

inline uint32_t *
dec_exasmb_message_type(exasmb_message_type *val, uint32_t *p)
{
    *val = (exasmb_message_type) ntohl(*p);

    return p + 1;
}

inline uint32_t *
enc_exasmb_message(exasmb_message *val, uint32_t *p)
{
    *p++ = htonl((uint32_t) *val);

    return p;
}

inline uint32_t *
dec_exasmb_message(exasmb_message *val, uint32_t *p)
{
    *val = (exasmb_message) ntohl(*p);

    return p + 1;
}

inline uint32_t *
enc_exasmb_message_error(exasmb_message_error *val, uint32_t *p)
{
    *p++ = htonl((uint32_t) *val);

    return p;
}

inline uint32_t *
dec_exasmb_message_error(exasmb_message_error *val, uint32_t *p)
{
    *val = (exasmb_message_error) ntohl(*p);

    return p + 1;
}

uint32_t *
enc_exasmb_message_header(exasmb_message_header *val, uint32_t *p)
{
	p = enc32(&val->domain, p);
	p = enc_exasmb_message_type(&val->message_type, p);
	p = enc_exasmb_message(&val->message_code, p);
	p = enc_exasmb_message_error(&val->message_error, p);
	*p++ = htonl(val->shmid);

    return p;
}

uint32_t *
dec_exasmb_message_header(exasmb_message_header *val, uint32_t *p)
{
	p = dec32(&val->domain, p);
	p = dec_exasmb_message_type(&val->message_type, p);
	p = dec_exasmb_message(&val->message_code, p);
	p = dec_exasmb_message_error(&val->message_error, p);
	val->shmid = ntohl(*p);

    return p + 1;
}

uint32_t *
enc_exasmb_conn_context(exasmb_conn_context *val, uint32_t *p)
{
	int len = strlen(val->inst_name) + 1;
	
    p = enc64(&val->vers, p);
	p = enc32(&val->inst_id, p);
	memcpy(p, val->inst_name, len);

    /* Add padding to 4 bytes */
    p += (len + sizeof(int32_t) - 1) / sizeof(int32_t);

    return p;
}

uint32_t *
dec_exasmb_conn_context(exasmb_conn_context *val, uint32_t *p)
{
    p = dec64(&val->vers, p);
	p = dec32(&val->inst_id, p);
	val->inst_name = (char *) p;

	/* Add padding to 4 bytes */
	p += (strlen(val->inst_name) + 1 + sizeof(int32_t) - 1) / sizeof(int32_t);

	return p;
}

uint32_t *
enc_exasmb_lr_context(const exasmb_lr_context *val, uint32_t *p)
{
    p = enc32(&val->pid, p);
    p = enc16(&val->tid, p);
    p = enc64(&val->dev, p);
    p = enc64(&val->inode, p);
	p = enc32(&val->file_id, p);
    p = enc64(&val->sequence_id, p);

    p = enc32(&val->data.len, p);
    memcpy(p, val->data.buf, val->data.len);
	/* Add padding to 4 bytes */
	p += (val->data.len + sizeof(int32_t) - 1) / sizeof(int32_t);
    return p;
}

uint32_t *
dec_exasmb_lr_context(exasmb_lr_context *val, uint32_t *p)
{
    p = dec32(&val->pid, p);
    p = dec16(&val->tid, p);
    p = dec64(&val->dev, p);
    p = dec64(&val->inode, p);
	p = dec32(&val->file_id, p);
    p = dec64(&val->sequence_id, p);

	p = dec32(&val->data.len, p);
	memcpy(val->data.buf, p, val->data.len);
	/* Add padding to 4 bytes */
	p += (val->data.len + sizeof(int32_t) - 1) / sizeof(int32_t);
    return p;
}


uint32_t *enc_exasmb_cleanup_request(exasmb_cleanup_request *val, uint32_t *p)
{
	p = enc32(&val->pid, p);
	
	return p;
}
uint32_t *dec_exasmb_cleanup_request(exasmb_cleanup_request *val, uint32_t *p)
{
	p = dec32(&val->pid, p);
	
	return p;
}


uint32_t *
enc_exasmb_simple_request(exasmb_simple_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->lock_count, p);
    
	return p;
}

uint32_t *
dec_exasmb_simple_request(exasmb_simple_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	p = dec32(&val->lock_count, p);
    
	return p;
}

uint32_t *
enc_exasmb_get_locks_request(exasmb_get_locks_request *val, uint32_t *p)
{
    p = enc_exasmb_lr_context(&val->lr, p);
    p = enc64(&val->cookie, p);
    p = enc32(&val->num_locks, p);

    return p;
}

uint32_t *
dec_exasmb_get_locks_request(exasmb_get_locks_request *val, uint32_t *p)
{
    p = dec_exasmb_lr_context(&val->lr, p);
    p = dec64(&val->cookie, p);
    p = dec32(&val->num_locks, p);

    return p;
}

uint32_t *
enc_exasmb_reply_vec_header(exasmb_reply_vec_header *val, uint32_t *p)
{
	p = enc64(&val->cookie, p);
	p = enc32(&val->num_entries, p);
	p = enc32(&val->delete_on_close, p);
	
	return p;
}

uint32_t *
dec_exasmb_reply_vec_header(exasmb_reply_vec_header *val, uint32_t *p)
{
	p = dec64(&val->cookie, p);
	p = dec32(&val->num_entries, p);
	p = dec32(&val->delete_on_close, p);
	
	return p;
}

uint32_t *
enc_exasmb_request_vec_header(exasmb_request_vec_header *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
    *p = htonl(val->num_entries);
	
	return p + 1;
}

uint32_t *
dec_exasmb_request_vec_header(exasmb_request_vec_header *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
    val->num_entries = ntohl(*p);
	
	return p + 1;
}

uint32_t *
enc_exasmb_time_val(const exasmb_time_val *val, uint32_t *p)
{
    p = enc32(&val->tv_sec, p);
    p = enc32(&val->tv_usec, p);

    return p;
}

uint32_t *
dec_exasmb_time_val(exasmb_time_val *val, uint32_t *p)
{
    p = dec32(&val->tv_sec, p);
    p = dec32(&val->tv_usec, p);

    return p;
}

size_t size_exasmb_time_val(exasmb_time_val *val)
{
	return xdr_sizeof(val->tv_sec) + xdr_sizeof(val->tv_usec);
}

uint32_t *
enc_exasmb_shared_lock(const exasmb_shared_lock *val, uint32_t *p)
{
    p = enc32(&val->pid, p);
	p = enc16(&val->node, p);
    p = enc32(&val->smbid, p);
	p = enc16(&val->op_mid, p);
    p = enc16(&val->op_type, p);
    p = enc32(&val->share_access, p);
    p = enc32(&val->access_mask, p);
    p = enc_exasmb_time_val(&val->open_time, p);
	p = enc64(&val->dev, p);
	p = enc64(&val->inode, p);
    p = enc32(&val->share_file_id, p);
	p = enc32(&val->private_options, p);
	p = enc32(&val->uid, p);
	p = enc16(&val->flags, p); 

    return p;
}

uint32_t *
dec_exasmb_shared_lock(exasmb_shared_lock *val, uint32_t *p)
{
    p = dec32(&val->pid, p);
	p = dec16(&val->node, p);
    p = dec32(&val->smbid, p);
	p = dec16(&val->op_mid, p);
    p = dec16(&val->op_type, p);
    p = dec32(&val->share_access, p);
    p = dec32(&val->access_mask, p);
    p = dec_exasmb_time_val(&val->open_time, p);
	p = dec64(&val->dev, p);
	p = dec64(&val->inode, p);
    p = dec32(&val->share_file_id, p);
	p = dec32(&val->private_options, p);
	p = dec32(&val->uid, p);
	p = dec16(&val->flags, p); 

    return p;
}

size_t size_exasmb_shared_lock(exasmb_shared_lock *val)
{
	(void) val;
	return xdr_sizeof(val->pid) +
		xdr_sizeof(val->node) +
		xdr_sizeof(val->smbid) +
		xdr_sizeof(val->op_mid) +
		xdr_sizeof(val->op_type) +
		xdr_sizeof(val->share_access) +
		xdr_sizeof(val->access_mask) +
		size_exasmb_time_val(&val->open_time) +
		xdr_sizeof(val->dev) +
		xdr_sizeof(val->inode) +
		xdr_sizeof(val->share_file_id) +
		xdr_sizeof(val->private_options) +
		xdr_sizeof(val->uid) +
		xdr_sizeof(val->flags);
}

uint32_t *
enc_exasmb_share_get_locks_reply_body(exasmb_share_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_share_get_locks_reply_body(exasmb_share_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_share_add_lock_request_body(exasmb_share_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_share_add_lock_request_body(exasmb_share_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_share_delete_lock_request_body(exasmb_share_delete_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_share_delete_lock_request_body(exasmb_share_delete_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_shared_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_share_delete_lock_reply(exasmb_share_delete_lock_reply *val, uint32_t *p)
{
   	p = enc32(&val->num_left, p);
	p = enc16(&val->locks_were_deleted, p);
	p = enc_exasmb_shared_lock(&val->lock_deleted, p);
	p = enc32(&val->delete_on_close, p);

	return p;
}

uint32_t *
dec_exasmb_share_delete_lock_reply(exasmb_share_delete_lock_reply *val, uint32_t *p)
{
    p = dec32(&val->num_left, p);
	p = dec16(&val->locks_were_deleted, p);
    p = dec_exasmb_shared_lock(&val->lock_deleted, p);
	p = dec32(&val->delete_on_close, p);

    return p;
}


uint32_t *
enc_exasmb_share_delete_by_pid(exasmb_share_delete_by_pid *val, uint32_t *p)
{
	p = enc32(&val->pid, p);
	p = enc32(&val->smbid, p);

	return p;
}

uint32_t *
dec_exasmb_share_delete_by_pid(exasmb_share_delete_by_pid *val, uint32_t *p)
{
    p = dec32(&val->pid, p);
    p = dec32(&val->smbid, p);

    return p;
}

uint32_t *
enc_exasmb_share_delete_lock_by_pid_request_body(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p)
{
   uint32_t i;

   for (i = 0; i < val->header.num_entries; i++) {
       p = enc_exasmb_share_delete_by_pid(&val->body.entries[i], p);
   }

   return p;
}

uint32_t *
dec_exasmb_share_delete_lock_by_pid_request_body(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p)
{
   uint32_t i;

   for (i = 0; i < val->header.num_entries; i++) {
       p = dec_exasmb_share_delete_by_pid(&val->body.entries[i], p);
   }

   return p;
}


uint32_t *
enc_exasmb_share_set_delete_on_close_request(const exasmb_share_set_delete_on_close_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->delete_on_close, p);
    
	return p;
}

uint32_t *
dec_exasmb_share_set_delete_on_close_request(exasmb_share_set_delete_on_close_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	p = dec32(&val->delete_on_close, p);
	
	return p;
}

uint32_t *
enc_exasmb_share_modify_oplock_request(const exasmb_share_modify_oplock_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->set_port, p);
   	p = enc_exasmb_shared_lock(&val->entry, p);
    
	return p;
}

uint32_t *
dec_exasmb_share_modify_oplock_request(exasmb_share_modify_oplock_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	p = dec32(&val->set_port, p);
   	p = dec_exasmb_shared_lock(&val->entry, p);
	
	return p;
}

uint32_t *
enc_exasmb_share_get_delete_on_close_request(const exasmb_share_get_delete_on_close_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	
	return p;
}

uint32_t *
dec_exasmb_share_get_delete_on_close_request(exasmb_share_get_delete_on_close_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	
	return p;
}

uint32_t *
enc_exasmb_share_get_delete_on_close_reply(const exasmb_share_get_delete_on_close_reply *val, uint32_t *p)
{
	p = enc32(&val->delete_on_close, p);
	
	return p;
}

uint32_t *
dec_exasmb_share_get_delete_on_close_reply(exasmb_share_get_delete_on_close_reply *val, uint32_t *p)
{
	p = dec32(&val->delete_on_close, p);
	
	return p;
}


/******************************************************************************/
/*                              Byte Range locks                              */
/******************************************************************************/

uint32_t *
enc_exasmb_br_lock(exasmb_br_lock *val, uint32_t *p)
{
	p = enc16(&val->smbpid, p);
	p = enc16(&val->tid, p);
	p = enc32(&val->pid, p);
	p = enc32(&val->node, p);
	p = enc32(&val->smbid, p);
    p = enc32(&val->fnum, p);
    p = enc64(&val->start, p);
    p = enc64(&val->size, p);
    p = enc32(&val->lock_type, p);

    return p;
}

uint32_t *
dec_exasmb_br_lock(exasmb_br_lock *val, uint32_t *p)
{
	p = dec16(&val->smbpid, p);
	p = dec16(&val->tid, p);
	p = dec32(&val->pid, p);
	p = dec32(&val->node, p);
	p = dec32(&val->smbid, p);
    p = dec32(&val->fnum, p);
    p = dec64(&val->start, p);
    p = dec64(&val->size, p);
    p = dec32(&val->lock_type, p);

    return p;
}

uint32_t *
enc_exasmb_br_get_locks_reply_body(exasmb_br_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
           p = enc_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_br_get_locks_reply_body(exasmb_br_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_br_chk_lock_request_body(exasmb_br_chk_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_br_chk_lock_request_body(exasmb_br_chk_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_br_add_lock_request_body(exasmb_br_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_br_add_lock_request_body(exasmb_br_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_br_delete_lock_request_body(exasmb_br_delete_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_br_delete_lock_request_body(exasmb_br_delete_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_br_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_file_close_request(exasmb_file_close_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->fnum, p);
	
    return p;
}

uint32_t *
dec_exasmb_file_close_request(exasmb_file_close_request *val, uint32_t *p)
{
    p = dec_exasmb_lr_context(&val->lr, p);
    p = dec32(&val->fnum, p);

    return p;
}

uint32_t *
enc_exasmb_blk_lock(exasmb_blk_lock *val, uint32_t *p)
{
	p = enc32(&val->com_type, p);
    p = enc32(&val->expire_time, p);
	p = enc32(&val->lock_num, p);
    p = enc32(&val->mid, p);
	p = enc32(&val->fnum, p);

    p = enc32(&val->length, p);
    memcpy(p, val->inbuf, val->length);

	/* Add padding to 4 bytes */
	p += (val->length + sizeof(int32_t) - 1) / sizeof(int32_t);

    return p;
}


uint32_t *
dec_exasmb_blk_lock(exasmb_blk_lock *val, uint32_t *p)
{
    p = dec32(&val->com_type, p);
    p = dec32(&val->expire_time, p);
    p = dec32(&val->lock_num, p);
    p = dec32(&val->mid, p);
    p = dec32(&val->fnum, p);

    p = dec32(&val->length, p);
    memcpy(val->inbuf, p, val->length);

	/* Add padding to 4 bytes */
	p += (val->length + sizeof(int32_t) - 1) / sizeof(int32_t);

    return p;
}

uint32_t *
enc_exasmb_blk_get_locks_reply_body(exasmb_blk_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_blk_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_blk_get_locks_reply_body(exasmb_blk_get_locks_reply *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = dec_exasmb_blk_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_blk_add_lock_request_body(exasmb_blk_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_blk_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
dec_exasmb_blk_add_lock_request_body(exasmb_blk_add_lock_request *val, uint32_t *p)
{
    uint32_t i;

    for (i = 0; i < val->header.num_entries; i++) {
        p = enc_exasmb_blk_lock(&val->body.entries[i], p);
    }

    return p;
}

uint32_t *
enc_exasmb_blk_delete_lock_by_fid_request(exasmb_blk_delete_lock_by_fid_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->fnum, p);

	return p;
}

uint32_t *
dec_exasmb_blk_delete_lock_by_fid_request(exasmb_blk_delete_lock_by_fid_request *val, uint32_t *p)
{
    p = dec_exasmb_lr_context(&val->lr, p);
    p = dec32(&val->fnum, p);

    return p;
}

uint32_t *
enc_exasmb_blk_delete_lock_by_mid_request(exasmb_blk_delete_lock_by_mid_request *val, uint32_t *p)
{
    p = enc_exasmb_lr_context(&val->lr, p);
    p = enc32(&val->mid, p);

    return p;
}

uint32_t *
dec_exasmb_blk_delete_lock_by_mid_request(exasmb_blk_delete_lock_by_mid_request *val, uint32_t *p)
{
    p = dec_exasmb_lr_context(&val->lr, p);
    p = dec32(&val->mid, p);

    return p;
}


uint32_t *
enc_exasmb_blk_check_timeout_request(exasmb_blk_check_timeout_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc32(&val->expire_time, p);
	
	return p;
}

uint32_t *
dec_exasmb_blk_check_timeout_request(exasmb_blk_check_timeout_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	p = dec32(&val->expire_time, p);
	
	return p;
}

uint32_t *
enc_exasmb_blk_check_timeout_reply(exasmb_blk_check_timeout_reply *val, uint32_t *p)
{
	p = enc16(&val->queue_empty, p);
	p = enc32(&val->time, p);

    return p;
}

uint32_t *
dec_exasmb_blk_check_timeout_reply(exasmb_blk_check_timeout_reply *val, uint32_t *p)
{
	p = dec16(&val->queue_empty, p);
	p = dec32(&val->time, p);

    return p;
}

uint32_t *
enc_exasmb_blk_delete_and_next_request(exasmb_blk_delete_and_next_request *val, uint32_t *p)
{
	p = enc_exasmb_lr_context(&val->lr, p);
	p = enc64(&val->cookie, p);

 	return p;
}

uint32_t *
dec_exasmb_blk_delete_and_next_request(exasmb_blk_delete_and_next_request *val, uint32_t *p)
{
	p = dec_exasmb_lr_context(&val->lr, p);
	p = dec64(&val->cookie, p);

	return p;
}

uint32_t *
enc_exasmb_blk_delete_and_next_reply(exasmb_blk_delete_and_next_reply *val, uint32_t *p)
{
	p = enc64(&val->cookie, p);

	return p;
}

uint32_t *
dec_exasmb_blk_delete_and_next_reply(exasmb_blk_delete_and_next_reply *val, uint32_t *p)
{
    p = dec64(&val->cookie, p);

    return p;
}
std::ostream& 
operator<<(std::ostream& out, exasmb_message code)
{
	switch(code){
	case EXASMB_CONNECT_NLM:
		out << "EXASMB_CONNECT_NLM";
		break;
	case EXASMB_CLEANUP:
		out << "EXASMB_CLEANUP";
		break;
	case EXASMB_SHARE_LOCK_LR:
		out << "EXASMB_SHARE_LOCK_LR";
		break;
	case EXASMB_SHARE_UNLOCK_LR:
		out << "EXASMB_SHARE_UNLOCK_LR";
		break;
	case EXASMB_SHARE_GET_LOCKS:
		out << "EXASMB_SHARE_GET_LOCKS";
		break;
	case EXASMB_SHARE_ADD_LOCK:
		out << "EXASMB_SHARE_ADD_LOCK";
		break;
	case EXASMB_SHARE_DELETE_LOCK:
		out << "EXASMB_SHARE_DELETE_LOCK";
		break;
	case EXASMB_SHARE_MODIFY_OPLOCK:
		out << "EXASMB_SHARE_MODIFY_OPLOCK";
		break;
	case EXASMB_SHARE_SET_DELETE_ON_CLOSE:
		out << "EXASMB_SHARE_SET_DELETE_ON_CLOSE";
		break;
	case EXASMB_SHARE_GET_DELETE_ON_CLOSE:
		out << "EXASMB_SHARE_GET_DELETE_ON_CLOSE";
		break;
	case EXASMB_BR_LOCK_LR:
		out << "EXASMB_BR_LOCK_LR";
		break;
	case EXASMB_BR_UNLOCK_LR:
		out << "EXASMB_BR_UNLOCK_LR";
		break;
	case EXASMB_BR_GET_LOCKS:
		out << "EXASMB_BR_GET_LOCKS";
		break;
	case EXASMB_BR_CHK_LOCK:
		out << "EXASMB_BR_CHK_LOCK";
		break;
	case EXASMB_BR_ADD_LOCK:
		out << "EXASMB_BR_ADD_LOCK";
		break;
	case EXASMB_BR_DELETE_LOCK:
		out << "EXASMB_BR_DELETE_LOCK";
		break;
	case EXASMB_BR_FILE_CLOSE:
		out << "EXASMB_BR_FILE_CLOSE";
		break;
	default:
		out << "unknown exasmb_code";
	}
	return out;
}
