#ifndef _EXASMB_XDR_H_
#define _EXASMB_XDR_H_

#include "smb/exasmb_types.h"
#include <stddef.h>

#ifdef __cplusplus
std::ostream& 
operator<<(std::ostream& out, exasmb_message code);
#endif

uint32_t *enc_exasmb_message_type(exasmb_message_type *val, uint32_t *p);
uint32_t *dec_exasmb_message_type(exasmb_message_type *val, uint32_t *p);


uint32_t *enc_exasmb_message(exasmb_message *val, uint32_t *p);
uint32_t *dec_exasmb_message(exasmb_message *val, uint32_t *p);


uint32_t *enc_exasmb_message_error(exasmb_message_error *val, uint32_t *p);
uint32_t *dec_exasmb_message_error(exasmb_message_error *val, uint32_t *p);


uint32_t *enc_exasmb_message_header(exasmb_message_header *val, uint32_t *p);
uint32_t *dec_exasmb_message_header(exasmb_message_header *val, uint32_t *p);


uint32_t *enc_exasmb_conn_context(exasmb_conn_context *val, uint32_t *p);
uint32_t *dec_exasmb_conn_context(exasmb_conn_context *val, uint32_t *p);


/* The lock resource context */
uint32_t *enc_exasmb_lr_context(const exasmb_lr_context *val, uint32_t *p);
uint32_t *dec_exasmb_lr_context(exasmb_lr_context *val, uint32_t *p);


/* A connection request to ExaStore */
static inline uint32_t *enc_exasmb_connect_request(exasmb_connect_request *val, uint32_t *p)
	{ return enc_exasmb_conn_context(&val->conn, p); }
static inline uint32_t *dec_exasmb_connect_request(exasmb_connect_request *val, uint32_t *p)
	{ return dec_exasmb_conn_context(&val->conn, p); }


/* Cleanup request for spesific host (pid) */
uint32_t *enc_exasmb_cleanup_request(exasmb_cleanup_request *val, uint32_t *p);
uint32_t *dec_exasmb_cleanup_request(exasmb_cleanup_request *val, uint32_t *p);


/* A simple request on a LR */
uint32_t *enc_exasmb_simple_request(exasmb_simple_request *val, uint32_t *p);
uint32_t *dec_exasmb_simple_request(exasmb_simple_request *val, uint32_t *p);


/*  Get entries from LR */
uint32_t *enc_exasmb_get_locks_request(exasmb_get_locks_request *val, uint32_t *p);
uint32_t *dec_exasmb_get_locks_request(exasmb_get_locks_request *val, uint32_t *p);


uint32_t *enc_exasmb_reply_vec_header(exasmb_reply_vec_header *val, uint32_t *p);
uint32_t *dec_exasmb_reply_vec_header(exasmb_reply_vec_header *val, uint32_t *p);


uint32_t *enc_exasmb_request_vec_header(exasmb_request_vec_header *val, uint32_t *p);
uint32_t *dec_exasmb_request_vec_header(exasmb_request_vec_header *val, uint32_t *p);

/******************************************************************************/
/*          					Shared locks                                  */
/******************************************************************************/

static inline uint32_t *enc_exasmb_share_lock_lr_request(exasmb_share_lock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_share_lock_lr_request(exasmb_share_lock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_share_unlock_lr_request(exasmb_share_unlock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_share_unlock_lr_request(exasmb_share_unlock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_share_get_locks_request(exasmb_share_get_locks_request *val, uint32_t *p)
	{ return enc_exasmb_get_locks_request(val, p); }
static inline uint32_t *dec_exasmb_share_get_locks_request(exasmb_share_get_locks_request *val, uint32_t *p)
	{ return dec_exasmb_get_locks_request(val, p); }


uint32_t *enc_exasmb_time_val(exasmb_time_val *val, uint32_t *p);
uint32_t *dec_exasmb_time_val(exasmb_time_val *val, uint32_t *p);


uint32_t *enc_exasmb_shared_lock(const exasmb_shared_lock *val, uint32_t *p);
uint32_t *dec_exasmb_shared_lock(exasmb_shared_lock *val, uint32_t *p);
size_t size_exasmb_shared_lock(exasmb_shared_lock *val);


static inline uint32_t *enc_exasmb_share_get_locks_reply_header(exasmb_share_get_locks_reply *val, uint32_t *p)
	{ return enc_exasmb_reply_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_share_get_locks_reply_header(exasmb_share_get_locks_reply *val, uint32_t *p)
	{ return dec_exasmb_reply_vec_header(&val->header, p); }

uint32_t *enc_exasmb_share_get_locks_reply_body(exasmb_share_get_locks_reply *val, uint32_t *p);
uint32_t *dec_exasmb_share_get_locks_reply_body(exasmb_share_get_locks_reply *val, uint32_t *p);


static inline uint32_t *enc_exasmb_share_add_lock_request_header(exasmb_share_add_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_share_add_lock_request_header(exasmb_share_add_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_share_add_lock_request_body(exasmb_share_add_lock_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_add_lock_request_body(exasmb_share_add_lock_request *val, uint32_t *p);

static inline uint32_t *enc_exasmb_share_delete_lock_request_header(exasmb_share_delete_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_share_delete_lock_request_header(exasmb_share_delete_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_share_delete_lock_reply(exasmb_share_delete_lock_reply *val, uint32_t *p);
uint32_t *dec_exasmb_share_delete_lock_reply(exasmb_share_delete_lock_reply *val, uint32_t *p);

uint32_t *enc_exasmb_share_delete_lock_request_body(exasmb_share_delete_lock_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_delete_lock_request_body(exasmb_share_delete_lock_request *val, uint32_t *p);


uint32_t *enc_exasmb_share_delete_by_pid(exasmb_share_delete_by_pid *val, uint32_t *p);
uint32_t *dec_exasmb_share_delete_by_pid(exasmb_share_delete_by_pid *val, uint32_t *p);


static inline uint32_t *
enc_exasmb_share_delete_lock_by_pid_request_header(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *
dec_exasmb_share_delete_lock_by_pid_request_header(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_share_delete_lock_by_pid_request_body(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_delete_lock_by_pid_request_body(exasmb_share_delete_lock_by_pid_request *val, uint32_t *p);


uint32_t *enc_exasmb_share_set_delete_on_close_request(const exasmb_share_set_delete_on_close_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_set_delete_on_close_request(exasmb_share_set_delete_on_close_request *val, uint32_t *p);

uint32_t *enc_exasmb_share_modify_oplock_request(const exasmb_share_modify_oplock_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_modify_oplock_request(exasmb_share_modify_oplock_request *val, uint32_t *p);

uint32_t *enc_exasmb_share_get_delete_on_close_request(const exasmb_share_get_delete_on_close_request *val, uint32_t *p);
uint32_t *dec_exasmb_share_get_delete_on_close_request(exasmb_share_get_delete_on_close_request *val, uint32_t *p);

uint32_t *enc_exasmb_share_get_delete_on_close_reply(const exasmb_share_get_delete_on_close_reply *val, uint32_t *p);
uint32_t *dec_exasmb_share_get_delete_on_close_reply(exasmb_share_get_delete_on_close_reply *val, uint32_t *p);



/******************************************************************************/
/*                              Byte Range locks                              */
/******************************************************************************/


static inline uint32_t *enc_exasmb_br_lock_lr_request(exasmb_br_lock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_br_lock_lr_request(exasmb_br_lock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_br_unlock_lr_request(exasmb_br_unlock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_br_unlock_lr_request(exasmb_br_unlock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_br_get_locks_request(exasmb_br_get_locks_request *val, uint32_t *p)
	{ return enc_exasmb_get_locks_request(val, p); }
static inline uint32_t *dec_exasmb_br_get_locks_request(exasmb_br_get_locks_request *val, uint32_t *p)
	{ return dec_exasmb_get_locks_request(val, p); }


uint32_t *enc_exasmb_br_lock(exasmb_br_lock *val, uint32_t *p);
uint32_t *dec_exasmb_br_lock(exasmb_br_lock *val, uint32_t *p);


static inline uint32_t *enc_exasmb_br_get_locks_reply_header(exasmb_br_get_locks_reply *val, uint32_t *p)
	{ return enc_exasmb_reply_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_br_get_locks_reply_header(exasmb_br_get_locks_reply *val, uint32_t *p)
	{ return dec_exasmb_reply_vec_header(&val->header, p); }

uint32_t *enc_exasmb_br_get_locks_reply_body(exasmb_br_get_locks_reply *val, uint32_t *p);
uint32_t *dec_exasmb_br_get_locks_reply_body(exasmb_br_get_locks_reply *val, uint32_t *p);

static inline uint32_t *enc_exasmb_br_chk_lock_request_header(exasmb_br_chk_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_br_chk_lock_request_header(exasmb_br_chk_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_br_chk_lock_request_body(exasmb_br_chk_lock_request *val, uint32_t *p);
uint32_t *dec_exasmb_br_chk_lock_request_body(exasmb_br_chk_lock_request *val, uint32_t *p);


static inline uint32_t *enc_exasmb_br_add_lock_request_header(exasmb_br_add_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_br_add_lock_request_header(exasmb_br_add_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_br_add_lock_request_body(exasmb_br_add_lock_request *val, uint32_t *p);
uint32_t *dec_exasmb_br_add_lock_request_body(exasmb_br_add_lock_request *val, uint32_t *p);


static inline uint32_t *enc_exasmb_br_delete_lock_request_header(exasmb_br_delete_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_br_delete_lock_request_header(exasmb_br_delete_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *dec_exasmb_br_delete_lock_request_body(exasmb_br_delete_lock_request *val, uint32_t *p);
uint32_t *enc_exasmb_br_delete_lock_request_body(exasmb_br_delete_lock_request *val, uint32_t *p);


uint32_t *enc_exasmb_file_close_request(exasmb_file_close_request *val, uint32_t *p);
uint32_t *dec_exasmb_file_close_request(exasmb_file_close_request *val, uint32_t *p);


/******************************************************************************/
/*                              Block locks                                   */
/******************************************************************************/


static inline uint32_t *enc_exasmb_blk_lock_lr_request(exasmb_blk_lock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_blk_lock_lr_request(exasmb_blk_lock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_blk_unlock_lr_request(exasmb_blk_unlock_lr_request *val, uint32_t *p)
	{ return enc_exasmb_simple_request(val, p); }
static inline uint32_t *dec_exasmb_blk_unlock_lr_request(exasmb_blk_unlock_lr_request *val, uint32_t *p)
	{ return dec_exasmb_simple_request(val, p); }


static inline uint32_t *enc_exasmb_blk_get_locks_request(exasmb_get_locks_request *val, uint32_t *p)
	{ return enc_exasmb_get_locks_request(val, p); }
static inline uint32_t *dec_exasmb_blk_get_locks_request(exasmb_get_locks_request *val, uint32_t *p)
	{ return dec_exasmb_get_locks_request(val, p); }


uint32_t *enc_exasmb_blk_lock(exasmb_blk_lock *val, uint32_t *p);
uint32_t *dec_exasmb_blk_lock(exasmb_blk_lock *val, uint32_t *p);


static inline uint32_t *enc_exasmb_blk_get_locks_reply_header(exasmb_blk_get_locks_reply *val, uint32_t *p)
	{ return enc_exasmb_reply_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_blk_get_locks_reply_header(exasmb_blk_get_locks_reply *val, uint32_t *p)
	{ return dec_exasmb_reply_vec_header(&val->header, p); }

uint32_t *enc_exasmb_blk_get_locks_reply_body(exasmb_blk_get_locks_reply *val, uint32_t *p);
uint32_t *dec_exasmb_blk_get_locks_reply_body(exasmb_blk_get_locks_reply *val, uint32_t *p);


static inline uint32_t *enc_exasmb_blk_add_lock_request_header(exasmb_blk_add_lock_request *val, uint32_t *p)
	{ return enc_exasmb_request_vec_header(&val->header, p); }
static inline uint32_t *dec_exasmb_blk_add_lock_request_header(exasmb_blk_add_lock_request *val, uint32_t *p)
	{ return dec_exasmb_request_vec_header(&val->header, p); }

uint32_t *enc_exasmb_blk_add_lock_request_body(exasmb_blk_add_lock_request *val, uint32_t *p);
uint32_t *dec_exasmb_blk_add_lock_request_body(exasmb_blk_add_lock_request *val, uint32_t *p);


uint32_t *enc_exasmb_blk_delete_lock_by_fid_request(exasmb_blk_delete_lock_by_fid_request *val, uint32_t *p);
uint32_t *dec_exasmb_blk_delete_lock_by_fid_request(exasmb_blk_delete_lock_by_fid_request *val, uint32_t *p);


uint32_t *enc_exasmb_blk_delete_lock_by_mid_request(exasmb_blk_delete_lock_by_mid_request *val, uint32_t *p);
uint32_t *dec_exasmb_blk_delete_lock_by_mid_request(exasmb_blk_delete_lock_by_mid_request *val, uint32_t *p);


uint32_t *enc_exasmb_blk_check_timeout_request(exasmb_blk_check_timeout_request *val, uint32_t *p);
uint32_t *dec_exasmb_blk_check_timeout_request(exasmb_blk_check_timeout_request *val, uint32_t *p);

uint32_t *enc_exasmb_blk_check_timeout_reply(exasmb_blk_check_timeout_reply *val, uint32_t *p);
uint32_t *dec_exasmb_blk_check_timeout_reply(exasmb_blk_check_timeout_reply *val, uint32_t *p);

uint32_t *enc_exasmb_blk_delete_and_next_request(exasmb_blk_delete_and_next_request *val, uint32_t *p);
uint32_t *dec_exasmb_blk_delete_and_next_request(exasmb_blk_delete_and_next_request *val, uint32_t *p);

uint32_t *enc_exasmb_blk_delete_and_next_reply(exasmb_blk_delete_and_next_reply *val, uint32_t *p);
uint32_t *dec_exasmb_blk_delete_and_next_reply(exasmb_blk_delete_and_next_reply *val, uint32_t *p);

#endif

