#ifndef _H_SID_STRUCT
#define _H_SID_STRUCT
#include "sid_identifier_authority.h"
#ifdef __cplusplus
	#include <string>
#endif

#define SID_SIZE 15

// uint8 and uint32 are already defined in samba's includes.h, we don't want to redeclare them
#ifndef uint8
typedef unsigned char uint8;
#endif

#ifndef uint32
typedef unsigned int uint32;
#endif

struct sid_struct {

#ifdef __cplusplus
	sid_struct() : 
	_revision(0),
	_sub_authority_count(0){
	}
	sid_struct(const sid_struct& other);
	sid_struct& operator=(const sid_struct& sid);
	bool operator==(const sid_struct& other) const;
	friend std::ostream& operator<< (std::ostream&, const sid_struct&);
	bool        sid_from_string(const std::string& sidstr);
	std::string to_string() const;
	uint32_t    rid() const {return(_sub_authority_count>0) ? (uint32_t)_sub_authority[_sub_authority_count - 1] : 0;}
	void        set_rid(uint32_t urid) {if (_sub_authority_count > 0) _sub_authority[_sub_authority_count - 1] = urid;}
	uint32_t    hash()const;
#endif
	uint8		                    _revision;
	uint8        		            _sub_authority_count;
	struct sid_identifier_authority _identifier_authority;
	uint32               		    _sub_authority[SID_SIZE];
};


/* 
	This struct is the array that holds all the SIDs that are to be transfered to the VFS/ACL server.
	All the information stored in the NT_USER_TOKEN should be parsed and transfered into this struct
	On location [0] - The first SID in the array is the user's SID
	On location [1] - The second SID in the array is the user's primary group SID
	The rest are the SIDs of all the other groups the user belongs to. Note: this includes SID history !
*/
typedef struct {
	unsigned int num_SIDs;
	struct sid_struct* sid_array;//[GROUPS_PER_USER];
} SID_Array;

#endif	//_H_SID_STRUCT

