#include "smb_acl_message.h"
#include "xnt_xdr_misc.h"
#include "unixntacl_xdr.h"
#include <string.h>

std::ostream& operator<<(std::ostream& out, smb_acl_message_code code)
{
	switch (code) {
	case MESSAGE_CODE_NULL:
		out << "NULL";
		break;
	case MESSAGE_CODE_SET_ACL:
		out << "SET_ACL" ;
		break;
	case MESSAGE_CODE_GET_ACL:
		out << "GET_ACL" ;
		break;
	case MESSAGE_CODE_GET_ACCESS:
		out << "GET_ACCESS" ;
		break;
	case MESSAGE_CODE_SENDFILE:
		out << "SENDFILE" ;
		break;
	case MESSAGE_CODE_READ:
		out << "READ" ;
		break;
	case MESSAGE_CODE_SHWRITE:
		out << "SHWRITE" ;
		break;
	case MESSAGE_CODE_RENAME:
		out << "RENAME" ;
		break;
	case MESSAGE_CODE_FSYNC:
		out << "FSYNC" ;
		break;
	case MESSAGE_CODE_DIRPLUS:
		out << "DIRPLUS" ;
		break;
	case MESSAGE_CODE_GET_SNAPSHOTS_TIMESTAMPS:
		out << "GET_SNAPSHOTS_TIMESTAMPS" ;
		break;
	case MESSAGE_CODE_GET_SNAPSHOT_NAME_BY_TIMESTAMP:
		out << "GET_SNAPSHOT_NAME_BY_TIMESTAMP" ;
		break;
	case MESSAGE_CODE_GET_ANTIVIRUS_INFO:
		out << "MESSAGE_CODE_GET_ANTIVIRUS_INFO" ;
		break;
	case MESSAGE_CODE_SET_ANTIVIRUS_INFO:
		out << "MESSAGE_CODE_SET_ANTIVIRUS_INFO" ;
		break;
	case MESSAGE_CODE_SET_ANTIVIRUS_INFO_GUARDED:
		out << "MESSAGE_CODE_SET_ANTIVIRUS_INFO_GUARDED" ;
		break;
	case MESSAGE_CODE_MAP_SID_TO_TEMP_ID:
		out << "MESSAGE_CODE_MAP_SID_TO_TEMP_ID";
		break;
	case MESSAGE_CODE_MAP_TEMP_ID_TO_SID:
		out << "MESSAGE_CODE_MAP_TEMP_ID_TO_SID";
		break;
	default:
		out << "unknown smb acl message code";

	}
	return out;
}
 
const uint32_t *
decode_message(smb_acl_message *msg, const uint32_t *p)
{
	p = xnt_dec32(&msg->type, p);
	switch (msg->type) {
	case MESSAGE_TYPE_REQUEST:
		p = xnt_dec32((uint32_t*)&msg->code, p);
		break;
	case MESSAGE_TYPE_REPLY:
		p = xnt_dec32((uint32_t*)&msg->error, p);
		break;
	}
	return p;
}


const uint32_t *
decode_request(smb_acl_request *req, const uint32_t *p)
{
	p = decode_message(req, p);
	if (req->type != MESSAGE_TYPE_REQUEST)
		throw std::exception();
	p = xnt_dec32(&req->fh_len, p);
	req->fh_data = reinterpret_cast<char*>( const_cast<uint32_t*>(p) );
	p += req->fh_len / sizeof(*p);
	if (req->fh_len & 3)
		++p;
	return p;
}

uint32_t *
encode_request(smb_acl_message_code code, uint32_t *p, const smb_acl_request& req)
{
	p = xnt_enc32(MESSAGE_TYPE_REQUEST, p);
	p = xnt_enc32(code, p);
	p = xnt_enc32(req.fh_len, p);
	memcpy(p, req.fh_data, req.fh_len);
	if (req.fh_len & 3) {
		memset(reinterpret_cast<char*>(p) + req.fh_len, 0, 4 - (req.fh_len & 3));
		++p;
	}
	p += req.fh_len / sizeof(*p);
	return p;
}


uint32_t *
encode_reply(smb_acl_error error, uint32_t *p)
{
	p = xnt_enc32(MESSAGE_TYPE_REPLY, p);
	p = xnt_enc32(error, p);
	return p;
}

const uint32_t *
decode_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = xnt_dec32(&rep->type, p);
	p = xnt_dec32((uint32_t*)&rep->error, p);
	return p;
}
uint32_t *encode_get_acl_request(const smb_acl_request *req, uint32_t *p)
{
	return encode_request(MESSAGE_CODE_GET_ACL, p, *req);
}

uint32_t *
encode_get_acl_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	const smb_acl_get_acl_reply *get_rep =
		static_cast<const smb_acl_get_acl_reply *>(rep);
	p = encode_unixntacl_acl(get_rep->acl, p);
	return p;
}

const uint32_t *
decode_get_acl_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (rep->error != SMBACL_MESSAGE_OK)
		return p;
	smb_acl_get_acl_reply *get_rep =
		static_cast<smb_acl_get_acl_reply *>(rep);
	p = decode_unixntacl_acl(get_rep->acl, p);
	return p;
}

uint32_t *
encode_set_acl_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_SET_ACL, p, *req);
	const smb_acl_set_acl_request *set_req =
		static_cast<const smb_acl_set_acl_request *>(req);
	p = encode_unixntacl_acl(set_req->acl, p);
	return p;
}

const uint32_t *
decode_set_acl_request(smb_acl_request *req, const uint32_t *p)
{
	smb_acl_set_acl_request *set_req =
		static_cast<smb_acl_set_acl_request *>(req);
	p = decode_unixntacl_acl(set_req->acl, p);
	return p;
}

uint32_t *
encode_set_acl_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	return p;
}

const uint32_t *
decode_set_acl_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	return p;
}


uint32_t *encode_get_access_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_GET_ACCESS, p, *req);
	const smb_acl_get_access_request *get_req =
		static_cast<const smb_acl_get_access_request *>(req);
	p = xnt_enc32(get_req->uid, p);
	p = xnt_enc32(get_req->ngroups, p);
	for (uint32_t i = 0; i < get_req->ngroups; ++i) {
		p = xnt_enc32(get_req->groups[i], p);
	}
	return p;
}

const uint32_t *decode_get_access_request(smb_acl_request *req, const uint32_t *p)
{
	smb_acl_get_access_request *get_req =
		static_cast<smb_acl_get_access_request *>(req);
	p = xnt_dec32(&get_req->uid, p);
	p = xnt_dec32(&get_req->ngroups, p);
	for (uint32_t i = 0; i < get_req->ngroups; ++i) {
		p = xnt_dec32(&get_req->groups[i], p);
	}
	return p;
}

uint32_t *encode_get_access_reply(const smb_acl_message *rep, uint32_t *p)
{
	const smb_acl_get_access_reply *get_rep =
		static_cast<const smb_acl_get_access_reply*>(rep);

	p = encode_reply(rep->error, p);
	if (!rep->error) {
		p = xnt_enc32(get_rep->access, p);
		p = xnt_encode_string(get_rep->access_troubleshooting_details, p);

	}
	return p;
}

const uint32_t *decode_get_access_reply(smb_acl_message *rep, const uint32_t *p)
{
	smb_acl_get_access_reply *get_rep = 
		static_cast<smb_acl_get_access_reply*>(rep);

	p = decode_reply(rep, p);
	if (!rep->error) {
		p = xnt_dec32(&get_rep->access, p);
		p = xnt_decode_string( p, get_rep->access_troubleshooting_details);
	}
	return p;
}

