#include <sys/types.h>
#include <unistd.h>
#include "exasmb_vfsfs_message.h"
#include "smb_acl_message.h"
#include "xnt_xdr_misc.h"
#include <string.h>
#include <iostream>

#define XNT_USING_STD_CERR	// dumping errors to std::cerr
#include "xnt_error_logging.h"

/*const uint32_t *
decode_message(smb_acl_message *msg, const uint32_t *p)
{
	p = xnt_dec32(&msg->type, p);
	switch (msg->type) {
	case MESSAGE_TYPE_REQUEST:
		p = xnt_dec32(&msg->code, p);
		break;
	case MESSAGE_TYPE_REPLY:
		p = xnt_dec32(&msg->error, p);
		break;
	}
	return p;
}


const uint32_t *
decode_request(smb_acl_request *req, const uint32_t *p)
{
	p = decode_message(req, p);
	if (req->type != MESSAGE_TYPE_REQUEST)
		throw std::exception();
	p = xnt_dec32(&req->fh_len, p);
	req->fh_data = reinterpret_cast<char*>( const_cast<uint32_t*>(p) );
	p += req->fh_len / sizeof(*p);
	if (req->fh_len & 3)
		++p;
	return p;
}

uint32_t *
encode_request(smb_acl_message_code code, uint32_t *p, const smb_acl_request& req)
{
	p = xnt_enc32(MESSAGE_TYPE_REQUEST, p);
	p = xnt_enc32(code, p);
	p = xnt_enc32(req.fh_len, p);
	memcpy(p, req.fh_data, req.fh_len);
	if (req.fh_len & 3) {
		memset(reinterpret_cast<char*>(p) + req.fh_len, 0, 4 - (req.fh_len & 3));
		++p;
	}
	p += req.fh_len / sizeof(*p);
	return p;
}


uint32_t *
encode_reply(smb_vfsfs_error error, uint32_t *p)
{
	p = xnt_enc32(MESSAGE_TYPE_REPLY, p);
	p = xnt_enc32(error, p);
	return p;
}smb_vfs_rename_request

const uint32_t *
decode_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = xnt_dec32(&rep->type, p);
	p = xnt_dec32(&rep->error, p);
	return p;
}

*/

/***********************************************************************/					   
/*	Read request/reply*/
/***********************************************************************/

uint32_t *encode_read_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_READ, p, *req);
	const smb_vfs_readfile_request *read_req =
		static_cast<const smb_vfs_readfile_request *>(req);
	p = xnt_enc64(&read_req->offset, p);
	p = xnt_enc64(&read_req->count, p);
	p = xnt_enc32((uint32_t*)&read_req->shmid, p);
    p = xnt_enc32((uint32_t*)&read_req->id, p);
	return p;
}

const uint32_t *decode_read_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_readfile_request *read_req =
		static_cast<smb_vfs_readfile_request *>(req);
	p = xnt_dec64(&read_req->offset, p);
	p = xnt_dec64(&read_req->count, p);
	p = xnt_dec32((uint32_t*)&read_req->shmid, p);
    p = xnt_dec32((uint32_t*)&read_req->id, p);
	return p;
}


uint32_t *encode_read_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_readfile_reply *read_rep =
			static_cast<const smb_vfs_readfile_reply *>(rep);
		p = xnt_enc64(&read_rep->offset, p);
		p = xnt_enc64(&read_rep->count, p);
	}
	return p;
}
const uint32_t *decode_read_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_readfile_reply *read_rep =
		static_cast<smb_vfs_readfile_reply *>(rep);
		p = xnt_dec64(&read_rep->offset, p);
		p = xnt_dec64(&read_rep->count, p);
	}
	return p;
}

/***********************************************************************/					   
/*	Sendfile request/reply	 */
/***********************************************************************/

//uint32_t *encode_sendfile_request(const smb_vfsfs_request *req, uint32_t *p)
uint32_t *encode_sendfile_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_SENDFILE, p, *req);
	const smb_vfs_sendfile_request *sendfile_req =
		static_cast<const smb_vfs_sendfile_request *>(req);
	p = xnt_enc64(&sendfile_req->offset, p);
	p = xnt_enc64(&sendfile_req->count, p);
    p = xnt_enc32((uint32_t*)&sendfile_req->id, p);
	return p;
}

//const uint32_t *decode_sendfile_request(smb_vfsfs_request *req, const uint32_t *p)
const uint32_t *decode_sendfile_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_sendfile_request *sendfile_req =
		static_cast<smb_vfs_sendfile_request *>(req);
	p = xnt_dec64(&sendfile_req->offset, p);
	p = xnt_dec64(&sendfile_req->count, p);
    p = xnt_dec32((uint32_t*)&sendfile_req->id, p);
	return p;
}

//uint32_t *encode_sendfile_reply(const smb_vfsfs_message *rep, uint32_t *p)
uint32_t *encode_sendfile_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_sendfile_reply *sendfile_rep =
			static_cast<const smb_vfs_sendfile_reply *>(rep);
		p = xnt_enc64(&sendfile_rep->count, p);
	}
	return p;
}

//const uint32_t *decode_sendfile_reply(smb_vfsfs_message *rep, const uint32_t *p)
const uint32_t *decode_sendfile_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_sendfile_reply *sendfile_rep =
			static_cast<smb_vfs_sendfile_reply *>(rep);
        p = xnt_dec64(&sendfile_rep->count, p);
	}
	return p;
}

// shwrite request/reply
uint32_t *encode_shwrite_request(const smb_acl_request *req, uint32_t *p){
    
    p = encode_request(MESSAGE_CODE_SHWRITE, p, *req);
	const smb_vfs_shwrite_request *shwrite_req =
		static_cast<const smb_vfs_shwrite_request *>(req);
	p = xnt_enc64(&shwrite_req->offset, p);
    p = xnt_enc64(&shwrite_req->buff_offset, p);
	p = xnt_enc32(&shwrite_req->count, p);
    p = xnt_enc32((uint32_t*) &shwrite_req->shmid, p);
    p = xnt_enc32((uint32_t*)&shwrite_req->id, p);
    p = xnt_enc32((uint32_t*)&shwrite_req->unstable_flag, p);
    p = xnt_enc32((uint32_t*)&shwrite_req->id_must_exist, p);
	p = xnt_enc32((uint32_t*)&shwrite_req->update_mtime, p);
	return p;
}

const uint32_t *decode_shwrite_request(smb_acl_request *req, const uint32_t *p){
    smb_vfs_shwrite_request *shwrite_req =
		static_cast<smb_vfs_shwrite_request *>(req);
	p = xnt_dec64(&shwrite_req->offset, p);
    p = xnt_dec64(&shwrite_req->buff_offset, p);
	p = xnt_dec32(&shwrite_req->count, p);
    p = xnt_dec32((uint32_t*)&shwrite_req->shmid, p);
    p = xnt_dec32((uint32_t*)&shwrite_req->id, p);
    p = xnt_dec32((uint32_t*)&shwrite_req->unstable_flag, p);
    p = xnt_dec32((uint32_t*)&shwrite_req->id_must_exist, p);
	p = xnt_dec32((uint32_t*)&shwrite_req->update_mtime, p);
	return p;
}

uint32_t *encode_shwrite_reply(const smb_acl_message *rep, uint32_t *p){
    p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_shwrite_reply *shwrite_rep =
			static_cast<const smb_vfs_shwrite_reply *>(rep);
        p = xnt_enc32(&shwrite_rep->count, p);
		p = xnt_enc32((uint32_t*)&shwrite_rep->unstable_pending, p);
	}
	return p;
}

const uint32_t *decode_shwrite_reply(smb_acl_message *rep, const uint32_t *p){
    p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_shwrite_reply *shwrite_rep =
			static_cast<smb_vfs_shwrite_reply *>(rep);
        p = xnt_dec32(&shwrite_rep->count, p);
		p = xnt_dec32((uint32_t*)&shwrite_rep->unstable_pending, p);
	}
	return p;

}

/*rename*/
uint32_t *encode_rename_request(const smb_acl_request *req, uint32_t *p){
    XNT_WARN("encode_rename_request:enter");
    p = encode_request(MESSAGE_CODE_RENAME, p, *req);
    const smb_vfs_rename_request *rename_req =
		static_cast< const smb_vfs_rename_request *>(req);
    
    p = xnt_enc32(&rename_req->from_name_len, p);
    memcpy(p, rename_req->from_name, rename_req->from_name_len);
    p = ptr_plus_n_bytes( p, rename_req->from_name_len );
	p = xnt_enc32(&rename_req->to_name_len, p);
    memcpy(p, rename_req->to_name,rename_req->to_name_len );
	/* Add padding to 4 bytes */
	p = ptr_plus_n_bytes( p, rename_req->to_name_len );
    return p;
}

const uint32_t *decode_rename_request(smb_acl_request *req, const uint32_t *p){
    XNT_WARN("decode_rename_request:enter\n");
    smb_vfs_rename_request *rename_req =
	    static_cast<smb_vfs_rename_request *>(req);
    
    p = xnt_dec32(&rename_req->from_name_len, p);
    XNT_WARN("rename_req->from_name_len is " << rename_req->from_name_len);
	memcpy(rename_req->from_name, p, rename_req->from_name_len);
    rename_req->from_name[rename_req->from_name_len] = '\0';
    XNT_WARN("rename_req->from_name is " << rename_req->from_name);
	p = ptr_plus_n_bytes( p, rename_req->from_name_len );

    p = xnt_dec32(&rename_req->to_name_len, p);
    XNT_WARN("rename_req->to_name_len is " << rename_req->to_name_len);
	memcpy(rename_req->to_name, p, rename_req->to_name_len);
    rename_req->to_name[rename_req->to_name_len] = '\0';
    XNT_WARN("rename_req->to_name is " << rename_req->to_name);
	p = ptr_plus_n_bytes( p, rename_req->to_name_len );

	return p;
}


const uint32_t *decode_rename_reply(smb_acl_message *rep, const uint32_t *p){
    p = decode_reply(rep, p);
    return p;
}

uint32_t *encode_rename_reply(const smb_acl_message *rep, uint32_t *p){
    p = encode_reply(rep->error, p);
    return p;
}


// fsync request/reply
uint32_t *encode_fsync_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_FSYNC, p, *req);
	const smb_vfs_fsync_request *fsync_req =
		static_cast<const smb_vfs_fsync_request *>(req);
	p = xnt_enc32((uint32_t*)&fsync_req->id, p);
	return p;
}

const uint32_t *decode_fsync_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_fsync_request *fsync_req =
		static_cast<smb_vfs_fsync_request *>(req);
	p = xnt_dec32((uint32_t*)&fsync_req->id, p);
	return p;
}

uint32_t *encode_fsync_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	return p;
}

const uint32_t *decode_fsync_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	return p;
}


// read dir plus request/reply
uint32_t *encode_dirplus_request(const smb_acl_request *req, uint32_t *p){
    
	p = encode_request(MESSAGE_CODE_DIRPLUS, p, *req);
	const smb_vfs_dirplus_request *dirplus_req =
		static_cast<const smb_vfs_dirplus_request *>(req);
	p = xnt_enc64(&dirplus_req->cookie, p);
	p = xnt_enc32(&dirplus_req->size, p);
	p = xnt_enc32((uint32_t*)&dirplus_req->dir_shmid, p);
	return p;
}

const uint32_t *decode_dirplus_request(smb_acl_request *req, const uint32_t *p){
	smb_vfs_dirplus_request *dirplus_req =
		static_cast<smb_vfs_dirplus_request *>(req);
	p = xnt_dec64(&dirplus_req->cookie, p);
	p = xnt_dec32(&dirplus_req->size, p);
	p = xnt_dec32((uint32_t*)&dirplus_req->dir_shmid, p);
 	return p;
}

uint32_t *encode_dirplus_reply(const smb_acl_message *rep, uint32_t *p){
    p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_dirplus_reply *dirplus_rep =
			static_cast<const smb_vfs_dirplus_reply *>(rep);
        p = xnt_enc64(&dirplus_rep->cookie, p);
	}
	return p;
}

const uint32_t *decode_dirplus_reply(smb_acl_message *rep, const uint32_t *p){
    p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_dirplus_reply *dirplus_rep =
			static_cast<smb_vfs_dirplus_reply *>(rep);
        p = xnt_dec64(&dirplus_rep->cookie, p);
	}
	return p;

}

// get_snapshots_timestamps request/reply
uint32_t *encode_get_snapshots_timestamps_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_GET_SNAPSHOTS_TIMESTAMPS, p, *req);
	const smb_vfs_get_snapshots_timestamps_request* get_snapshots_timestamps_req =
		static_cast<const smb_vfs_get_snapshots_timestamps_request*>( req );
	p = xnt_encode_string( get_snapshots_timestamps_req->volume_name, p );
	p = xnt_enc32(get_snapshots_timestamps_req->max_number_of_snapshots, p);
	return p;
}

const uint32_t *decode_get_snapshots_timestamps_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_get_snapshots_timestamps_request* get_snapshots_timestamps_req =
		static_cast<smb_vfs_get_snapshots_timestamps_request*>( req );
	p = xnt_decode_string( p, get_snapshots_timestamps_req->volume_name );
	p = xnt_dec32(&(get_snapshots_timestamps_req->max_number_of_snapshots), p);
	return p;
}

uint32_t *encode_get_snapshots_timestamps_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_get_snapshots_timestamps_reply* get_snapshots_timestamps_rep =
			static_cast<const smb_vfs_get_snapshots_timestamps_reply*>( rep );
		p = xnt_enc32(get_snapshots_timestamps_rep->number_of_snapshots, p);
		for ( uint32_t i = 0 ; i < get_snapshots_timestamps_rep->number_of_snapshots ; ++i ) {
			p = xnt_enc32(get_snapshots_timestamps_rep->snapshots_timestamps[i], p);
		}
	}
	return p;
}

const uint32_t *decode_get_snapshots_timestamps_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_get_snapshots_timestamps_reply *get_snapshots_timestamps_rep =
			static_cast<smb_vfs_get_snapshots_timestamps_reply *>(rep);
		p = xnt_dec32(&(get_snapshots_timestamps_rep->number_of_snapshots), p);
		for ( uint32_t i = 0 ; i < get_snapshots_timestamps_rep->number_of_snapshots ; ++i ) {
			p = xnt_dec32(&(get_snapshots_timestamps_rep->snapshots_timestamps[i]), p);
		}
	}
	return p;
}

// get_snapshot_name_by_timestamp get_snapshot_name_by_timestamp request/reply
uint32_t *encode_get_snapshot_name_by_timestamp_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_GET_SNAPSHOT_NAME_BY_TIMESTAMP, p, *req);
	const smb_vfs_get_snapshot_name_by_timestamp_request* get_snapshot_name_by_timestamp_req =
		static_cast<const smb_vfs_get_snapshot_name_by_timestamp_request*>( req );
	p = xnt_encode_string( get_snapshot_name_by_timestamp_req->volume_name, p );
	p = xnt_enc32(get_snapshot_name_by_timestamp_req->snapshot_timestamp, p);
	return p;
}

const uint32_t *decode_get_snapshot_name_by_timestamp_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_get_snapshot_name_by_timestamp_request* get_snapshot_name_by_timestamp_req =
		static_cast<smb_vfs_get_snapshot_name_by_timestamp_request*>( req );
	p = xnt_decode_string( p, get_snapshot_name_by_timestamp_req->volume_name );
	p = xnt_dec32(&(get_snapshot_name_by_timestamp_req->snapshot_timestamp), p);
	return p;
}

uint32_t *encode_get_snapshot_name_by_timestamp_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_get_snapshot_name_by_timestamp_reply* get_snapshot_name_by_timestamp_rep =
			static_cast<const smb_vfs_get_snapshot_name_by_timestamp_reply*>( rep );
		p = xnt_encode_bool(get_snapshot_name_by_timestamp_rep->snapshot_found, p);
		if ( get_snapshot_name_by_timestamp_rep->snapshot_found ) {
			p = xnt_encode_string( get_snapshot_name_by_timestamp_rep->snapshot_name, p );
		}
	}
	return p;
}

const uint32_t *decode_get_snapshot_name_by_timestamp_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_get_snapshot_name_by_timestamp_reply *get_snapshot_name_by_timestamp_rep =
			static_cast<smb_vfs_get_snapshot_name_by_timestamp_reply *>(rep);
		p = xnt_decode_bool(&(get_snapshot_name_by_timestamp_rep->snapshot_found), p);
		if ( get_snapshot_name_by_timestamp_rep->snapshot_found ) {
			p = xnt_decode_string( p, get_snapshot_name_by_timestamp_rep->snapshot_name );
		}
	}
	return p;
}

uint32_t *encode_get_antivirus_info_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_GET_ANTIVIRUS_INFO, p, *req);
	return p;
}

const uint32_t *decode_get_antivirus_info_request(smb_acl_request *req, const uint32_t *p)
{
	return p;
}

uint32_t *encode_get_antivirus_info_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_get_antivirus_info_reply* get_antivirus_info_rep =
			static_cast<const smb_vfs_get_antivirus_info_reply*>( rep );
		p = xnt_enc32(get_antivirus_info_rep->val, p);
	}
	return p;
}

const uint32_t *decode_get_antivirus_info_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_get_antivirus_info_reply* get_antivirus_info_rep =
			static_cast<smb_vfs_get_antivirus_info_reply*>( rep );
		p = xnt_dec32(&(get_antivirus_info_rep->val), p);
	}
	return p;
}

uint32_t *encode_set_antivirus_info_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_SET_ANTIVIRUS_INFO, p, *req);
	const smb_vfs_set_antivirus_info_request *set_antivirus_info_req =
		static_cast<const smb_vfs_set_antivirus_info_request *>(req);
	p = xnt_enc32(set_antivirus_info_req->val, p);
	return p;
}

const uint32_t *decode_set_antivirus_info_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_set_antivirus_info_request *set_antivirus_info_req =
		static_cast<smb_vfs_set_antivirus_info_request *>(req);
	p = xnt_dec32(&(set_antivirus_info_req->val), p);
	return p;
}

uint32_t *encode_set_antivirus_info_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	return p;
}

const uint32_t *decode_set_antivirus_info_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	return p;
}

uint32_t *encode_set_antivirus_info_guarded_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_SET_ANTIVIRUS_INFO_GUARDED, p, *req);
	const smb_vfs_set_antivirus_info_guarded_request *set_antivirus_info_guarded_req =
		static_cast<const smb_vfs_set_antivirus_info_guarded_request *>(req);
	p = xnt_enc32(set_antivirus_info_guarded_req->val, p);
	p = xnt_enc32(set_antivirus_info_guarded_req->guard, p);
	return p;
}

const uint32_t *decode_set_antivirus_info_guarded_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_set_antivirus_info_guarded_request *set_antivirus_info_guarded_req =
		static_cast<smb_vfs_set_antivirus_info_guarded_request *>(req);
	p = xnt_dec32(&(set_antivirus_info_guarded_req->val), p);
	p = xnt_dec32(&(set_antivirus_info_guarded_req->guard), p);
	return p;
}

uint32_t *encode_set_antivirus_info_guarded_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_set_antivirus_info_guarded_reply* set_antivirus_info_guarded_rep =
			static_cast<const smb_vfs_set_antivirus_info_guarded_reply*>( rep );
		p = xnt_enc32(set_antivirus_info_guarded_rep->val, p);
	}
	return p;
}

const uint32_t *decode_set_antivirus_info_guarded_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_set_antivirus_info_guarded_reply* set_antivirus_info_guarded_rep =
			static_cast<smb_vfs_set_antivirus_info_guarded_reply*>( rep );
		p = xnt_dec32(&(set_antivirus_info_guarded_rep->val), p);
	}
	return p;
}

uint32_t* encode_map_sid_to_temp_id_request( const smb_acl_request* p_acl_req, uint32_t* p_buffer )
{
	p_buffer = encode_request( MESSAGE_CODE_MAP_SID_TO_TEMP_ID, p_buffer, *p_acl_req );
	const smb_vfs_map_sid_to_temp_id_request* p_req =
		static_cast<const smb_vfs_map_sid_to_temp_id_request*>( p_acl_req );
	p_buffer = xnt_encode_sid( p_req->sid, p_buffer );
	return p_buffer;
}

const uint32_t* decode_map_sid_to_temp_id_request( smb_acl_request* p_acl_req, const uint32_t* p_buffer )
{
	smb_vfs_map_sid_to_temp_id_request* p_req =
		static_cast<smb_vfs_map_sid_to_temp_id_request*>( p_acl_req );
	p_buffer = xnt_decode_sid( p_buffer, p_req->sid );
	return p_buffer;
}

uint32_t* encode_map_sid_to_temp_id_reply( const smb_acl_message* acl_rep, uint32_t* p_buffer )
{
	p_buffer = encode_reply( acl_rep->error, p_buffer );
	if ( acl_rep->error == SMBACL_MESSAGE_OK ) {
		const smb_vfs_map_sid_to_temp_id_reply* rep =
			static_cast<const smb_vfs_map_sid_to_temp_id_reply*>( acl_rep );
		p_buffer = xnt_enc32( rep->temp_id, p_buffer );
	}
	return p_buffer;
}

const uint32_t* decode_map_sid_to_temp_id_reply( smb_acl_message* acl_rep, const uint32_t* p_buffer )
{
	p_buffer = decode_reply( acl_rep, p_buffer );
	if ( acl_rep->error == SMBACL_MESSAGE_OK ) {
		smb_vfs_map_sid_to_temp_id_reply* rep =
			static_cast<smb_vfs_map_sid_to_temp_id_reply*>( acl_rep );
		p_buffer = xnt_dec32( &(rep->temp_id), p_buffer );
	}
	return p_buffer;
}

uint32_t* encode_map_temp_id_to_sid_request( const smb_acl_request* p_acl_req, uint32_t* p_buffer )
{
	p_buffer = encode_request( MESSAGE_CODE_MAP_TEMP_ID_TO_SID, p_buffer, *p_acl_req );
	const smb_vfs_map_temp_id_to_sid_request* p_req =
		static_cast<const smb_vfs_map_temp_id_to_sid_request*>( p_acl_req );
	p_buffer = xnt_enc32( p_req->temp_id, p_buffer );
	return p_buffer;
}

const uint32_t* decode_map_temp_id_to_sid_request( smb_acl_request* p_acl_req, const uint32_t* p_buffer )
{
	smb_vfs_map_temp_id_to_sid_request* p_req =
		static_cast<smb_vfs_map_temp_id_to_sid_request*>( p_acl_req );
	p_buffer = xnt_dec32( &(p_req->temp_id), p_buffer );
	return p_buffer;
}

uint32_t* encode_map_temp_id_to_sid_reply( const smb_acl_message* acl_rep, uint32_t* p_buffer )
{
	p_buffer = encode_reply( acl_rep->error, p_buffer );
	if ( acl_rep->error == SMBACL_MESSAGE_OK ) {
		const smb_vfs_map_temp_id_to_sid_reply* rep =
			static_cast<const smb_vfs_map_temp_id_to_sid_reply*>( acl_rep );
		p_buffer = xnt_encode_sid( rep->sid, p_buffer );
	}
	return p_buffer;
}

const uint32_t* decode_map_temp_id_to_sid_reply( smb_acl_message* acl_rep, const uint32_t* p_buffer )
{
	p_buffer = decode_reply( acl_rep, p_buffer );
	if ( acl_rep->error == SMBACL_MESSAGE_OK ) {
		smb_vfs_map_temp_id_to_sid_reply* rep =
			static_cast<smb_vfs_map_temp_id_to_sid_reply*>( acl_rep );
		p_buffer = xnt_decode_sid( p_buffer, rep->sid );
	}
	return p_buffer;
}

// statfs request/reply
uint32_t *encode_statfs_request(const smb_acl_request *req, uint32_t *p){
    
	p = encode_request(MESSAGE_CODE_STATFS, p, *req);
	return p;
}

const uint32_t *decode_statfs_request(smb_acl_request *req, const uint32_t *p){
//	nothing to decode...
    return p;
}

uint32_t *encode_statfs_reply(const smb_acl_message *rep, uint32_t *p){
    p = encode_reply(rep->error, p);
	if (!rep->error) {
		const smb_vfs_statfs_reply *statfs_rep =
			static_cast<const smb_vfs_statfs_reply *>(rep);
        p = xnt_enc64(&statfs_rep->bsize, p);
		p = xnt_enc64(&statfs_rep->dfree, p);
		p = xnt_enc64(&statfs_rep->dsize, p);
	}
	return p;
}

const uint32_t *decode_statfs_reply(smb_acl_message *rep, const uint32_t *p){
    p = decode_reply(rep, p);
	if (!rep->error) {
		smb_vfs_statfs_reply *statfs_rep =
			static_cast<smb_vfs_statfs_reply *>(rep);
        p = xnt_dec64(&statfs_rep->bsize, p);
		p = xnt_dec64(&statfs_rep->dfree, p);
		p = xnt_dec64(&statfs_rep->dsize, p);
	}
	return p;

}


uint32_t *encode_set_create_time_request(const smb_acl_request *req, uint32_t *p)
{
	p = encode_request(MESSAGE_CODE_SET_CREATE_TIME, p, *req);
	const smb_vfs_set_create_time_request *set_createtime_req =
		static_cast<const smb_vfs_set_create_time_request *>(req);
	
	p = xnt_enc32(set_createtime_req->create_time_sec, p);
	p = xnt_enc32(set_createtime_req->create_time_nsec, p);
	return p;

}

const uint32_t *decode_set_create_time_request(smb_acl_request *req, const uint32_t *p)
{
	smb_vfs_set_create_time_request *set_createtime_req =
		static_cast<smb_vfs_set_create_time_request *>(req);
	p = xnt_dec32(&set_createtime_req->create_time_sec, p);
	p = xnt_dec32(&set_createtime_req->create_time_nsec, p);
	return p;
}

uint32_t *encode_set_create_time_reply(const smb_acl_message *rep, uint32_t *p)
{
	p = encode_reply(rep->error, p);
    return p;
}

const uint32_t *decode_set_create_time_reply(smb_acl_message *rep, const uint32_t *p)
{
	p = decode_reply(rep, p);
    return p;
}
