#include <sys/types.h>
#include <netinet/in.h>
#include <string.h>
#include <byteswap.h>

#include "unixntacl_xdr.h"
/******************************************************************************/
/*                          Basic types decode/encode functions			 	  */
/******************************************************************************/

#define xdr_sizeof(x) (sizeof(x) > sizeof(uint32_t) ? \
	sizeof(x) : sizeof(uint32_t))

static inline const uint32_t *
dec16(uint16_t *val, const uint32_t *p)
{
    *val = (uint16_t) ntohl(*p);

    return p + 1;
}

static inline uint32_t *
enc16(const uint16_t *val, uint32_t *p)
{
    *p = htonl((uint32_t) *val);

    return p + 1;
}

static inline const uint32_t *
dec32(uint32_t *val, const uint32_t *p)
{
    *val = ntohl(*p);

    return p + 1;
}

static inline uint32_t *
enc32(const uint32_t *val, uint32_t *p)
{
    *p = htonl(*val);

    return p + 1;
}

static inline const uint32_t *
dec64(uint64_t *val, const uint32_t *p)
{
#ifdef WORDS_BIGENDIAN
    *val = *(uint64_t *) p;
#else
    uint32_t *val32 = (uint32_t *)val;
    *val32++ = ntohl(p[1]);
    *val32 = ntohl(p[0]);
#endif
    return p + 2;
}

static inline uint32_t *
enc64(const uint64_t *val, uint32_t *p)
{
#ifdef WORDS_BIGENDIAN
    *(uint64_t *) p = *val;
#else
    uint32_t *val32 = (uint32_t *)val;
    p[1] = htonl(*val32); val32++;
    p[0] = htonl(*val32);
#endif
    return p + 2;
}

/*****************************************************************************/

inline uint32_t *
encode_ace_type(const ace_type *x, uint32_t *p)
{
	*p = htonl(*x);
	
	return p + 1;
}

inline const uint32_t *
decode_ace_type(ace_type *x, const uint32_t *p)
{
	*x = (ace_type)ntohl(*p);
	
	return p + 1;
}

inline uint32_t *
encode_trustee_type(const trustee_type *x, uint32_t *p)
{
	*p = htonl(*x);
	
	return p + 1;
}

inline const uint32_t *
decode_trustee_type(trustee_type *x, const uint32_t *p)
{
	*x = (trustee_type) ntohl(*p);
	
	return p + 1;
}

inline uint32_t *
encode_unixntacl_acl_ace(const unixntacl_acl_ace *x, uint32_t *p)
{
	p = encode_ace_type(&x->type, p);
	p = enc32(&x->flags, p);
	p = enc32(&x->mask, p);
	p = enc32(&x->trustee.id, p);
	return p;
}

inline const uint32_t *
decode_unixntacl_acl_ace(unixntacl_acl_ace *x, const uint32_t *p)
{
	p = decode_ace_type(&x->type, p);
	p = dec32(&x->flags, p);
	p = dec32(&x->mask, p);
	p = dec32(&x->trustee.id, p);
	return p;
}

inline uint32_t *
encode_unixntacl_acl_header(const unixntacl_acl *x, uint32_t *p)
{
	p = enc32(&x->type, p);
	p = enc32(&x->ace_count, p);
	return p;
}

inline uint32_t *
encode_unixntacl_acl_aces(const unixntacl_acl *x, uint32_t *p)
{
	int i;
	for (i = 0; i < x->ace_count; ++i) {
		p = encode_unixntacl_acl_ace(&x->aces[i], p);
	}
	return p;
}

uint32_t *
encode_unixntacl_acl(const unixntacl_acl *x, uint32_t *p)
{
	p = encode_unixntacl_acl_header(x, p);
	p = encode_unixntacl_acl_aces(x, p);
	return p;
}


const uint32_t *
decode_unixntacl_acl_header(unixntacl_acl *x, const uint32_t *p)
{
	p = dec32(&x->type, p);
	p = dec32(&x->ace_count, p);
	return p;
}

const uint32_t *
decode_unixntacl_acl_aces(unixntacl_acl *x, const uint32_t *p)
{
	int i;
	for (i = 0; i < x->ace_count; ++i) {
		p = decode_unixntacl_acl_ace(&x->aces[i], p);
	}
	return p;
}

const uint32_t *
decode_unixntacl_acl(unixntacl_acl *x, const uint32_t *p)
{
	p = decode_unixntacl_acl_header(x, p);
	p = decode_unixntacl_acl_aces(x, p);
	return p;
}

